from struct import pack
from atop.tpython import NamedProperty

class MetaFlags(type):
    def __new__(C, name, bases, attrs):
        mask = 1l
        for flagname in attrs['flags']:
            attrs[flagname] = Flag(flagname, mask)
            mask <<= 1
        T = type.__new__(C, name,bases,attrs)
        return T


class Flag(object):
    def __init__(self, flagname, mask):
        self.flagname = flagname
        self.mask = mask

    def __get__(self, oself, type=None):
        return bool(oself.getBits(self.mask))

    def __set__(self, oself, value, type=None):
        oself.setBits(self.mask, value)

class Flags:
    __metaclass__ = MetaFlags
    flags = []
    _flags = 0

    def getBits(self, mask):
        return self._flags & mask

    def setBits(self, mask, on=True):
        if on:
            self._flags |= mask
        else:
            self._flags &= ~mask

    def __iter__(self):
        for flag in self.flags:
            yield getattr(self, flag)

    def __len__(self):
        return len(self.flags)

    def __repr__(self):
        bytes = [getattr(self, flag) and '1' or '0' for flag in self.flags]
        bytes.reverse()
        return '<%s %s>' % (self.__class__.__name__,
                            ''.join(bytes))

    def pack(self):
        return packint(self._flags)


def iterint(integer, base=256):
    remaining = integer
    while remaining:
        remaining, value = divmod(remaining, base)
        yield value


def packint(integer, bigendian=True):
    packlist = list(iterint(integer))
    if packlist:
        packlist.reverse()
    return ''.join(map(chr, packlist))


def unpackint(S, bigendian=True):
    base = 256
    if bigendian:
        S = list(S)
        S.reverse()
    number = 0
    for offset, character in enumerate(S):
        number += ord(character) * (base ** offset)
    return number

class StoredFlagsWrapper(Flags):
    flags = ()
    def _flags():
        def get(self):
            return getattr(self.storable, self.intStorageName)
        def set(self, value):
            return setattr(self.storable, self.intStorageName, value)
        return get, set
    _flags = property(*_flags())

    def __init__(self, storable, intStorageName='_flags'):
        self.storable = storable
        self.intStorageName = intStorageName

class StoredFlags(NamedProperty):
    def __init__(self, *flagnames):
        NamedProperty.__init__(self)
        class SFW(StoredFlagsWrapper):
            flags = flagnames
        self.sfw = SFW
    def __get__(self, oself, type=None):
        name = self.getName(oself)
        oname = self.prefix + name
        if not hasattr(oself, oname):
            setattr(oself, oname, 0)
        return self.sfw(oself, oname)
    def __set__(self, oself, value, type=None):
        name = self.getName(oself)
        raise AttributeError("%s.%s is a read-only variable" % (self.classObj.__name__, name))


def tests():
    class SampleFlags(Flags):
        flags = ['one',
                 'two',
                 'three',
                 'four']


    class AllFlags(Flags):
        flags = ['_%d' %x  for x in range(32)]

    class GotSomeFlags:
        _named_flags = 0
        flags = StoredFlags('one', 'two', 'three', 'four')

    foo = GotSomeFlags()

