# -*- test-case-name: atop.test.test_credup -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

from twisted.python.components import implements as doesImplement, Interface
from twisted.cred.portal import IRealm
from twisted.cred.credentials import IUsernamePassword, IUsernameHashedPassword
from twisted.cred.checkers import ICredentialsChecker, ANONYMOUS
from twisted.cred.error import UnauthorizedLogin
from twisted.python import log

from atop.ifcompat import backwardsCompatImplements, implements
from atop.powerup import Storeup
from atop.store import SubStore, Pool, CaseInsensitiveNameIndex, referenced, indexed, INDEX
from atop.fwd import ItemForwardingComponent
from atop import iatop

def _localpartDomainToDBEmail(localpart, domain):
    """Format a username and domain name suitably for use as a database key.
    """
    dparts = domain.split('.')
    dparts.reverse()
    revdomain = '.'.join(dparts)
    return (revdomain+'@'+localpart)

class IAvatar(Interface):
    pass

class NoSuchDomain(UnauthorizedLogin):
    pass

class NoSuchUser(UnauthorizedLogin):
    pass

class BadCredentials(UnauthorizedLogin):
    pass

class UsernameIndex(CaseInsensitiveNameIndex):
    pass

class DomainIndex(CaseInsensitiveNameIndex):
    def keyToSortableString(key):
        return '.'.join(CaseInsensitiveNameIndex.keyToSortableString(key).split('.')[::-1])

    def sortableStringToKey(sstr, pool):
        return '.'.join(sstr.split('.')[::-1])

class Avatar(SubStore):
    implements(IAvatar)

    name = indexed(UsernameIndex)

    securePassword = None
    insecurePassword = None

    def setSecurePassword(self, passwd):
        """Save away a string for use with protocols that do not expose your
        password on the network.
        """
        self.touch()
        self.securePassword = passwd

    def setInsecurePassword(self, passwd):
        """Save away a string for use with protocols that expose your password
        on the internet.
        """
        self.touch()
        self.insecurePassword = passwd

    def getPassword(self, passwordSecure):
        """Retrieve a password.
        """
        if passwordSecure:
            return self.securePassword
        else:
            return self.insecurePassword

    def getDomains(self):
        return [d for d in self.iterpools() if isinstance(d, Domain)]

    def __str__(self):
        return '<Avatar %r at %d in %s>' % (self.name, self.storeID, self.store)

backwardsCompatImplements(Avatar)
# XXX HACK for compyCompat compatibility
from twisted.python.components import registerAdapter
from atop.powerup import PowerStation, IPowerStation
registerAdapter(PowerStation, Avatar, IPowerStation)
from atop.faraway import IFarIndex, farIndexForSubStore
registerAdapter(farIndexForSubStore, Avatar, IFarIndex)


class AnonymousUser:
    ## Maybe this is pointless
    ## Maybe the adapter being registered on this class
    ## Should just be registered on the Store itself
    def __init__(self, store):
        self.store = store

class DomainPool(Pool):
    def __init__(self, store, name):
        Pool.__init__(self, store, name)
        self.addIndex(DomainIndex)

class Domain(Pool):
    name = indexed(DomainIndex)

    def __init__(self, store, name):
        Pool.__init__(self, store, name)
        self.addIndex(UsernameIndex)

class Preauthenticated:
    implements(IUsernameHashedPassword)

    def __init__(self, username):
        self.username = username

    def checkPassword(self, password):
        return True
backwardsCompatImplements(Preauthenticated)
def preauth(store, username):
    """Utility function to retrieve an avatar from a database by name.
    """
    return store.getItemByID(
        ICredentialsChecker(store).
        requestAvatarId(Preauthenticated(username)))

class FwdChecker(ItemForwardingComponent):
    implements(ICredentialsChecker)
    def __init__(self, fwdTo):
        ItemForwardingComponent.__init__(self, fwdTo)
        self.credentialInterfaces = fwdTo.getItem().credentialInterfaces
backwardsCompatImplements(FwdChecker)
class FwdRealm(ItemForwardingComponent):
    implements(IRealm)
backwardsCompatImplements(FwdRealm)
class StoreCredentialsChecker(Storeup):
    implements(ICredentialsChecker)

    powerupName = "Credentials Checker"

    credentialInterfaces = (IUsernamePassword, IUsernameHashedPassword)

    domains = referenced

    def setUpPools(self, store):
        self.domains = DomainPool(store, 'domains')
        store.setComponent(ICredentialsChecker, self)

    def deleteFromStore(self):
        self.store.unsetComponent(ICredentialsChecker)
        return Storeup.deleteFromStore(self)

    def getDomains(self):
        """Retrieve an iterator of all domains in this store.
        """
        return self.domains.queryIndex(index=DomainIndex)

    def getAvatars(self):
        """Retrieve an iterator of all avatars regardless of domain.

        This will yield certain avatars twice.
        """
        for domain in self.getDomains():
            for avatar in domain.queryIndex():
                yield avatar

    def getDomain(self, name=None):
        dp = self.domains
        for dom in dp.queryIndex(DomainIndex, startKey=name):
            if name is None:
                return dom
            if name.lower() == dom.name.lower():
                return dom
        else:
            raise NoSuchDomain(name)

    def addDomain(self, domainName):
        dp = self.domains
        x = dp.getIndexItem(DomainIndex, domainName, None)
        if x is None:
            x = Domain(self.store, domainName)
            dp.addItem(x)
        return x

    def addAvatar(self, avatar, domainName=None):
        def _():
            domain = self.getDomain(domainName)
            av2 = domain.getIndexItem(UsernameIndex, avatar.name, None)
            if av2 is not None:
                raise UnauthorizedLogin("You can't choose that username in that domain.")
            domain.addItem(avatar)
        return self.store.transact(_)
    addAvatarToDomain = addAvatar

    def getAvatar(self, localpart, domainName=None):
        def _():
            domain = self.getDomain(domainName)
            try:
                return domain.getIndexItem(UsernameIndex, localpart)
            except KeyError:
                raise NoSuchUser(localpart, domainName)
        return self.store.transact(_)
    getAvatarFromDomain = getAvatar

    def requestAvatarId(self, credentials):
        passwordSecure = doesImplement(credentials, IUsernameHashedPassword)
        usplit = credentials.username.split('@', 1)
        avatar = self.getAvatarFromDomain(*usplit)
        password = avatar.getPassword(passwordSecure)
        if credentials.checkPassword(password):
            return avatar.storeID
        else:
            raise BadCredentials()

backwardsCompatImplements(StoreCredentialsChecker)

class StoreRealm(Storeup):
    implements(IRealm)

    powerupName = "Realm"
    dependencies = [StoreCredentialsChecker]
    anonymousFactory = AnonymousUser

    def setUpPools(self, store):
        store.setComponent(IRealm, self)

    def deleteFromStore(self):
        self.store.unsetComponent(IRealm)
        return Storeup.deleteFromStore(self)

    ##
    ## Actual IRealm implementation
    ##

    def requestAvatar(self, avatarId, mind, *interfaces):
        """
        """
        def _():
            if avatarId is ANONYMOUS:
                avatar = self.anonymousFactory(self.store)
            else:
                avatar = self.store.getItemByID(avatarId)
            for iface in interfaces:
                avatarAspect = iface(avatar, default=None)
                if avatarAspect is None:
                    continue

                logout = getattr(avatarAspect, 'logout', noLogout)
                log.msg(interface=iatop.IStatEvent, name='auth', stat_authorized_logins=1, login=avatarId, iface=iface, mind=mind)
                return (iface, avatarAspect, logout)
            log.msg(interface=iatop.IStatEvent, name='auth', stat_unauthorized_logins=1, login=None, iface=iface, mind=mind)
            raise UnauthorizedLogin("You're not allowed to access that.")
        return self.store.transact(_)

backwardsCompatImplements(StoreRealm)

def noLogout():
    pass
