# -*- test-case-name: atop.test.test_notify -*-
"""Notifying proxies of mutable sequence and mapping types.
"""

import itertools

def passthru(x):
    return x

class NotifyList(object):
    __slots__ = ['_data',
                 '_added',
                 '_removed',
                 '_nodups',
                 '_loadValue',
                 '_storeValue']

    def __init__(self, data, added, removed, nodups=True,
                 storeValue=None, loadValue=None):
        """
        data: the list to proxy for
        added: a one-arg function to call when an item is added
        removed: a one-arg function to call when an item is removed

        nodups: don't give removed notifications if the item is still in the
        list somewhere else, and don't give added notifications if the item was
        in the list before.
        """
        self._data = data
        self._added = added
        self._removed = removed
        self._nodups = nodups
        if loadValue is None:
            loadValue = passthru
        if storeValue is None:
            storeValue = passthru
        self._storeValue = storeValue
        self._loadValue = loadValue

    # Mutation methods

    def _addnotify(self, svi):
        """Called before the list is modified.  Should I notify about this item
        being added after it's been done?
        """
        if self._nodups:
            return svi not in self._data
        else:
            return True

    def _multiadd(self, other):
        if not self._nodups:
            return other
        L = []
        for item in other:
            svi = self._storeValue(item)
            if svi not in self._data and item not in L:
                L.append(item)
        return L

    def _remnotify(self, item):
        if (not self._nodups) or not (self._storeValue(item) in self._data):
            self._removed(item)

    def append(self, item):
        svi = self._storeValue(item)
        notify = self._addnotify(svi)
        self._data.append(svi)
        if notify:
            self._added(item)

    def insert(self, i, item):
        svi = self._storeValue(item)
        notify = self._addnotify(svi)
        self._data.insert(i, svi)
        if notify:
            self._added(item)

    def pop(self, i=-1):
        svi = self._data.pop(i)
        item = self._loadValue(svi)
        self._remnotify(item)
        return item

    def remove(self, item):
        svi = self._storeValue(item)
        self._data.remove(svi)
        self._remnotify(item)

    def extend(self, other):
        tonotify = self._multiadd(other)
        self._data.extend(map(self._storeValue, other))
        for item in tonotify:
            self._added(item)

    def __imul__(self, n):
        olen = len(self._data)
        self._data *= n
        if not self._nodups:
            for svi in self._data[0:olen]:
                self._added(self._loadValue(svi))
        return self

    def __iadd__(self, other):
        self.extend(other)
        return self

    def __setitem__(self, i, item):
        oldsvi = self._data[i]
        olditem = self._loadValue(oldsvi)
        newsvi = self._storeValue(item)
        notify = self._addnotify(newsvi)
        self._data[i] = newsvi
        self._remnotify(olditem)
        if notify:
            self._added(item)

    def __delitem__(self, i):
        svi = self._data[i]
        item = self._loadValue(svi)
        del self._data[i]
        self._remnotify(item)

    def __delslice__(self, i, j):
        i = max(i, 0); j = max(j, 0)
        removed = map(self._loadValue, self._data[i:j])
        del self._data[i:j]
        for item in removed:
            self._remnotify(item)

    def __setslice__(self, i, j, other):
        i = max(i, 0); j = max(j, 0)
        removedsvi = self._data[i:j]
        removed = map(self._loadValue, removedsvi)
        added = self._multiadd(other)
        if self._nodups:
            nremoved = []
            for item in removed:
                if item not in other:
                    nremoved.append(item)
            removed = nremoved
        self._data[i:j] = map(self._storeValue, other)
        for item in added:
            self._added(item)
        for item in removed:
            self._removed(item)


    ######## Idempotent methods


    def __repr__(self):
        L = map(self._loadValue, self._data)
        return 'NotifyList'+repr(L)+''

    def count(self, item):
        return self._data.count(self._storeValue(item))

    def index(self, item, *args):
        return self._data.index(self._storeValue(item), *args)

    def reverse(self):
        self._data.reverse()

    def sort(self, *args):
        self._data.sort(*args)

    def __lt__(self, other):
        return self._data <  self.__cast(other)

    def __le__(self, other):
        return self._data <= self.__cast(other)

    def __eq__(self, other):
        return self._data == self.__cast(other)

    def __ne__(self, other):
        return self._data != self.__cast(other)

    def __gt__(self, other):
        return self._data >  self.__cast(other)

    def __ge__(self, other):
        return self._data >= self.__cast(other)

    def __cast(self, other):
        if isinstance(other, NotifyList) and other._storeValue == self._storeValue:
            return other._data
        else:
            return map(self._storeValue,other)

    def __cmp__(self, other):
        return cmp(self._data, self.__cast(other))

    def __contains__(self, item):
        return self._storeValue(item) in self._data

    def __len__(self):
        return len(self._data)

    def __getitem__(self, i):
        return self._loadValue(self._data[i])

    def __getslice__(self, i, j):
        i = max(i, 0); j = max(j, 0)
        return map(self._loadValue,self._data[i:j])

    def __add__(self, other):
        if isinstance(other, NotifyList) and other._storeValue == self._storeValue:
            return NotifyList(self._data + other._data, passthru, passthru, loadValue = self._loadValue, storeValue = self._storeValue)
        else:
            L = map(self._loadValue, self._data)
            L.extend(other)
            return L

    def __radd__(self, other):
        if isinstance(other, NotifyList) and other._storeValue == self._storeValue:
            return NotifyList(other._data + self._data, passthru, passthru, loadValue = self._loadValue, storeValue = self._storeValue)
        else:
            L = list(other)
            L.extend(iter(self))
            return L

    def __mul__(self, n):
        return self._data*n
    __rmul__ = __mul__

    def __iter__(self):
        for svi in self._data:
            yield self._loadValue(svi)


class NotifyDict(object):
    __slots__ = ['_data', '_added', '_removed']

    def __init__(self, data, added, removed):
        """
        data: the list to proxy for
        added: a two-arg function to call when a key/value pair is added
        removed: a two-arg function to call when a key/value pair is removed
        """
        self._data = data
        self._added = added
        self._removed = removed

    def __repr__(self):
        return repr(self._data)

    # Mutating methods.

    def __setitem__(self, key, item):
        if self.has_key(key):
            doremove = True
            removed = self._data[key]
        else:
            doremove = False
        self._data[key] = item
        if doremove:
            self._removed(key, removed)
        self._added(key, item)

    def update(self, dict):
        if isinstance(dict, NotifyDict):
            dict = dict._data
        for k, v in dict.iteritems():
            self[k] = v

    def __delitem__(self, key):
        self._removed(key, self._data.pop(key))

    def clear(self):
        for k in self._data.keys():
            del self[k]

    def setdefault(self, key, failobj=None):
        if not self.has_key(key):
            self[key] = failobj
        return self[key]

    _nopop = object()

    def pop(self, key, default=_nopop):
        value = self._data.pop(key, default)
        if value is NotifyDict._nopop:
            if default is NotifyDict._nopop:
                raise KeyError(key)
            else:
                return default
        self._removed(key, value)
        return value

    def popitem(self):
        item = self._data.popitem()
        self._removed(*item)
        return item

    ### Idempotent Methods

    def __cmp__(self, dict):
        if isinstance(dict, NotifyDict):
            return cmp(self._data, dict._data)
        else:
            return cmp(self._data, dict)

    def __len__(self):
        return len(self._data)

    def __getitem__(self, key):
        return self._data[key]

    def copy(self):
        return self._data.copy()

    def keys(self):
        return self._data.keys()

    def items(self):
        return self._data.items()

    def iteritems(self):
        return self._data.iteritems()

    def iterkeys(self):
        return self._data.iterkeys()

    def itervalues(self):
        return self._data.itervalues()

    def values(self):
        return self._data.values()

    def has_key(self, key):
        return self._data.has_key(key)

    def get(self, key, failobj=None):
        if not self.has_key(key):
            return failobj
        return self[key]

    def __contains__(self, key):
        return key in self._data


def identity(x):
    return x

class FilteringDict(object):
    __slots__ = ['_data',
                 '_loadKey',
                 '_storeKey',
                 '_loadValue',
                 '_storeValue']
    def __init__(self, data,
                 loadKey=identity, storeKey=identity,
                 loadValue=identity, storeValue=identity):
        """
        Filter a dictionary.
        """
        self._data = data
        self._loadKey = loadKey
        self._storeKey = storeKey
        self._loadValue = loadValue
        self._storeValue = storeValue

    def __repr__(self):
        return 'FilteringDict('+repr(self.copy())+')'

    # Mutating methods.

    def __setitem__(self, key, item):
        self._data[self._storeKey(key)] = self._storeValue(item)

    def update(self, dict):
        for k, v in dict.iteritems():
            self[k] = v

    def __delitem__(self, key):
        del self._data[self._storeKey(key)]

    def clear(self):
        self._data.clear()

    def setdefault(self, key, failobj=None):
        skey = self._storeKey(key)
        if not self._data.has_key(skey):
            self._data[skey] = failobj
        return self._loadValue(self._data[skey])

    def pop(self, key, default=NotifyDict._nopop):
         value = self._data.pop(self._storeKey(key), NotifyDict._nopop)
         if value is NotifyDict._nopop:
             if default is NotifyDict._nopop:
                 raise KeyError(key)
             else:
                 return default
         return self._loadValue(value)

    def popitem(self):
        item = self._data.popitem()
        return self._loadItem(item)

    ### Idempotent Methods

    def __cmp__(self, dict):
        if isinstance(dict, NotifyDict):
            return cmp(self._data, dict._data)
        else:
            return cmp(self.copy(), dict)

    def __len__(self):
        return len(self._data)

    def __getitem__(self, key):
        return self._loadValue(self._data[self._storeKey(key)])

    def _loadItem(self, (key, value)):
        return self._loadKey(key), self._loadValue(value)

    def copy(self):
        return dict(map(self._loadItem, self._data.iteritems()))

    def keys(self):
        return map(self._loadKey, self._data.iterkeys())

    def items(self):
        return map(self._loadItem, self._data.iteritems())

    def values(self):
        return map(self._loadValue, self._data.values())

    def iteritems(self):
        return iter(self.items())

    def iterkeys(self):
        return iter(self.keys())

    def itervalues(self):
        return iter(self.values())

    def has_key(self, key):
        return self._data.has_key(self._storeKey(key))

    def get(self, key, failobj=None):
        if not self.has_key(key):
            return failobj
        return self[key]

    def __contains__(self, key):
        return self._storeKey(key) in self._data


class TransformationList(object):
    """Something like a list, but applies some function before putting things
    in, and applies another function when taking them out.

    Hopefully, these two functions are mutually inverse.

    XXX this code was written and then turned out to be unneeded. Currently
    unused, and not thoroughly tested.
    """

    inFunction = lambda x: x
    outFunction = lambda x: x

    def __init__(self, initializer=[]):
        self.things = [self.inFunction(thing) for thing in initializer]

    def __delitem__(self, key):
        del self.things[key]

    def __delslice__(self, *args):
        self.things.__delitem__(*args)

    def __eq__(self, other):
        if not isinstance(other, TransformationList):
            return False
        return self.things == other.things

    def __ne__(self, other):
        return not self.__eq__(other)

    def __getitem__(self, key):
        return self.outFunction(self.things[key])

    def __iter__(self):
        for itemRef in self.things:
            yield self.outFunction(itemRef)

    def __repr__(self):
        return '%s.%s(%r)' % (self.__module__, self.__class__.__name__, getattr(self, 'things', None))

    def __setitem__(self, key, value):
        self.things[key] = self.inFunction(value)

    def append(self, thing):
        self.things.append(self.inFunction(thing))

    def count(self, thing):
        return self.things.count(self.inFunction(thing))

    def extend(self, items):
        self.things.extend(itertools.imap(self.inFunction, items))

    def index(self, thing):
        return self.things.index(self.inFunction(thing))

    def insert(self, index, thing):
        self.things.insert(index, self.inFunction(thing))

    def pop(self, *args):
        return self.outFunction(self.things.pop(*args))

    def remove(self, thing):
        self.things.remove(self.inFunction(thing))

    def reverse(self):
        self.things.reverse()
