# -*- test-case-name: atop.test.test_powerup -*-

# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

from twisted.python.components import Interface, Adapter, registerAdapter
from twisted.python import reflect

from atop.ifcompat import implements, backwardsCompatImplements

from atop.store import MetaItem, Item, Pool, NameIndex, SubStore, Store, deleteIfNotDeleted, _sim
from atop.tpython import amap

class IPowerup(Interface):

    property("name", doc="""
    The nickname for this powerup.  This may or may not be mutable: if it is a
    system-configured powerup like 'to-do', it needs its name to be consistent
    so that multiple things can find the root todo pool to add items to it.  If
    it is a user-created pool like 'Bob's Favorite Photos From his Vacation in
    Hali', then it is user-defined and therefore mutable.
    """)

    property("commandPrefixes",
             doc="""
    This property is a list of strings which are the commands that this powerup
    will recognize.  For example, for a To-Do system, this might be ['todo',
    't'], which would mean that from the web-UI, the `search box' would
    activate the To-Do system's commandLine method with either `/t' or `/todo'.
    """)

    property("dependencies",
             doc="""
    A list of other implementors of IPowerup which must be installed on the
    avatar before we can consider it ready for this powerup to be finished
    instantiating.  Elements of this list may also be strings representing
    fully qualified class names.
    """)

    property("storeDependencies",
             doc="""
    A list of other implementors of IPowerup which must be installed on the
    avatar's store before we can consider it ready for this powerup to be
    finished instantiating.  Elements of this list may also be strings 
    representing fully qualified class names.
    """)

    def commandLine(self, commander, context, text):
        """
        @param commander: The object that initiated this command-line action.

        @param context: The context in which this action was invoked.  In the
        case of a to-do being created from the web, the context might be the
        current resource.

        @param text: the text following the command string and separator.  So
        if the user types `/todo call dentist', this parameter is 'call
        dentist'.
        """

def _powerupByName(pool, name):
    for p in pool.queryIndex():
        if name == p.name:
            return p

class PowerupType(MetaItem):
    powerupNames = {}
    def __new__(cls, name, bases, attrs):
        # Normalize the dependencies list
        for a in ('dependencies', 'storeDependencies'):
            if a in attrs:
                deps = []
                for d in attrs[a]:
                    if not isinstance(d, str):
                        d = reflect.qual(d)
                    assert d not in deps
                    deps.append(d)
                attrs[a] = deps
        t = super(PowerupType, cls).__new__(cls, name, bases, attrs)
        if not hasattr(t, 'powerupName'):
            raise ValueError("%s has no powerupName" % (name,))
        elif t.powerupName is None:
            del t.powerupName
        else:
            qname = reflect.qual(t)
            if cls.powerupNames.setdefault(t.powerupName, qname) != qname:
                raise ValueError("Powerup named %s exists already." % (t.powerupName,))
        return t

class PowerupBase(Item):
    """Absolute base-est class for powerups.  Use this if you need to execute
    before the PowerStation is installed on the avatar.
    """
    __metaclass__ = PowerupType
    implements(IPowerup)
    dependencies = ()
    storeDependencies = ()
    powerupName = None

    def __init__(self, avatar):
        Item.__init__(self, avatar, self.powerupName)
        self.loadDependencies(avatar)
        if avatar.store:
            self.loadStoreDependencies(avatar.store)
        self.setUpPools(avatar)

    def _loadListedPowerups(self, st, listName):
        pm = IPowerStation(st)
        L = []
        reflect.accumulateClassList(self.__class__, listName, L)
        for dep in L:
            dep = reflect.namedClass(dep)
            pm.installPowerup(dep)

    def loadDependencies(self, avatar):
        return self._loadListedPowerups(avatar, 'dependencies')

    def loadStoreDependencies(self, store):
        return self._loadListedPowerups(store, 'storeDependencies')

    def _calculateDependants(self):
        dep = []
        ps = IPowerStation(self.store)
        for pupRef in ps.powerups.values():
            pup = pupRef.getItem()
            if reflect.qual(self.__class__) in pup.dependencies:
                dep.append(pup.getItemPath())
        return dep

    def deleteFromStore(self):
        return amap(deleteIfNotDeleted, self._calculateDependants()
            ).addCallback(*_sim(self._cbDeletedDependants)
            )

    def _cbDeletedDependants(self, result):
        self.store.touch()
        ps = IPowerStation(self.store)
        del ps.powerups[reflect.qual(self.__class__)]
        return Item.deleteFromStore(self)

    def setUpPools(self, avatar):
        """Initialize this powerup.

        Initialization usually involves the creation of pools, thus the name.
        However, this method is simply called at the end of the default
        implementation of __init__; there is nothing magic about it.
        """

    def getPlugIns(self, interface):
        """
        Return iterable of implementors of the given interface.

        See the documentation distributed with this package for more
        details.
        """
        return None

backwardsCompatImplements(PowerupBase)

class IPowerStation(Interface):
    """Tag interface for PowerStation, don't implement me.
    """

class PowerStation(Adapter):
    """Adapter around a store or substore which manages the creation and
    dependency of other powerups.
    """
    def __init__(self, original):
        Adapter.__init__(self, original)
        self.powerups = {}

    def installPowerup(self, powerupClass):
        assert issubclass(powerupClass, PowerupBase), ("Can only install Powerups on PowerStations, not", powerupClass)
        self.original.touch()
        cn = reflect.qual(powerupClass)
        if self.powerups.has_key(cn):
            return self.powerups[cn].getItem()
        else:
            pup = powerupClass(self.original)
            self.powerups[cn] = pup.referenceTo()
            pups = self.getPowerupsPool()
            if pup not in pups:
                pups.addItem(pup)
            return pup

    def getPowerupsPool(self):
        return self._powerupsPool.getItem()

    def getPlugIns(self, interface):
        """
        Return iterator of implementors of the given interface, querying
        each powerup in turn.

        See the documentation distributed with this package for more
        details.
        """
        for powerup in self.getPowerupsPool().queryIndex():
            subplg = powerup.getPlugIns(interface)
            if subplg is not None:
                for plugin in subplg:
                    yield plugin

registerAdapter(PowerStation, SubStore, IPowerStation)

class MetaPowerup(PowerupBase):
    invisible = True
    powerupName = 'MetaPowerup'
    def setUpPools(self, avatar):
        pm = IPowerStation(avatar)
        if isinstance(avatar, Store):
            # short-circuit for now, while the implementation of PowerStation is
            # divergent...
            return
        p = Pool(avatar, "powerups")
        p.invisible = True
        p.addIndex(NameIndex)
        p.suggestIndex(NameIndex, False)
        pm._powerupsPool = p.referenceTo()
        avatar.rootPoolRef.getItem().addItem(p)
        self.invisible = True

class Powerup(PowerupBase):
    """
    Base class for normal powerups.
    """
    dependencies = [MetaPowerup]
    powerupName = None

class Storeup(Powerup):
    """I am a store powerup.  I can be set as a component on the root store.
    """
    powerupName = None
    implementsForStore = None

    def setindex_powerup(self, newpowerup):
        self.implementsForStore = newpowerup
        return newpowerup

    def index_powerup(self, pool):
        return self.implementsForStore
