# -*- test-case-name: atop.test.test_sched -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import time, warnings

from twisted.application import service
from twisted.python.components import Interface
from twisted.python import log

from atop.ifcompat import *

from atop.powerup import Powerup, Storeup, IPowerStation
from atop.store import INDEX_AND_ITEM, Pool, DateIndex, ItemMixIn, Item, referenced, indexed, DeletedItem
from atop import tpython

class IScheduler(Interface):
    """Marker interface for Scheduler.
    """
    def callItemLater(self, item):
        pass
    def clock(self):
        pass

class IScheduledEvent(Interface):
    """Interface which scheduled events must implement.

    Subclass Schedulable and don't worry about it, okay?  Also, subclass item.

    I know, this isn't really a component.  Live with it.
    """

class ScheduledIndex(DateIndex):
    methodPrefix = 'scheduled'
class ScheduledErrorsIndex(DateIndex):
    methodPrefix = 'scheduled_errors'


class Schedulable(ItemMixIn):
    """I am a mix-in for items which wish to schedule timed tasks.
    """

    implements(IScheduledEvent)

    status = ''

    scheduled = indexed(ScheduledIndex)
    scheduledErrors = indexed(ScheduledErrorsIndex)

    def schedule(self, newTime):
        """Schedule my run() method to be executed at the given time.

        @type newTime: atop.tpython.Time
        @param newTime: when to call self.run()

        The old docstring said this, and it's wrong: "Schedule my run() method
        to be executed at the given local time.  (Values returned from
        time.time() are local times, so use that and adjust.)"

        In fact, time.time() is not local time, it is a POSIX timestamp. POSIX
        timestamps count the number of non-leap seconds since midnight, january
        1 1970. The number is the same in any timezone. This stupidness is why
        you should use tpython.Time.
        """
        if isinstance(newTime, tpython.Time):
            newTime = newTime.asPOSIXTimestamp()
        else:
            warnings.warn('using floats for Schedulable.schedule is deprecated. Use an atop.tpython.Time instance instead.', DeprecationWarning)

        if self.scheduled != newTime:
            # print 'Scheduling', self, 'for', time.ctime(newTime)
            self.scheduled = newTime
            assert self.scheduled == newTime
            IScheduler(self.store).callItemLater(self)

    def delay(self, secondsLater):
        """Delay the scheduled call.
        @param secondsLater: how many seconds from its current firing time to delay
        """
        self.schedule(self.getIndexKey(None, "scheduled") + secondsLater)

    def reset(self, secondsFromNow):
        """Reset the scheduled call's timer.
        @param secondsFromNow: how many seconds from now it should fire,
            equivalent to C{self.cancel()} and then doing another
            C{reactor.callLater(secondsLater, ...)}
        """
        self.schedule(IScheduler(self.store).clock() + secondsFromNow)

    def unschedule(self):
        self.setIndexKey(ScheduledIndex, None)
        p = IScheduler(self.store).pool
        if self in p:
            p.removeItem(self)

    def run(self):
        """Implement
        """
        raise NotImplementedError("")

backwardsCompatImplements(Schedulable)

class SchedulerBase(Schedulable):
    implements(IScheduler)

    clock = staticmethod(time.time)

    MAXIMUM_LOCKS = 50

    pool = referenced

    def setUpPools(self, store):
        p = Pool(store, "Scheduled tasks")
        p.invisible = True
        p.addIndex(ScheduledIndex)
        p.addIndex(ScheduledErrorsIndex)
        self.pool = p
        getattr(store, 'rootRef', getattr(store, 'rootPoolRef', None)).getItem().addItem(p)

    def getEvents(self):
        return self.pool.queryIndex(ScheduledIndex, result=INDEX_AND_ITEM)

    def callItemLater(self, item):
        """Call an item later, according to its date index.

        You probably want to use Schedulable.schedule(newTime) rather than
        calling this directly.
        """
        p = self.pool
        if item not in p:
            # These are almost always rescheduled items, not new
            # items, so the check saves us a function call, not
            # to mention cleans up our log output.
            p.addItem(item)
        for localtime, item in self.getEvents():
            self.schedule(localtime)
            return

    def nextEvent(self):
        try:
            return self.getEvents().next()
        except StopIteration:
            return None

    def _oneEvent(self, now, alreadyRan):
        next = self.nextEvent()
        if next is not None and next[0] < now:
            localtime, item = next
            scheduled = item.getIndexKey(None, ScheduledIndex)
            if int(localtime) != int(scheduled):
                log.msg("Non-fatal (maybe) database inconsistency in the scheduled items pool")
                log.msg("The item in question is %s" % (item,))
                log.msg("The pool believes it should run at %f (%s)" % (localtime, time.ctime(localtime)))
                log.msg("The item believes it should run at %f (%s)" % (scheduled, time.ctime(scheduled)))
                log.msg("I am erasing the pool's belief.")
                item.removeFromIndex(ScheduledIndex, localtime)
                return
            elif item.storeID in alreadyRan:
                # We're done for this tick
                return True
            alreadyRan[item.storeID] = None
            item.setIndexKey(ScheduledIndex, None)
            try:
                self.store.transact(IScheduledEvent(item).run)
            except:
                # TODO: later this should also actually catch the Failure
                # here and pickle it somewhere.
                log.err()
                item.setIndexKey(ScheduledIndex, None)
                item.setIndexKey(ScheduledErrorsIndex, localtime)
            else:
                if not isinstance(item, DeletedItem):
                    if item.getIndexKey(None, ScheduledIndex) is None:
                        if item in self.pool:
                            self.pool.removeItem(item)
        else:
            return True

    def tick(self):
        now = self.clock()
        alreadyRan = {}
        self.store.batch(self._oneEvent, now, alreadyRan)
        next = self.nextEvent()
        if next is not None:
            # print 'Next event is', next[1], 'and scheduled at', time.ctime(next[0])
            return next[0]
        # print 'No more events, returning None'
        return None

backwardsCompatImplements(SchedulerBase)

class SubTicker(Item, Schedulable):
    substore = referenced

    def __init__(self, substore):
        Item.__init__(self, substore.store)
        self.substore = substore

    def __str__(self):
        if self.substore is not None:
            return '<SubTicker (%s)>' % (self.substore.name,)
        return '<SubTicker (uninitialzed)>'

    def run(self):
        when = IScheduler(self.substore).tick()
        if when is None:
            self.unschedule()
        else:
            self.schedule(when)

class SystemScheduler(SchedulerBase, Storeup):
    """Top-level scheduling object stored in a database.

    I inherit scheduling functionality from SchedulerBase, which I
    share in common with SubScheduler.  All SubSchedulers are
    registered as timed events with me, so timed events are managed in
    a hierarchical fashion: startService and stopService start and
    shut down a reactor.callLater which I use to schedule myself, and
    when my timed event fires, I propogate to the next SubScheduler
    which has work to do at the time.  In turn, each SubScheduler
    keeps a list of Schedulables for its SubStore.

    """
    implements(service.IService)

    powerupName = "System Scheduler"
    persistenceForgets = ['nextCall']
    dependencies = ['atop.service.Serviceup']

    running = False

    def setUpPools(self, store):
        SchedulerBase.setUpPools(self, store)
        store.setComponent(IScheduler, self)
        service.IServiceCollection(store).addService(self)

    def startService(self):
        self.touch()
        self.running = True
        self.callLaterEr()

    def stopService(self):
        self.touch()
        self.running = False
        if self.nextCall is not None:
            self.nextCall.cancel()
            self.nextCall = None

    def setsStoreInterfaces(self):
        return [IScheduler]

    nextCall = None

    def callLater(self, secs, f, *a, **b):
        from twisted.internet import reactor
        return reactor.callLater(secs, f, *a, **b)

    def schedule(self, newTime):
        if isinstance(newTime, tpython.Time):
            newTime = newTime.asPOSIXTimestamp()
        else:
            warnings.warn('using floats for Schedulable.schedule is deprecated. Use an atop.tpython.Time instance instead.', DeprecationWarning)

        if self.nextCall is not None:
            self.nextCall.cancel()
            self.nextCall = None
        if self.running and newTime is not None:
            delay = max(newTime - self.clock(), 0)
            self.nextCall = self.callLater(delay, self.callLaterEr)

    def callLaterEr(self):
        self.nextCall = None
        nexttime = self.store.transact(self.tick)
        self.schedule(nexttime)

backwardsCompatImplements(SystemScheduler)

class SubScheduler(SchedulerBase, Powerup):
    powerupName = 'Scheduled Tasks'

    storeDependencies = [SystemScheduler]

    def setUpPools(self, substore):
        SchedulerBase.setUpPools(self, substore)
        substore.setComponent(IScheduledEvent, SubTicker(substore))
        substore.setComponent(IScheduler, self)

    def deleteFromStore(self):
        self.store.unsetComponent(IScheduler)
        # XXX Ugg why are these things so mixed together?
        se = IScheduledEvent(self.store)
        self.store.unsetComponent(IScheduledEvent)
        se.deleteFromStore()
        return Powerup.deleteFromStore(self)

    def schedule(self, newTime):
        IScheduledEvent(self.store).schedule(newTime)

class CallingItem(Item, Schedulable):

    """Helper for transaction-safe reactor.callLater()

    Instead of reactor.callLater(n, f, *a, **kw) simply
    use CallingItem(store, n, f, *a, **kw) in a transaction.
    Note that f, a, and kw must all be serializable.
    """

    def __init__(self, store, n, f, *a, **kw):
        Item.__init__(self, store)
        self.f = f
        self.a = a
        self.kw = kw
        self.reset(n)

    def run(self):
        self.unschedule()
        try:
            self.f(*self.a, **self.kw)
        finally:
            self.deleteFromStore()
