# -*- test-case-name: atop.test.test_store.Storables -*-

from twisted.python import context, reflect
from atop.tpython import NamedProperty
from atop.store import SEP, MULTIPASS

def noop():
    pass

class StoredAttribute(NamedProperty):
    def __get__(self, oself, type=None):
        name = self.getName(oself)
        if hasattr(oself, "__temporary__"):
            return oself.__temporary__[name]
        else:
            return self.getter(oself, self.prefix + name, oself.getStoredData(name))

    def __set__(self, oself, value, type=None):
        name = self.getName(oself)
        if hasattr(oself, "__temporary__"):
            # Oops, we're not in the database yet!
            oself.__temporary__[name] = value
        else:
            oself.setStoredData(name, self.setter(oself, name, value))

    def getter(self, oself, name, value):
        """Convert a string value from the database back into a python object.  Inverse
        of 'setter'.
        """
        return value

    def setter(self, oself, name, value):
        """Convert a python object to a string suitable to be stored in the
        database. Inverse of 'getter'.
        """
        return value

class StoredUnicode(StoredAttribute):
    encoding = 'utf-8'
    def getter(self, oself, name, value):
        return value.decode(self.encoding)
    def setter(self, oself, name, value):
        return value.encode(self.encoding)

class StoredInteger(StoredAttribute):
    def getter(self, oself, name, value):
        return str(value)
    def setter(self, oself, name, value):
        return repr(value)

class StoredReference(StoredAttribute):
    def getter(self, oself, name, value):
        return oself.store.getItemByID(int(value))
    def setter(self, oself, name, value):
        return str(value.storeID)

class UnstoredAttribute(NamedProperty):
    def getter(self, oself, name, value): return value
    def setter(self, oself, name, value): return value

class MetaStorable(type):
    def __new__(cls, name, bases, attrs):
        assert not '__slots__' in attrs
        S = attrs['__slots__'] = []
        oldS = []
        for base in bases:
            oldS.extend(getattr(base,'__slots__',[]))

        for special in ['__temporary__', '__weakref__']:
            if special not in oldS:
                S.append(special)
        for aname, attr in attrs.items():
            if isinstance(attr, UnstoredAttribute):
                sname = attr.prefix + aname
                if sname not in oldS:
                    S.append(sname)
        return super(MetaStorable, cls).__new__(cls, name, bases, attrs)

class Storable(object):
    __metaclass__ = MetaStorable

    store = UnstoredAttribute()
    storeID = UnstoredAttribute()
    _attrStoragePrefix = UnstoredAttribute()

    def __new__(cls, *a, **k):
        i = super(Storable, cls).__new__(cls, *a,**k)
        i.__temporary__ = {}
        return i

    def __finalizer__(self):
        return noop

    def _getItemPath(self):
        return self.store.getItemPath() + (self.storeID,)

    def getStoredData(self, name):
        return self.store.db.get(self._attrStoragePrefix + name,
                                 txn=context.get('txn'))

    def setStoredData(self, name, data):
        return self.store.db.put(self._attrStoragePrefix + name, data,
                                 txn=context.get('txn'))

    def _attrsFromStore(self, store, storeID):
        self.store = store
        self.storeID = storeID
        path = self._getItemPath()
        self._attrStoragePrefix = self.store._formatIDPath(path) + SEP

    def wasFromStore(self, store, storeID):
        self._attrsFromStore(store, storeID)
        del self.__temporary__ # <-- craziness like that can happen here

    def addToStore(self, store, storeID=None):
        if storeID is None:
            storeID = store.nextID()
        self._attrsFromStore(store, storeID)
        self.store.cacheItem(self, self._getItemPath())
        self.store.db.put(self._attrStoragePrefix[:-1],
                          MULTIPASS + reflect.qual(self.__class__),
                          txn=context.get('txn'))
        for k, v in self.__temporary__.iteritems():
            setattr(self, k, v)
        del self.__temporary__

    def fromStorage(Class, store, storeID):
        i = Class.__new__(Class)
        i.wasFromStore(store, storeID)
        return i
    fromStorage = classmethod(fromStorage)

    def addedToPool(self, pool):
        pass


class OverlayStringMap:
    """Simple mapping of string keys to string values.

    Keys and values are stored directly in the database, allowing for a
    large number of them to be present without unduly increasing the size of
    this item in a serialized object graph.
    """
    def __init__(self, store, prefix):
        self.db = store.db
        self.prefix = prefix

    def __getstate__(self):
        return self.prefix

    def __setstate__(self, state):
        self.prefix = state
        self.db = context.get('store').db

    def __getitem__(self, key):
        """OverlayStringMap[key] -> value

        If the key is not present, None is returned.
        """
        if not isinstance(key, str):
            raise TypeError("OverlayStringMap.__getitem__ requires a string, got %s" % (type(key).__name__,))
        return self.db.get(self.prefix + key,
                           txn=context.get('txn'))

    def __setitem__(self, key, value):
        """OverlayStringMap[key] = value
        """
        for o in (key, value):
            if not isinstance(o, str):
                raise TypeError("OverlayStringMap.__setitem__ requires strings, got %s" % (type(o).__name__,))
        self.db.put(self.prefix+key, value,
                    txn=context.get('txn'))

    def __delitem__(self, key):
        if not isinstance(key, str):
            raise TypeError("OverlayStringMap.__delitem__ requires a string, got %s" % (type(key).__name__,))
        self.db.delete(self.prefix + key,
                       txn=context.get('txn'))

    def __contains__(self, key):
        return self[key] is not None

    def has_key(self, key):
        return key in self

    def iterkeys(self):
        return self.queryitems(withValues=False)

    def itervalues(self):
        for k, v in self.queryitems(withValues=True):
            yield v

    def iteritems(self):
        return self.queryitems(withValues=True)

    def keys(self):
        return list(self.iterkeys())

    def values(self):
        return list(self.itervalues())

    def items(self):
        return list(self.iteritems())

    def queryitems(self, prefix='', withValues=True):
        return _OverlayStringMapIterator(self, prefix, withValues)

class _OverlayStringMapIterator:
    startKey = ''
    cursorMethod = None
    cursorFactory = None

    def __init__(self, map, prefix, withValues=True):
        self.withValues = withValues
        self.prefix = prefix
        self.allKeysBeginWith = map.prefix + prefix
        self.lastItem = (self.allKeysBeginWith, '')

    def _nextKey(self):
        firstTime = self.cursorMethod is None
        currentFactory = context.get("CursorFactory")
        if self.cursorFactory is not currentFactory:
            self.cursorFactory = currentFactory
            cursor = currentFactory.cursor()
            self.cursorMethod = cursor.next
            lastKey, lastValue = self.lastItem
            val = newKey, newValue = cursor.set_range(lastKey)
            if newKey == lastKey and not firstTime:
                val = self.cursorMethod()
        else:
            val = self.cursorMethod()
        self.lastItem = val
        return val

    def next(self):
        k, v = self._nextKey()
        if not k.startswith(self.allKeysBeginWith):
            raise StopIteration()
        key = k[len(self.allKeysBeginWith):]
        if self.withValues:
            return key, v
        else:
            return key

    def __iter__(self):
        return self


class OverlayTupleMap(OverlayStringMap):

    def __contains__(self, key):
        return OverlayStringMap.__contains__(self, self._formatTupleKey(key))

    def __getitem__(self, key):
        return OverlayStringMap.__getitem__(self, self._formatTupleKey(key))

    def __setitem__(self, key, value):
        return OverlayStringMap.__setitem__(self, self._formatTupleKey(key), value)

    def __delitem__(self, key):
        return OverlayStringMap.__delitem__(self, self._formatTupleKey(key))

    def _formatTupleKey(self, tupleKey):
        if isinstance(tupleKey, basestring):
            # Allow the syntax foo['a'], as it is logically similar to foo['a', 'b']
            tupleKey = (tupleKey,)
        assert not [x for x in tupleKey if SEP in x], \
            "Don't put SEP in tupleKey (or contribute a patch)"
        stringKey = SEP.join(tupleKey)
        return stringKey

    def queryitems(self, prefix=(), withValues=True):
        for x in _OverlayStringMapIterator(self, self._formatTupleKey(prefix)+SEP, withValues):
            if withValues:
                yield tuple(x[0].split(SEP)), x[1]
            else:
                yield tuple(x.split(SEP))

class StoredStringMapping(NamedProperty):
    wrapper = OverlayStringMap
    def __get__(self, oself, type=None):
        name = self.getName(oself)
        return self.wrapper(oself.store, oself._attrStoragePrefix + name + SEP)

    def __set__(self, oself, value, type=None):
        raise AttributeError("You can't change the value of a stored mapping.")

class StoredTupleMapping(StoredStringMapping):
    wrapper = OverlayTupleMap
