# -*- test-case-name: atop.test.test_store -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

"""
Storage classes for ATOP.
"""

__version__ = '$Revision: 1.135 $'

import warnings, time, os, sys, new, shutil, atexit, inspect, itertools, operator
from os.path import join as opj

from bsddb import db as libdb

from twisted.python import log, context
from twisted.python.components import Interface, Adapter, Componentized, implements as doesImplement, CannotAdapt

from twisted.python.util import sibpath
from twisted.application import service
from twisted.internet import defer, task

from twisted.python import reflect

from weakref import ref

from twisted.persisted.styles import Ephemeral

from atop.ifcompat import implements, backwardsCompatImplements, Attribute, getAdapterFactory

from atop.tpython import mkgmtime, FilesystemLock, amap, Time
from atop.fwd import ItemForwardingComponent, makeForwarder
from atop import versioning
from atop import iatop
from atop import notify

# the encoding used to serialize unicode objects
STRING_ENCODING = 'utf-8'

_indexUpgradeTable = {
    'uid': 'atop.store.UIDIndex',
    'received': 'atop.store.DateIndex',
    'scheduled': 'atop.sched.ScheduledIndex',
    'scheduled_errors': 'atop.sched.ScheduledErrorsIndex',
    'lastName': 'quotient.smarts.person.LastNameIndex',
    'firstName': 'quotient.smarts.person.FirstNameIndex',
    }

def _deprecatedIndexLookup(name, stacklevel=3):
    if isinstance(name, str):
        warnings.warn("Use indexes for indexes, not names.",
                      DeprecationWarning, stacklevel=stacklevel)
        return reflect.namedClass(_indexUpgradeTable[name])
    elif isinstance(name, tuple):
        return MultiIndex(*name)
    else:
        return name

_indexToStringTable = { }
_stringToIndexTable = { }

PICKLE = '\x80'
MULTIPASS = '\x81'

def noop():
    pass

class IPoolIndex(Interface):

    methodPrefix = Attribute("""
    A nickname for use in deprecated indexing methods.
    """)

    def keyToSortableString(self, key):
        """Convert a given 'key' object (such as a date) to a fixed-length
        sortable string: e.g. the date represented by the float
        1066147720.700978 might be converted to '2003/10/14/12/08/40'
        """

    def sortableStringToKey(self, sortableString, pool):
        """Convert a value produced by keyToSortableString back into its
        original format.
        """

def stringseqpack(esses):
    Ls = len(esses)
    assert Ls < 256, "What the HELL are you doing???"
    return ''.join(esses)+''.join([chr(len(x)) for x in esses])+chr(Ls)

def stringsequnpack(S):
    howmany = ord(S[-1])
    r = range(howmany)
    offt = 0
    for R in r:
        Loff = -1-howmany+R
        clen = ord(S[Loff])
        r[R] = S[offt:offt+clen]
        offt += clen
    return r

class MultiIndex(object):
    """I am an index that maps over several underlying indexes.
    """

    useValueOf = property(lambda x: x)
    __name__ = property(lambda x: '<'+('&'.join(map(str,x.indexes)))+'>')

    implements(IPoolIndex)
    def __init__(self, *indexes):
        """
        """
        self.indexes = indexes

    def keyToSortableString(self, key):
        """
        """
        assert len(key) == len(self.indexes), "Multi-index requires key argument to match number of indexes."
        return stringseqpack([i.keyToSortableString(k) for i, k in zip(self.indexes, key)])

    def sortableStringToKey(self, sortableString, pool):
        """
        """
        return tuple(
            [i.sortableStringToKey(s, pool)
             for i, s
             in zip(self.indexes, stringsequnpack(sortableString))])

    def keyFromItem(self, pool, item, default=None):
        keys = []
        for idx in self.indexes:
            k = item.getIndexKey(pool, idx)
            if k is None:
                return None
            keys.append(k)
        return keys

    def __hash__(self):
        return hash(self.indexes)

    def __eq__(self, other):
        if not isinstance(other, MultiIndex):
            return False
        return self.indexes == other.indexes

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        return '<MultiIndex (%s)>' % (
            ' '.join([idx.__name__ for idx in self.indexes]),)

    __repr__ = __str__

backwardsCompatImplements(MultiIndex)

class MetaIndex(type):
    implements(IPoolIndex)

    def _getself(self):
        warnings.warn("Don't use the .index attribute any more.",
                      DeprecationWarning, stacklevel=2)
        return self

    index = property(_getself)

    def _useval(self):
        if self.__dict__.has_key('useValueOf'):
            return self.__dict__['useValueOf']
        else:
            return self

    useValueOf = property(_useval)

    def __setstate__(self, state):
        """Ignore __setstate__: this is an ugly and unfortunate result of some
        old persistence decisions.
        """

    def __new__(klass, name, bases, props):
        # creating
        for descr, method in ((staticmethod, 'sortableStringToKey'),
                              (staticmethod, 'keyToSortableString'),
                              (staticmethod, 'storeKey'),
                              (staticmethod, 'loadKey')):
            if props.has_key(method):
                props[method] = descr(props[method])
        nt = type.__new__(klass, name, bases, props)
        Q = reflect.qual(nt)
        if 'methodPrefix' in props:
            assert nt.methodPrefix not in _stringToIndexTable, ("_stringToIndexTable already had", nt.methodPrefix)
            _stringToIndexTable[nt.methodPrefix] = nt
            if _indexUpgradeTable.has_key(nt.methodPrefix):
                a = _indexUpgradeTable[nt.methodPrefix]
                b = Q
                # assert a == b, "%s != %s [%s]" % (a, b, nt.methodPrefix)
            else:
                _indexUpgradeTable[nt.methodPrefix] = Q
        _stringToIndexTable[Q] = nt
        _indexToStringTable[nt] = Q
        return nt

    def __call__(self, name=None, maybestore=None):
        warnings.warn("Just use the index, don't instantiate it", DeprecationWarning,
                      stacklevel=2)
        if name is not None:
            subcl = _deprecatedIndexLookup(name)
            if not issubclass(subcl, self):
                warnings.warn("%s not subclass of %s" % (subcl, self), DeprecationWarning)
            return subcl
        return self

backwardsCompatImplements(MetaIndex)

# helpful descriptors
class indexed(object):
    def __init__(self, index, callOnChange=None):
        """
        @param index: the index class this attribute represents
        @param callOnChange: a callable taking two arguments, the instance
        being changed (self, probably) and the new value. This will be called
        when the value is changed.

        | class FuIndex(StringIndex): pass
        |
        | class Fu(Item):
        |     def fuChanged(self, newFu):
        |         print 'fu has changed to %r' % (newFu,)
        |     fu = indexed(FuIndex, fuChanged)
        """
        self.index = index
        self.callOnChange = callOnChange

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        return self.index.loadKey(obj.getIndexKey(None, self.index))

    def __set__(self, obj, value):
        obj.setIndexKey(self.index, self.index.storeKey(value))
        if self.callOnChange:
            self.callOnChange(obj, value)


class indexedList(indexed):
    def __get__(self, item, type=None):
        return item._multiValueProxy(self.index)

    def __set__(self, item, value):
        self.__get__(item)[:] = value

class indexedPoolDict(indexed):
    def __get__(self, item, type=None):
        return item._differentValueProxy(self.index)

    def __set__(self, obj, value):
        pdolip = self.__get__(obj)
        pdolip.clear()
        pdolip.update(value)

class indexedPoolDictOfLists(indexed):

    def __get__(self, item, type=None):
        return item._differentMultiValueProxy(self.index)

    def __set__(self, obj, value):
        pdolip = self.__get__(obj)
        pdolip.clear()
        pdolip.update(value)

class PoolIndex:
    __metaclass__ = MetaIndex
    def storeKey(obj):
        return obj

    def loadKey(obj):
        return obj

    def toString(K):
        return _indexToStringTable[K]
    toString = classmethod(toString)

    _noDefaultIndex = object()
    def fromString(K, S, default=_noDefaultIndex):
        try:
            return _stringToIndexTable[S]
        except KeyError:
            if default is K._noDefaultIndex:
                raise
            return default
    fromString = classmethod(fromString)

    def keyFromItem(K, pool, item, default=None):
        return default
    keyFromItem = classmethod(keyFromItem)


class DateIndex(PoolIndex):
    methodPrefix = 'received'

    def sortableStringToKey(sstr, pool):
        return mkgmtime(tuple(map(int, sstr.split(".")))+(0,0,0))

    def keyToSortableString(key):
        if isinstance(key, Time):
            gmt = key.asStructTime()
        elif isinstance(key, (list, tuple, time.struct_time)):
            warnings.warn("DateIndex keys should be atop.tpython.Time instances, not 9-tuples", stacklevel=5)
            gmt = key
        else:
            warnings.warn("DateIndex keys should be atop.tpython.Time instances, not something else like floats or integers", stacklevel=5)
            gmt = time.gmtime(key)
        return "%04d.%02d.%02d.%02d.%02d.%02d" % gmt[:6]

class IntIndex(PoolIndex):
    methodPrefix = 'int'

    def keyToSortableString(key):
        assert isinstance(key, (int, long)), "%r is not an int" % (key,)
        return '%0.10d' % key

    def sortableStringToKey(sstr, pool):
        return int(sstr)


class HierarchyIndex(PoolIndex):
    def keyToSortableString(key):
        "tuple of positive ints -> \x02-joined list of 10-digit strings"
        return '\x02'.join(['%010d'%x for x in key])

    def sortableStringToKey(sstr, pool):
        "\x02-separated list of 10-digit strings to tuple of ints"
        return tuple(map(int, sstr.split('\x02')))

class HierarchyBrowser:
    def __init__(self, pool):
        self.pool = pool

    def queryAtDepth(self, pfx, descent=0):
        # print '-------------------------------'

        def _next(x):
            return x + (int(1e10) - 1,) * descent + (0,)

        query = self.pool.queryIndex(HierarchyIndex,
                                     result=INDEX_AND_STORE_ID,
                                     startKey=pfx,
                                     endKey=_next(pfx))

        while True:
            try:
                index, storeID = query.next()
            except StopIteration:
                return
            else:
                if len(index) - len(pfx) > descent:
                    # print index, 'is too deep, skipping to', _next(index), ':', _next(pfx)
                    query = self.pool.queryIndex(HierarchyIndex,
                                                 result=INDEX_AND_STORE_ID,
                                                 startKey=_next(index),
                                                 endKey=_next(pfx))
                else:
                    # print 'Good value in', index
                    yield index, storeID

class UIDIndex(IntIndex):
    methodPrefix = 'uid'

class StringIndex(PoolIndex):
    """Index for character strings (unicode objects).

    Keys are encoded using the module level STRING_ENCODING.  Since this
    value is not stored in the database anywhere (at this time), changing
    it will lead to horribly broken indexes, so _don't change it_.  For backwards
    compatibility reasons, ascii-only str objects are allowed without a warning
    (although a warning may be added in the future), and non-ascii str objects
    are allowed, but with a warning, and decoded using the 'charmap' codec.
    """
    def keyToSortableString(key):
        if isinstance(key, str):
            try:
                key = key.decode('ascii')
            except UnicodeDecodeError:
                warnings.warn("Non-unicode non-ascii not supported by StringIndex", stacklevel=5)
                key = key.decode('charmap')
        return key.encode(STRING_ENCODING)

    def sortableStringToKey(sstr, pool):
        return unicode(sstr, STRING_ENCODING, 'replace')

class NameIndex(StringIndex):
    methodPrefix = 'name'

class CaseInsensitiveStringIndex(StringIndex):
    def keyToSortableString(key):
        return StringIndex.keyToSortableString(key.strip().lower())

class CaseInsensitiveNameIndex(CaseInsensitiveStringIndex):
    pass

CaseInsensitiveIndex = CaseInsensitiveNameIndex

class PowerupIndex(NameIndex):
    methodPrefix = 'powerup'

class ItemReferenceIndex(PoolIndex):
    '''Allow access to a "foreign key"
    '''
    def keyToSortableString(key):
        # key ought to be an ItemReference.
        return '%0.10d' % key.storeID

    def sortableStringToKey(sstr, pool):
        return ItemReference(pool.store, int(sstr))

    def loadKey(ref):
        if ref is not None:
            return ref.getItem()

    def storeKey(item):
        if item is not None:
            return item.referenceTo()

NAME = NameIndex

# This is a value that is greater than anything that might show up inside
# one part of a key string, but less than a special magic value that lets
# us know when a particular prefix has ended.  It cannot show up in any parts
# of any key names inserted into the database.
SEP = chr(0x00)
END = chr(0xff)

# We use this because index keys contain their storeIDs to allow multiple
# values per key.  When we do queries we need to make sure that they ignore the
# storeID.

MAGIC = chr(0x01)

def md(dirname):
    if os.path.isdir(dirname):
        return False
    os.makedirs(dirname)
    return True

MAX_LOCKS = 1000
MAX_LOCKERS = 1000
MAX_OBJECTS = 1000

LOCK_TIMEOUT = 1
TXN_TIMEOUT = 1

DB_LOCK_NAME = 'lock'

class DatabaseAlreadyOpen(RuntimeError):
    def __str__(self):
        return 'Database Already Open: %s' % RuntimeError.__str__(self)

# This flag is really new.  Not using it will slow things
# down a lot, but at least it will work.
try:
    libdb.DB_TXN_WRITE_NOSYNC
except AttributeError:
    log.msg("Warning: DB_TXN_WRITE_NOSYNC not available.")
    libdb.DB_TXN_WRITE_NOSYNC = 0

class _EnvAndDB:
    cleanedUp = False

    def __init__(self, env, db, lock = None):
        self.env = env
        self.db = db
        self.lock = lock
        atexit.register(self._cleanup)

    def __len__(self):
        if self.lock is not None:
            return 3
        return 2

    def __getitem__(self, index):
        if self.lock is not None:
            return (self.env, self.db, self.lock)[index]
        return (self.env, self.db)[index]

    def _cleanup(self):
        if self.cleanedUp:
            return
        log.msg("Warning: Cleaning up after dirty database/environment")
        self.cleanup()

    def cleanup(self):
        # XXX - In a transaction, this code path is a SEGV.
        assert context.get('txn') is None
        self.db.close()
        self.env.close()
        if self.lock is not None:
            self.lock.unlock()
        self.db = self.env = self.lock = None
        self.cleanedUp = True

class DatabaseEnvironment:
    # Name of the lock file to protect this environment with
    lockFileName = None

    maxLocks = None
    maxLockers = None
    maxLockObjects = None

    lockTimeout = None
    txnTimeout = None

    envJoinFlags = libdb.DB_JOINENV

    envOpenFlags = (libdb.DB_INIT_MPOOL   # Init shared memory subsystem
                    | libdb.DB_INIT_LOG   # Init logging subsystem
                    | libdb.DB_INIT_LOCK  # Initialize locking subsystem
                    | libdb.DB_INIT_TXN   # Make environment transactional
                    | libdb.DB_CREATE)    # Create files as necessary

    envFlags = (libdb.DB_TXN_WRITE_NOSYNC
                | libdb.DB_AUTO_COMMIT)

    normalRecoveryFlags = libdb.DB_RECOVER
    fatalRecoveryFlags = libdb.DB_RECOVER_FATAL

    def prepareEnvironment(self, env):
        env.set_lk_max_locks(self.maxLocks or MAX_LOCKS)
        env.set_lk_max_lockers(self.maxLockers or MAX_LOCKERS)
        env.set_lk_max_objects(self.maxLockObjects or MAX_OBJECTS)
        env.set_timeout(self.lockTimeout or LOCK_TIMEOUT, libdb.DB_SET_LOCK_TIMEOUT)
        env.set_timeout(self.txnTimeout or TXN_TIMEOUT, libdb.DB_SET_TXN_TIMEOUT)
        env.set_flags(self.envFlags, 1)

    def openEnvironment(self, envdir, lockClean):
        if lockClean:
            # This *could* be 0 (no flag), but it doesn't
            # hurt (except it's slow) to recover a consistent
            # database like this, so we'll play it safe for
            # the time being.
            extraFlags = self.normalRecoveryFlags
        else:
            log.msg("Lock acquired uncleanly.  Performing fatal recovery.")
            extraFlags = self.fatalRecoveryFlags

        env = libdb.DBEnv()
        try:
            self.prepareEnvironment(env)
            try:
                log.msg("Opening environment...")
                env.open(envdir, self.envOpenFlags | extraFlags)
            except libdb.DBRunRecoveryError:
                if lockClean:
                    log.msg("FATAL RECOVERY FAILED.  CALL SLEEPYCAT IMMEDIATELY.")
                else:
                    log.msg("Regular recovery failed.")
                raise
            log.msg("Environment opened.")
        except:
            env.close()
            raise
        return env

    def openDatabase(self, env, datafile):
        db = libdb.DB(env)
        try:
            txn = env.txn_begin()
            try:
                log.msg("Opening database...")
                db.open(datafile, "db", libdb.DB_BTREE, libdb.DB_CREATE, txn=txn)
            except:
                txn.abort()
                db.close()
                env.close()
                raise
            txn.commit()
            log.msg("Database opened.")
        except:
            db.close()
            raise
        return db

    def join(self, envdir, datafile='db'):
        envdir = os.path.abspath(envdir)
        env = libdb.DBEnv()
        try:
            env.open(envdir, self.envJoinFlags)
            db = self.openDatabase(env, datafile)
        except:
            env.close()
            raise
        return _EnvAndDB(env, db)

    def fixupOldEnvironment(self, datadir):
        """Backwards compatibility upgrader
        """
        envdir = opj(datadir, 'env')
        if os.path.isdir(envdir):
            for fname in os.listdir(envdir):
                os.rename(opj(envdir, fname), opj(datadir, fname))
            os.rmdir(envdir)
            log.msg("Re-arranged directory structure beneath " + datadir)

    def open(self, envdir, datafile='db'):
        """Open a BSDDB environment and a single database.
        """
        envdir = os.path.abspath(envdir)
        md(envdir)

        lockfile = opj(envdir, self.lockFileName or DB_LOCK_NAME)
        lockObj = FilesystemLock(lockfile)

        if not lockObj.lock():
            raise DatabaseAlreadyOpen(envdir, lockfile)
        try:
            self.fixupOldEnvironment(envdir)

            env = self.openEnvironment(envdir, lockObj.clean)
            try:
                db = self.openDatabase(env, datafile)
            except:
                env.close()
                raise
        except:
            if lockObj.clean:
                lockObj.unlock()
            raise
        return _EnvAndDB(env, db, lockObj)

_dbenv = DatabaseEnvironment()
envdblock = _dbenv.open
envdbjoin = _dbenv.join

def retrieveIDKey(db, keyName):
    return int(db.get(keyName, txn=context.get('txn')) or 0)

def updateIDKey(db, keyName):
    txn = context.get('txn')
    if txn is None:
        raise NotImplementedError("Updating ID outside of transaction")
    val = db.get(keyName, flags=libdb.DB_RMW, txn=txn)
    if val is None:
        val = 0
    ret = int(val)
    db.put(keyName, str(ret + 1), txn=txn)
    return ret

class LoginNotFound(KeyError):
    pass

class DuplicateLogin(KeyError):
    pass

class NoSuchIndex(KeyError):
    """Raised when a requested index is not found.
    """

class DuplicateIndex(Warning):
    """Raised when an index is added twice.
    """

class Inconsistency(Exception):
    """
    A slightly weaker warning than an assert.  Something bad happened, but it
    was recoverable, and it's the kind of thing we need to consistently check
    for at runtime (it can't be disabled because it's not really specified by
    any invariant, and could be exploited by attackers).
    """

class StoreClosed(Inconsistency):
    """Operation on a closed store.
    """

class AtomicFile(file):
    """I am a file which is moved from temporary to permanent storage when it
    is closed.

    After I'm closed, I will have a 'finalpath' property saying where I went.
    """
    def __init__(self, tempname, destdir):
        self._destdir = destdir
        file.__init__(self, tempname, 'wb')

    def close(self, destpath):
        file.close(self)
        fullpath = self._destdir + (os.sep + destpath)
        md(os.path.dirname(fullpath.path))
        self.finalpath = fullpath
        os.rename(self.name, fullpath.path)
        return fullpath

    def abort(self):
        os.unlink(self.name)

tempCounter = itertools.count()

def noLogout():
    pass

import traceback
def _fcremover(w, k, f):
    def remove(o):
        # Weakref callbacks cannot raise exceptions
        # or DOOM ensues
        try:
            try:
                f()
            except:
                log.msg("Exception in finalizer cannot be propegated")
                log.err()
            self = w()
            if self is not None:
                del self.data[k]
        except:
            try:
                traceback.print_exc(file=file('_fcremover.log', 'a'))
            except:
                try:
                    fh = file('_fcremover.log.log', 'a')
                    fh.write('Doom doom doom\n')
                    t, i, tb = sys.exc_info()
                    fh.write(str(i) + '\n')
                    fh.flush()
                    fh.close()
                except:
                    try:
                        fh = file('_fcremover.log.log.log', 'a')
                        fh.write('Tried really hard and failed to log an exception')
                        fh.flush()
                        fh.close()
                    except:
                        # Jesus christ.
                        pass
    return remove

class StrongCache(service.Service):
    """Holds strong references to objects, for a while.
    """

    interval = 10
    index = 0

    maxBufferSize = 1000
    defaultRingSize = 100

    def __init__(self, size=None):
        self.ringSize = size

    def startService(self):
        service.Service.startService(self)
        self.data = [{} for x in range(self.ringSize or self.defaultRingSize)]
        self.call = task.LoopingCall(self.tick)
        self.dcall = self.call.start(self.interval)
        self.dcall.addErrback(self._ebLoop)

    def stopService(self):
        service.Service.stopService(self)
        self.data = None
        if self.call is not None:
            self.call.stop()

    def cache(self, path, item):
        self.data[self.index][path] = item
        if len(self.data[self.index]) >= self.maxBufferSize:
            self.tick()

    def grow(self, amount=1):
        self.data[self.index + 1:self.index + 1] = [{} for x in range(amount)]

    def tick(self):
        self.index += 1
        self.index %= len(self.data)
        self.data[self.index].clear()

    def _ebLoop(self, failure):
        self.call = None
        self.dcall = None
        log.err(failure)

class FinalizingCache:
    """Possibly useful for infrastructure?  This would be a nice addition (or
    perhaps even replacement) for twisted.python.finalize.
    """
    def __init__(self):
        self.data = {}
        if not versioning.PROFILING:
            # see docstring for 'has'
            self.has = self.data.has_key

    def cache(self, key, value):
        fin = value.__finalizer__()
        self.data[key] = ref(value, _fcremover(ref(self), key, fin))
        return value

    def uncache(self, key):
        del self.data[key]

    def has(self, key):
        """Does the cache have this key?

        (This implementation is only used if the system is being profiled, due
        to bugs in Python's old profiler and its interaction with weakrefs.
        Set the module attribute PROFILING to True at startup for this.)
        """
        if key in self.data:
            o = self.data[key]()
            if o is None:
                del self.data[key]
                return False
            return True
        return False

    def get(self, key):
        o = self.data[key]()
        if o is None:
            raise Inconsistency("FinalizingCache has %r but its value is no more." % (key,))
        log.msg(interface=iatop.IStatEvent, name='database', stat_cache_hits=1, key=key)
        return o

class StoreRelativePath:
    """A path that keeps itself relative to the file directory of a store.

    @type path: C{str}
    @ivar path: The string to use when a string path is required.
    """
    def __init__(self, relative, store=None):
        self.__setstate__(relative, store)

    def __getstate__(self):
        return self.relative

    def __setstate__(self, relative, store=None):
        if store is None:
            store = context.get('store')
        if store is None:
            raise RuntimeError("No Store in context and no Store given to StoreRelativePath()")
        while store.store:
            store = store.store
        self.relative = relative
        self.path = opj(store.filedir, relative)

    def __add__(self, other):
        return StoreRelativePath(self.relative + other)

class Store(object):
    name = '(Store)'
    store = None
    persistenceForgets = ['cache', 'env', 'db']

    lock = None
    closed = False

    service = None

    HYGENE_PERIOD = 60 * 60
    _databaseHygeneLoop = None

    def __init__(self, datadir, filedir, archdir=None, join=False):
        """
        @param datadir: The directory in which to instantiate database files.
        @param filedir: The directory into which to place message files.
        @param archdir: The directory into which environment log files which
        are no longer needed will be moved.
        """
        self.filedir = filedir
        self.datadir = datadir
        self.tempdir = os.path.join(self.filedir, 'tmp')

        if archdir is None:
            archdir = sibpath(self.filedir, 'archive')
        self.archdir = archdir

        md(self.tempdir)
        dataWasntThere = md(self.datadir)
        self.setup(join)
        try:
            if dataWasntThere:
                self.transact(self.live)
            else:
                self.transact(self.resurrect)
        except:
            self.close()
            raise

    def resurrect(self):
        """Re-instantiate a database.
        """
        self.rootRef = ItemReference(self, 0)
        pupool = self.rootRef.getItem().getIndexItem(NameIndex, 'powerups')
        self.powerupsRef = pupool.referenceTo()
        from atop.powerup import PowerStation
        self.powerstation = PowerStation(self)
        self.powerstation._powerupsPool = self.powerupsRef

        # really, this ought to be handled by a different implementation of
        # IPowerStation considering the extra specialness that Store has to do,
        # but this at least fixes incorrect behavior with a relatively small
        # change... -gl

        for pup in pupool.queryIndex():
            self.powerstation.powerups[reflect.qual(pup.__class__)] = pup.referenceTo()

        self._databaseHygeneLoop = task.LoopingCall(self._databaseHygene)
        self._databaseHygeneLoop.start(self.HYGENE_PERIOD)

    def live(self):
        """This is the first time this store has been initialized.  Perform
        in-database initialization to create various objects which we will
        need.
        """

        # This is being done in this apparently round-about way (with root
        # vs. powerup pool) because we _need_ to be able to resurrect these
        # items deterministically if we lose track of the original Store
        # object.  If we lose the .tap, we need to be able to re-create it

        root = Pool(self, "root")
        root.addIndex(NameIndex)
        powerups = Pool(self, "powerups")
        powerups.addIndex(PowerupIndex)
        root.addItem(powerups)
        from atop.powerup import PowerStation
        self.powerstation = PowerStation(self)
        self.rootRef = root.referenceTo()
        self.powerupsRef = powerups.referenceTo()
        self.powerstation._powerupsPool = self.powerupsRef

        self._databaseHygeneLoop = task.LoopingCall(self._databaseHygene)
        self._databaseHygeneLoop.start(self.HYGENE_PERIOD)

    def _databaseHygene(self):
        # XXX - I /think/ self.backup() belongs in the middle
        self.checkpoint()
        # self.backup()
        self.archive()

    def listArchivable(self):
        """Return a list of filenames of archivable log files.

        The returned names are relative to C{self.datadir}.
        """
        return self.env.log_archive()

    def archive(self):
        """Archive all archivable log files.
        """
        arch = self.listArchivable()
        if not os.path.exists(self.archdir):
            os.mkdir(self.archdir)
        for fname in arch:
            shutil.move(opj(self.datadir, fname), opj(self.archdir, fname))
        if arch:
            log.msg("Archived " + ', '.join(arch))

    def checkpoint(self):
        """Checkpoint the database environment.

        This allows old environment logfiles to be archived.
        """
        log.msg("Checkpointing database...")
        self.env.txn_checkpoint()
        log.msg("Database checkpointed.")

    def touch(self):
        """This is a no-op for convenience.
        """

    def getComponent(self, interface, registry=None, default=None):
        # I am my own registry, really.
        # XXX - Not really!  This precludes external adapter registration
        # against Stores.
        from atop.powerup import IPowerStation
        if interface == IPowerStation:
            return self.powerstation
        ppool = self.powerupsRef.getItem()
        if context.get('txn') is None:
            warnings.warn("Calling Store.getComponent() outside a transaction is deprecated!",
                          DeprecationWarning, stacklevel=3)
        pitem = ppool.getIndexItem(PowerupIndex, reflect.qual(interface), default=default)
        if isinstance(pitem, ItemForwardingComponent):
            return pitem.fwdTo.getItem()
        return pitem

    __conform__ = getComponent

    def setComponent(self, interface, newItemComponent):
        if newItemComponent.store is not self:
            raise Inconsistency("You may only store top-level items as top-level powerups.")
        pups = self.powerupsRef.getItem()
        if newItemComponent not in pups:
            pups.addItem(newItemComponent)
        newItemComponent.appendToIndex(PowerupIndex, reflect.qual(interface))

    def unsetComponent(self, interfaceClass):
        try:
            oldItemComponent = interfaceClass(self)
        except CannotAdapt:
            raise KeyError(interfaceClass)
        else:
            pups = self.powerupsRef.getItem()
            oldItemComponent.removeFromIndex("powerup", reflect.qual(interfaceClass))
            for (iface, storeID) in pups.queryIndex('powerup', result=INDEX_AND_STORE_ID):
                if oldItemComponent.storeID == storeID:
                    break
            else:
                pups.removeItem(oldItemComponent)

    def removeComponent(self, itemComponent):
        if itemComponent.store is not self:
            raise Inconsistency("You cannot remove a foreign component from the store.")
        pups = self.powerupsRef.getItem()
        ifaces = []
        for (ifaceName, storeID) in pups.queryIndex(result=INDEX_AND_STORE_ID):
            if itemComponent.storeID == storeID:
                iface = reflect.namedAny(ifaceName)
                self.unsetComponent(iface)
                ifaces.append(iface)
        return ifaces

    def newSubFileForItem(self, substore):
        tmpname = os.path.join(self.tempdir, str(tempCounter.next())+".tmp")
        destpath = StoreRelativePath(str(substore.storeID), self)
        return AtomicFile(tmpname, destpath)

    def newSubDirForItem(self, item, name):
        thisPath = StoreRelativePath(os.path.join(str(item.storeID), name), self)
        md(thisPath.path)
        return thisPath

    def __getstate__(self):
        if context.get('item') is not None:
            fmt = "Serializing the Store from an Item (%r) in the Store (Check out %s)."
            item = context.get('item')
            raise Inconsistency(fmt % (item, _refPath(item, self)))
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__ = state
        self.setup()

    def setup(self, join=False):
        self.cache = FinalizingCache()
        if join:
            dbAndEnv = envdbjoin(self.datadir)
        else:
            dbAndEnv = envdblock(self.datadir)
        self.env, self.db = dbAndEnv.env, dbAndEnv.db
        self.cleanup = dbAndEnv.cleanup
        self.closed = False

    def close(self):
        if self.closed:
            raise StoreClosed()

        self.cleanup()
        if getattr(self, '_databaseHygeneLoop', None) is not None:
            self._databaseHygeneLoop.stop()
        self._databaseHygeneLoop = self.db = self.env = self.cleanup = None
        self.closed = True

    def getItemPath(self):
        return ()

    def nextID(self, allocate=True):
        """Allocate and return the next available store ID.
        """
        f = allocate and updateIDKey or retrieveIDKey
        return f(self.db, "next" + SEP + "id")

    def _idKeyName(self, item):
        return self._formatIDPath(item.getItemPath()) + SEP + 'next' + SEP + 'id'

    def nextSubID(self, item, allocate=True):
        """Allocate and return the next available store ID for the given item.

        These IDs are independent of any other IDs, including the ones for this
        Store instance.  If allocate is False, the ID is not actually allocated
        (a subsequent call to this function with the same arguments may return
        the same value).
        """
        txn = context.get('txn')
        keyName = self._idKeyName(item)
        value = self.getSubID(item, txn, keyName)
        if allocate:
            self.setSubID(item, value + 1, txn, keyName)
        return value

    def getSubID(self, item, _txn=None, _keyName=None):
        """Return the next available store ID for the given item.
        """
        if _txn is None:
            _txn = context.get('txn')
        if _txn is None:
            raise NotImplementedError("Updating ID outside of transaction")
        if _keyName is None:
            _keyName = self._idKeyName(item)
        return int(self.db.get(_keyName, flags=libdb.DB_RMW, txn=_txn) or 0)

    def setSubID(self, item, value, _txn=None, _keyName=None):
        """Change the next available store ID for the given item.
        """
        if _txn is None:
            _txn = context.get('txn')
        if _txn is None:
            raise NotImplementedError("Updating ID outside of transaction")
        if _keyName is None:
            _keyName = self._idKeyName(item)
        self.db.put(_keyName, str(value), txn=_txn)

    def setSubKey(self, item, key, value):
        ipfx = self._formatIDPath(item.getItemPath())
        self.db.put(ipfx + key, value, txn=context.get('txn'))

    def getSubKey(self, item, key):
        ipfx = self._formatIDPath(item.getItemPath())
        return self.db.get(ipfx + key, txn=context.get('txn'))

    def delSubKey(self, item, key):
        ipfx = self._formatIDPath(item.getItemPath())
        self.db.delete(ipfx + key, txn=context.get('txn'))

    def hasSubKey(self, item, key):
        ipfx = self._formatIDPath(item.getItemPath())
        return self.db.get(ipfx + key, doff=0, dlen=0, txn=context.get('txn')) is not None

    def _loadItemByPath(self, ctxstore, path, alreadyLoadedItem=None):
        """
        Retrieve an item given a sequence of integers.

        @param alreadyLoadedItem: None, or an already-loaded-already-cached
        version of this item.
        """
        stringKey = self._formatIDPath(path)
        stringValue = self.db.get(stringKey, txn=context.get('txn'))
        if stringValue is None:
            return None
        if stringValue[0] == PICKLE:
            myObject = context.call(
                {"store": ctxstore,
                 "item": alreadyLoadedItem}, versioning.loadWithVersions,
                stringValue, withRevert=alreadyLoadedItem,
                beforeAfterUpgrade=lambda x: self.cacheItem(x, path))
        elif stringValue[0] == MULTIPASS:
            klass = reflect.namedAny(stringValue[1:])
            myObject = klass.fromStorage(self, path[-1])
        else:
            assert False, ("Don't know how to load", repr(stringValue))

        log.msg(interface=iatop.IStatEvent, name='database',
                stat_load_bytes=len(stringValue), path=path,
                item=myObject, pickle=stringValue)

        if alreadyLoadedItem is None:
            self.cacheItem(myObject, path)
            ctx = context.get('NewlyLoadedContext')
            if ctx is None:
                warnings.warn("Loading items outside of a transaction is deprecated!",
                    DeprecationWarning, stacklevel=4)
            else:
                ctx[path] = None
            if stringValue[0] == PICKLE:
                myObject.activate()
        return myObject


    def cacheItem(self, item, path):
        assert isinstance(path, tuple)
        if self.cache.has(path):
            assert self.cache.get(path) is item, "inconsistency!?!  the db manages this now.  Cached item %r at %r is not %r" % (self.cache.get(path), path, item)
        else:
            self.cache.cache(path, item)

    def uncacheItem(self, item, path):
        assert isinstance(path, tuple)
        assert self.cache.get(path) is item, "inconsistency!?!  the db manages this now.  Trying to uncache item %r at %r is not %r" % (self.cache.get(path), path, item)
        self.cache.uncache(path)


    def _saveItemToPath(self, item, path):
        self.cacheItem(item, path)
        stringKey = self._formatIDPath(path)
        stringValue = context.call({'item': item}, versioning.saveWithVersions, item)
        log.msg(interface=iatop.IStatEvent, name='database', stat_store_bytes=len(stringValue), path=path, item=item, pickle=stringValue)
        self.db.put(stringKey, stringValue, txn=context.get('txn'))

    def _removeItemFromPath(self, item, path):
        self.uncacheItem(item, path)
        stringKey = self._formatIDPath(path)
        self.db.delete(stringKey, txn=context.get('txn'))
        try:
            self.db.delete(stringKey + SEP + 'next' + SEP + 'id', txn=context.get('txn'))
        except KeyError:
            pass
        log.msg(interface=iatop.IStatEvent, name='database', stat_deletes=1, path=path, item=item)

    def _getItemByPath(self, ctxstore, *path):
        if self.cache.has(path):
            return self.cache.get(path)
        else:
            obj = self._loadItemByPath(ctxstore, path, None)
            if obj is None:
                raise KeyError(path)
            return obj

    def getItemByID(self, itemID):
        """Return the item from this Store with the given ID.

        @type itemID: C{int}
        """
        return self._getItemByPath(self, itemID)

    def getSubItem(self, subStore, subItemID):
        return self._getItemByPath(subStore, *subStore.getItemPath() + (subItemID,))

    def lockItem(self, subItem):
        self.lockSubItem(self, subItem)

    def deleteItem(self, item):
        path = item.getItemPath()
        try:
            self._removeItemFromPath(item, path)
        except KeyError, e:
            # It has not been added to the database yet.
            pass
        context.get('UnstoredFinalizationContext')[path] = item.removedFromStore()
        context.get('AntiStorageContext')[path] = item
        store = item.store
        storeID = item.storeID
        was = str(item)
        item.__dict__.clear()
        item.__class__ = DeletedItem
        vars(item)['store'] = store
        vars(item)['storeID'] = storeID
        vars(item)['was'] = was

    def lockSubItem(self, subStore, subItem):
        p = subItem.getItemPath()
        stringKey = self._formatIDPath(p)
        self.db.get(stringKey, flags=libdb.DB_RMW, dlen=0, doff=0, txn=context.get('txn'))
        log.msg(interface=iatop.IStatEvent, name='database', stat_locks=1, path=p, item=subItem)

    def _inMemoryRevert(self, collection, skip):
        """Revert a collection of objects to their state in the database.
        """
        for (path, item) in collection:
            if path in skip:
                if self.cache.has(path):
                    self.cache.uncache(path)
            else:
                dupItem = self._loadItemByPath(item.store, path, item)
                if dupItem is not None:
                    assert dupItem is item
                elif self.cache.has(path):
                    self.cache.uncache(path)
                else:
                    log.msg("Reversion peculiarity: %s" % (item,))

    def new_txn(self):
        return self.env.txn_begin(context.get('txn'))

    def _subtransact(self, f, *a, **k):
        """Open a database transaction object and commit it.
        """

        # Start a Berkeley DB transaction.  All database calls made in
        # this Atop transaction will use this BDB transaction object.
        txn = self.new_txn()

        # Items modified in this transaction are added as values to this
        # dictionary (Their item path is the key) via the Item.touch()
        # method.  Items belonging to it at the end of the transaction are
        # serialized and committed to the database.
        storageContext = {}

        # Items deleted in this transaction are added as values to this
        # dictionary (Their item path is the key) via the Store.deleteItem()
        # method.  Items are deleted immediately upon the call to
        # Store.deleteItem().  This dictionary is keyed on tupled item paths
        # as well.
        antiStorageContext = {}

        # Items deleted in this transaction have the return value of the
        # `removedFromStore' added as values to this dictionary (Their item
        # path is the key).  If the transaction is committed, each callable
        # is invoked.
        unstoredFinalizationContext = {}

        # Items which were not cached before this transaction began but were
        # loaded into memory during this transaction are added as values to
        # this dictionary (Their item path is the key).  These Items do not
        # need to be reverted if the transaction is aborted.  They are
        # simply discarded instead.
        newlyLoadedContext = {}

        # A factory for creating cursor objects.  These cursors are all
        # closed when the transaction is committed or aborted.
        cursorFactory = CursorFactory(self, txn)
        try:
            retvalue = context.call({'store': self,
                                     'txn': txn,
                                     'StorageContext': storageContext,
                                     'AntiStorageContext': antiStorageContext,
                                     'UnstoredFinalizationContext': unstoredFinalizationContext,
                                     'NewlyLoadedContext': newlyLoadedContext,
                                     'CursorFactory': cursorFactory},
                                    f, *a, **k)
            context.call({'txn': txn},
                         self._inMemoryCommit,
                         storageContext.iteritems(),
                         unstoredFinalizationContext)
        except:
            t, v, tb = sys.exc_info()
            # log.err(failure.Failure(v, t, tb))
            cursorFactory.abort()
            txn.abort()
            log.msg(interface=iatop.IStatEvent, name='database', stat_transaction_aborts=1)
            # is there a way to make this transaction read-only?  -glyph
            revertReadTxn = self.new_txn()
            try:
                try:
                    context.call({'txn': revertReadTxn},
                                 self._inMemoryRevert,
                                 itertools.chain(storageContext.iteritems(), antiStorageContext.iteritems()),
                                 newlyLoadedContext)
                finally:
                    revertReadTxn.abort()
            except:
                log.msg("In memory revert failed.  Things are going to get weird.")
                log.err()
            raise t, v, tb
        else:
            cursorFactory.commit()
            txn.commit()            # this can fail with a RUNRECOVERY - that would be bad.
            if antiStorageContext:
                self._postCommitCleanup(unstoredFinalizationContext)
            log.msg(interface=iatop.IStatEvent, name='database', stat_transaction_commits=1)
            return retvalue

    def _inMemoryCommit(self, storeCollection, skip):
        for (path, item) in storeCollection:
            if path not in skip:
                # save the item by ID, either overwriting or adding
                self._saveItemToPath(item, path)

    def _postCommitCleanup(self, unstoredItems):
        for (path, finalizer) in unstoredItems.iteritems():
            try:
                finalizer()
            except:
                log.msg("Error in post-commit cleanup for item %s:" % (path,))
                log.err()

    def transact(self, f, *a, **k):
        """
        Run a function or method within the context of the storage system.

        This will keep track of all items created or touched during the
        duration of this function and save or revert them based on whether the
        function completes successfully.
        """
        if context.get('txn'):
            return f(*a,**k)
        if self.closed:
            raise StoreClosed("can't run %r %r %r" % (f, a, k))
        return self._subtransact(f, *a, **k)

    def transback(self, result, f, *a, **k):
        """Utility method for using 'transact' as a callback to a Deferred.

        All this does is reverse the order of the first two arguments to
        'transact'.  Its common usage is:

            d = doSomethingLater()
            d.addCallback(store.transback, F, 'hello')

        As you might imagine, this causes F to be called as

            F(deferred-result, 'hello')

        Note that this typically causes the Deferred to become
        unserializable, making this method primarily of use for
        code dealing with ephemeral state (for example, protocol
        implementations).
        """
        return self.transact(f, result, *a, **k)

    def batch(self, f, *a, **kw):
        """Run a batch processing function as many times as is possible.

        f(*a, **kw) will be invoked until time or database limits are reached.
        Each invocation of f should perform a small amount of work towards a
        larger goal.  If f(*a, **kw) returns a True value, processing will cease.

        @return: The number of times the function was called.
        """
        count = 1
        stat = self.env.lock_stat
        def locksRemaining():
            lockInfo = stat()
            return lockInfo['maxlocks'] - lockInfo['nlocks']
        startLocks = locksRemaining()
        if f(*a, **kw):
            return count
        lockChange = startLocks - locksRemaining()
        minLocks = max(startLocks - 20, lockChange)
        while locksRemaining() > minLocks:
            count += 1
            if f(*a, **kw):
                return count
        return count

    def _formatIDPath(self, path):
        L = ['%0.10d'] * len(path)
        L.insert(0, 'id')
        return ((SEP.join(L)) % tuple(path))

class CursorFactory:
    def __init__(self, store, txn):
        self.store = store
        self.txn = txn
        self.cursors = []
        self.done = False

    def cursor(self):
        if self.done:
            raise Inconsistency("Re-use of a storage factory.  You suck!")
        self.cursors.append(self.store.db.cursor(self.txn))
        return self.cursors[-1]

    def closeCursors(self):
        self.done = True
        for cursor in self.cursors:
            cursor.close()

    def commit(self):
        self.closeCursors()

    def abort(self):
        self.closeCursors()

class ItemReference:
    def __init__(self, store, storeID, iface=None):
        self.store = store
        self.storeID = storeID
        self.iface = iface

    def getItem(self):
        if self.iface is not None:
            return self.iface(self.store.getItemByID(self.storeID))
        return self.store.getItemByID(self.storeID)

    def __repr__(self):
        return '<ItemReference to %s, %d>' % (getattr(self.store, 'storeID', None), self.storeID)

    def __hash__(self):
        return hash((self.store, self.storeID))

    def __setstate__(self, state):
        self.__dict__ = state
        if not hasattr(self, 'store'):
            st = context.get('store')
            if st is None:
                raise Inconsistency("You're loading an item reference outside the context of a store.")
            self.store = st

    def __ne__(self, other):
        #PYTHON IS SO AWESOME
        return not self.__eq__(other)

    def __eq__(self, other):
        if not isinstance(other, ItemReference):
            return False
        return (other.store is self.store) and (other.storeID == self.storeID)

    def __getstate__(self):
        # get the store we're saving to.
        topitem = context.get('item')
        if topitem is None:
            return self.__dict__
##             raise Inconsistency("Why are you saving an item reference outside the scope of an item?  "
##                                 "Consider using RemoteItemReference (when it becomes available.  "
##                                 "Talk to glyph.")
        elif topitem.store is self.store:
            # If the store we're saving to is the same store as the one for the
            # item we're storing a reference to, remember it from context on
            # load.
            dct = self.__dict__.copy()
            del dct['store']
            return dct
        elif topitem is self.store:
            # Most likely, what's happening here is that we're saving a
            # reference to a sub-item of the avatar within the avatar itself.
            # This will only happen in substores, because 'topitem is
            # self.store' means it's a sub-item of the top item which we are
            # currently saving.

            # In this case it's perfectly safe to return a synchronous
            # reference to the store, because it's going to be remembered by
            # the pickler.
            return self.__dict__
        else:
            if self.getItem() is topitem.store:
                raise Inconsistency("Trying to store an ItemReference on %s to its own store! Don't *do* that." % (topitem,))

            fmt = "Ref to %s stored from %s in %s should be %s (Check out %s)."
            raise Inconsistency(fmt % (self.getItem(),
                                       topitem,
                                       topitem.store,
                                       self.getItem().store,
                                       _refPath(topitem, self)))

def _refPath(a, b):
    # Convenience la la
    from twisted.python.reflect import objgrep, isSame
    return objgrep(a, b, isSame, showUnknowns=True)


class referenced(object):
    """Use this on an Item to automatically refer to other items.  For example::

    class Foo(Item):
        ...
    class Bar(Item):
        fooX = referenced
        fooY = referenced
        def setFoos(self, x, y):
            'x and y are Foo instances'
            self.fooX = x
            self.fooY = y
            self.touch()
        def getFoos(self):
            'returns a tuple of 2 Foo instances'
            return self.fooX, self.fooY

    Note that referenced does not automatically .touch() your object.
    """

    def __init__(self, iface):
        self.iface = iface

    def setName(self, attrname):
        self.name = "_"+attrname+"Ref"

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        x = getattr(obj, self.name, None)
        if x:
            if self.iface:
                return self.iface(x.getItem())
            else:
                return x.getItem()

    def __set__(self, obj, value):
        if value is not None:
            if hasattr(obj, 'store') and value.store is not obj.store:
                raise Inconsistency("You can't reference foreign items directly.")
            if self.iface:
                value = IPersistable(value).referenceTo()
            else:
                value = value.referenceTo()
        setattr(obj, self.name, value)


def transacted(method):
    """Use this on an Item to automatically run a method in a transaction.
    For example::

    class Foo(Item):
        def bar(self, aPool):
            aPool.addItem(self)
        bar = transacted(bar)
    """
    def _(self, *a, **kw):
        return self.store.transact(method, self, *a, **kw)
    return new.function(_.func_code, _.func_globals, _.func_name,
                        inspect.getargspec(_), _.func_closure)

class MetaItem(type):
    """I automate the various indexed attributes of an Item.
    """
    def __new__(metaclass, name, bases, ns):
        for k, v in ns.items():
            if v is referenced:
                r = v(None)
                r.setName(k)
                ns[k] = r
            if isinstance(v, referenced):
                ns[k] = v
                v.setName(k)
        return super(MetaItem, metaclass).__new__(metaclass, name, bases, ns)


def deleteIfNotDeleted(path, deep=False):
    item = context.get('store')
    for segment in path[:-1]:
        item = item.getItemByID(segment)
    try:
        item = item.getItemByID(path[-1])
    except KeyError:
        pass
    else:
        if deep and isinstance(item, Pool) and item.countIndex(UIDIndex):
            from atop.upgrade import demolishPool
            return demolishPool(item
                ).addCallback(*_sim(item._deleteHelper)
                )
        return item.deleteFromStore()

def _deleteContainedItems(ignored, paths, deep):
    return amap(deleteIfNotDeleted, paths, [deep] * len(paths))


class ItemMixIn(object):
    """I am a base class for mix-ins designed for Items in order to create appropriate properties.
    """
    __metaclass__ = MetaItem


class Item(versioning.Versionable):

    __metaclass__ = MetaItem

    persistenceForgets = ['store']
    contextRemembers = [('store', 'store')]

    name = indexed(NAME)

    def __init__(self, store, name=None):
        self.storeID = store.nextID()
        self.store = store
        self._indexes = {}
        self.store.cacheItem(self, self.getItemPath())
        if name is not None:
            self.name = name
        self.dateCreated = Time()
        self.touch()

    def _str(self, fmt):
        try:
            kls = self.__class__.__name__
        except:
            kls = '<error: class name unreachable>'
        try:
            name = self.name
        except:
            name = '<error: item name unreachable>'
        try:
            path = self.getItemPath()
        except:
            path = '<error: item path unreachable>'
        return fmt % {'class': kls, 'name': name, 'path': path, 'id': id(self)}


    def __str__(self):
        return self._str('<%(class)s %(name)r at 0x%(id)x>')

    def __repr__(self):
        return self._str('<%(class)s %(name)r %(path)r>')

    def __statecheck__(self):
        if self is not context.get('item'):
            fmt = "You were saving %s but I found %s in it (Check out %s)."
            item = context.get('item')
            raise Inconsistency(fmt % (item, self, _refPath(item, self)))

    def __getstate__(self):
        self.__statecheck__()
        return versioning.Versionable.__getstate__(self)

    def fwd(self, *interfaces):
        if not interfaces:
            interfaces = self.__implements__
        return makeForwarder(interfaces)(self.referenceTo())

    def _deleteContained(self):
        shallow = []
        deep = []
        for cls in [self.__class__] + reflect.allYourBase(self.__class__):
            if 'deleteShallow' in vars(cls):
                shallow.extend([o.getItemPath() for o in cls.deleteShallow(self)])
            if 'deleteDeep' in vars(cls):
                deep.extend([o.getItemPath() for o in cls.deleteDeep(self)])
        return _deleteContainedItems(None, shallow, False
            ).addCallback(_deleteContainedItems, deep, True
            )

    def _unsetComponents(self):
        for cls in [self.__class__] + reflect.allYourBase(self.__class__):
            if 'setsStoreInterfaces' in vars(cls):
                for iface in cls.setsStoreInterfaces(self):
                    self.store.unsetComponent(iface)

    def deleteFromStore(self):
        """Remove this Item from the database.
        """
        for p in self.iterpools():
            p.removeItem(self)
        return self._deleteContained(
            ).addCallback(*_sim(self._cbDeletedContained, str(self))
            )

    def _deleteHelper(self, ignored):
        return self.deleteFromStore()

    def _cbDeletedContained(self, result, s):
        self._unsetComponents()
        self.store.deleteItem(self)

    def removedFromStore(self):
        """
        Called after the transaction in which deleteFromStore() was invoked
        has been committed.

        This should remove any ancillary data not present within the database itself.  This
        method is _not_ invoked in a transaction, and transactional code should _not_ be
        run from this method.
        """
        return noop

    def referenceTo(self, iface=None):
        return ItemReference(self.store, self.storeID, iface)
    getRef = referenceTo

    def getItemPath(self):
        """Return a list describing this item's location in the main store.
        """
        ids = [self.storeID]
        ss = self.store
        while ss.store is not None:
            ids.append(ss.storeID)
            ss = ss.store
        ids.reverse()
        return tuple(ids)

    def setDataKey(self, key, value):
        return self.store.setSubKey(self, key, value)

    def getDataKey(self, key):
        return self.store.getSubKey(self, key)

    def delDataKey(self, key):
        return self.store.delSubKey(self, key)

    def hasDataKey(self, key):
        return self.store.hasSubKey(self, key)

    def __finalizer__(self):
        return noop

    def activate(self):
        """This method is called when this item has been loaded from the DB.
        """

    def reactivate(self, oldAttributes):
        """This method is called when this item has been re-loaded from the DB
        during a transaction revert.  Try to avoid putting complex code here;
        bugs here will likely be very difficult to understand.

        @param oldAttributes: a dict of the old {names: values} of attributes
        specified in persistenceForgets.
        """
        self.__dict__.update(oldAttributes)

    def touch(self):
        """Notify that this thingy is dirty and in need of persistament.

        This must be called prior to modifying any persistent state of an item
        within a transaction.  It is needed so that atop knows what items must
        be saved to commit the transaction, or reverted if the transaction
        fails.
        """
        sctx = context.get('StorageContext')
        if sctx is None:
            raise Inconsistency("You're touching an object outside a transaction.")
        ikey = self.getItemPath()
        sctx[ikey] = self

    def iterpools(self):
        return self.poolToUID.iterkeys()

    def change(self, notificationData=None):
        """Send out a notification that something about this item has changed.
        This will notify observers on the pools which this item inhabits.

        @param notificationData: an object that represents the change.  XXX
        TODO: donovan, what kind of change object do you want here?
        """
        for value in self.iterpools():
            value._announceItem(self, "itemChanged", notificationData)

    def lock(self):
        """Prepare this object for database update.
        """
        self.touch()
        self.store.lockItem(self)

    def setIndexKey(self, index, newKey, notify=False):
        """Change the value of a single-value index key on this item to the new
        key.
        """
        index = _deprecatedIndexLookup(index)
        self._assertSingleIndex(index)
        self.touch()
        for pool in self.iterpools():
            oldKey = self.getIndexKey(pool, index)
            pool.changeKeyForItem(self, index, oldKey, newKey, notify=notify)
        self._setIndexKeyData(index, newKey)

    def _iterOf(self, onlyOnePool=None):
        if onlyOnePool is not None:
            if self in onlyOnePool:
                itr = iter([onlyOnePool])
            else:
                itr = iter(())
        else:
            itr = self.iterpools()
        return itr

    def appendToIndex(self, index, indexKey, onlyOnePool=None):
        """Be careful with this method.  The store will not automatically
        remove state added using this method from a pool.  removeFromIndex,
        similarly.
        """
        for pool in self._iterOf(onlyOnePool):
            pool.changeKeyForItem(self, index, None, indexKey)

    def removeFromIndex(self, index, indexKey, onlyOnePool=None):
        """The inverse of appendToIndex.
        """
        for pool in self._iterOf(onlyOnePool):
            pool.changeKeyForItem(self, index, indexKey, None)

    def getIndexKey(self, pool, index, default=None):
        """Get a key value for a particular index.  The key value must match
        what the index expects.
        """
        index = _deprecatedIndexLookup(index)
        if hasattr(index, 'methodPrefix'):
            m = getattr(self, "index_" + index.methodPrefix, None)
            if m:
                warnings.warn_explicit("prefixed methods are deprecated",
                                       DeprecationWarning,
                                       m.im_func.func_code.co_filename,
                                       m.im_func.func_code.co_firstlineno,
                                       registry=_indexMethodWarnReg)
                return m(pool)
        uvo = index.useValueOf
        result = self._indexes.get(uvo, Item._noitem)
        if result is not Item._noitem:
            return result
        if pool is not None:
            diffkey = ('diff', uvo)
            if self._indexes.has_key(diffkey):
                di = self._indexes[diffkey]
                if di.has_key(pool.storeID):
                    return di[pool.storeID]
            iis = IInvertedIndexStorage(pool, None)
            if iis:
                iisk = iis.getItemIndexKey(self, index)
                if iisk is not None:
                    return iisk
        key = index.keyFromItem(pool, self)
        if key is not None:
            return key
        return default

    def getIndexKeys(self, pool, index):
        """Get all applicable keys for a particular index.
        """
        uvo = index.useValueOf
        mi = self._indexes.get(('multi', uvo))
        if mi:
            return mi
        iis = IInvertedIndexStorage(pool, None)
        if iis:
            nval = iis.getItemIndexKeys(self, index)
            if nval is not None:
                return nval
        single = self.getIndexKey(pool, index)
        if single is not None:
            return [single]
        return []

    _noitem = object()
    _indexes = None

    def _assertSingleIndex(self, index):
        if (self._indexes.has_key(('multi', index.useValueOf)) or
            self._indexes.has_key(('multi-diff', index.useValueOf))):
            raise RuntimeError("There is more than one value for %s in %s" % (
                index, reflect.qual(self.__class__)))

    def _setIndexKeyData(self, index, newValue):
        if hasattr(index, 'methodPrefix'):
            m = getattr(self, "setindex_" + index.methodPrefix, None)
            if m:
                warnings.warn_explicit("prefixed methods are deprecated",
                                       DeprecationWarning,
                                       m.im_func.func_code.co_filename,
                                       m.im_func.func_code.co_firstlineno,
                                       registry=_indexMethodWarnReg)
                return m(newValue)
        self._indexes[index.useValueOf] = newValue

    class _DictWatcher(object):
        __slots__ = 'item index'.split()

        def __init__(self, item, index):
            self.item = item
            self.index = index

        def added(self, addedPool, addedKey):
            self.item.appendToIndex(self.index, addedKey, addedPool)

        def removed(self, removedPool, removedKey):
            self.item.removeFromIndex(self.index, removedKey, removedPool)

    class _ListWatcher(object):
        """I am a mostly invisible helper object.  I hold state for callbacks
        related to proxies which represent lists of key values.

        I am a nested class and also an object with slots on purpose: I should
        NEVER be pickled.
        """

        __slots__ = 'item index pool'.split()
        def __init__(self, item, index, pool=None):
            self.item = item
            self.index = index
            self.pool = pool

        def added(self, addedKey):
            return self.item.appendToIndex(self.index, addedKey, self.pool)

        def removed(self, removedKey):
            return self.item.removeFromIndex(self.index, removedKey, self.pool)

    def _multiValueProxy(self, index):
        """Return a list-like object that proxies for multiple values of an
        index.  Initialize the index for multiple values if it hasn't been
        already.  This may touch this item but it will try not to.
        """
        touched = False
        mkey = ('multi', index.useValueOf)
        if self._indexes.has_key(index):
            self.touch()
            touched = True
            # We are upgrading from a single index.
            del self._indexes[index]
        if not self._indexes.has_key(mkey):
            if not touched:
                self.touch()
            self._indexes[mkey] = []
        lw = Item._ListWatcher(self, index)
        return notify.NotifyList(self._indexes[mkey], lw.added, lw.removed, True, storeValue=index.storeKey, loadValue=index.loadKey)

    def _differentValueProxy(self, index):
        """Return a dict-like object that proxies for different values of an
        index in different pools.
        """
        # self._indexes[index] is the default value
        self._assertSingleIndex(index)
        dkey = ('diff', index.useValueOf)
        if not self._indexes.has_key(dkey):
            self.touch()
            self._indexes[dkey] = {}
        dw = Item._DictWatcher(self, index)
        return notify.NotifyDict(notify.FilteringDict(self._indexes[dkey],
                                        loadKey=self.store.getItemByID,
                                        storeKey=_getstoreid,
                                        loadValue=index.loadKey,
                                        storeValue=index.storeKey),
                          dw.added, dw.removed)

    def _differentMultiValueProxy(self, index):
        mdkey = ('multi-diff', index.useValueOf)
        if not self._indexes.has_key(mdkey):
            self.touch()
            self._indexes[mdkey] = {}
        return Item._MultiValueDictEmulator(self, index, self._indexes[mdkey])

    class _MultiValueDictEmulator(object):
        __slots__ = ['item',
                     'index',
                     'data']

        def __init__(self, item, index, data):
            self.item = item
            self.index = index
            self.data = data

        def __getitem__(self, pool):
            assert isinstance(pool, Pool), "I don't know how how to handle anything but pools"
            if not self.data.has_key(pool.storeID):
                self.item.touch()
                self.data[pool.storeID] = []
            lw = Item._ListWatcher(self.item, self.index, pool)
            return notify.NotifyList(self.data[pool.storeID], lw.added, lw.removed, True, storeValue=self.index.storeKey, loadValue=self.index.loadKey)

        def __setitem__(self, pool, data):
            self[pool][:] = data

    def addedToPool(self, pool, uid=None):
        if pool.store != self.store:
            raise Inconsistency(
                "You can't add foreign items directly to pools.")
        if pool not in self.poolToUID:
            self.touch()
            if uid is None:
                uid = pool.nextUID()
            self.poolToUID[pool] = uid
        # else: already in pool?  what to do?

    def removedFromPool(self, pool):
        if pool.store != self.store:
            raise Inconsistency(
                "You can't add foreign items directly to pools.")
        self.touch()
        del self.poolToUID[pool]

    poolToUID = indexedPoolDict(UIDIndex)

    def _bcgetpools(self):
        warnings.warn("don't use item.pools, use item.poolToUID.  "
                      "the user-visible values have changed from {int:int} to {item:item}",
                      DeprecationWarning, stacklevel=2)
        return notify.FilteringDict(self.poolToUID,
                             loadKey=_getstoreid,
                             storeKey=self.store.getItemByID)

    pools = property(_bcgetpools)

    dateCreated = indexed(DateIndex)

    def newCacheDir(self, dirname):
        cur = self.store
        while cur.store is not None:
            cur = cur.store
        thisPath = os.path.join(cur.filedir, str(self.store.storeID), 'cache', str(self.storeID), dirname)
        md(thisPath)
        return thisPath

def _getstoreid(x):
    return x.storeID

_indexMethodWarnReg = {}

class DeletedItem(Item):
    """When an in-memory Item is deleted, it becomes an instance of this class.

    Ideally, nothing but the code which deleted the instance refernces the
    Item. By setting the deleted instance's class to DeletedItem, anything that
    references the now deleted, non-existant Item will break loudly.

    Eventually, deletion in atop will be handled automatically, and this can go
    away.
    """
    def __str__(self):
        return '<Deleted Item @ %X (was %r)>' % (id(self), self.was)

    def __repr__(self):
        return '<Deleted Item @ %X (was %r)>' % (id(self), self.was)

    def __setattr__(self, name, value):
        raise AttributeError(name)

    def setIndexKey(self, key, value):
        pass

    def deleteFromStore(self):
        raise RuntimeError("You can't delete something twice.")

class IComponentForwarder(Interface):
    """A forwarder which can be usd with SubStore's implementation of
    getComponent.

    The use-case for this class is to provide a way for provisioned powerups to
    return a 'I'm not implemented' component in cases where the provisioning
    has run out.
    """
    def forwardComponent(self, interface, default):
        """Return a proxy for myself which may vary in certain situations.
        """

class IInvertedIndexStorage(Interface):
    """I am an interface used by alternative pool implementations to indicate
    pool-specific indexes which are stored externally to an item.
    """

    def getItemIndexKeys(self, item, index):
        """Return multiple values for an item's index key.
        """

    def getItemIndexKey(self, item, index):
        """Return a single value for an item's index key.
        """

class SubStore(Item):
    def __init__(self, *a, **b):
        Item.__init__(self, *a, **b)
        self._adapterCache = {}
        p = Pool(self, name='Root Pool')
        p.addIndex(NameIndex)
        p.suggestIndex(NameIndex, False)
        self.rootPoolRef = p.referenceTo()

    def db(self):
        return self.store.db
    db = property(db)

    def _formatIDPath(self, path):
        return self.store._formatIDPath(path)

    def deleteFromStore(self):
        # Remove from all pools so it cannot be accessed
        for p in self.iterpools():
            p.removeItem(self)
        from atop import upgrade
        return upgrade.SubStoreDeletion(self).start()

    def getRootPool(self):
        """Return the root pool for this store.
        """
        return self.rootPoolRef.getItem()

    def __getstate__(self):
        state = Item.__getstate__(self)

        x = state['_adapterCache'] = state['_adapterCache'].copy()
        for k,v in x.items():
            if isinstance(v, Ephemeral):
                del x[k]
        return state

    # make sure that all appropriate interfaces touch me before making a
    # potentially destructive change!!  TODO: add special-casing for ephemerals

    def setComponent(self, interfaceClass, component):
        self.touch()
        if isinstance(component, Item):
            # if it's going to be automatic, it might as well be symmetrical
            component = ItemForwardingComponent(component.referenceTo())
        self._adapterCache[interfaceClass] = component

    def getComponent(self, interface, registry=None, default=None):
        if interface in self._adapterCache:
            component = self._adapterCache[interface]
            if isinstance(component, ItemForwardingComponent):
                component = component.fwdTo.getItem()
            if doesImplement(component, IComponentForwarder):
                component = component.forwardComponent(interface)
            return component
        else:
            if registry is None:
                adapterFactory = getAdapterFactory(self.__class__, interface, None)
                if adapterFactory is not None:
                    adapter = adapterFactory(self)
                else:
                    return default
            else:
                # the registry will not be None in old Twisted, so we're in 1.3
                # land now
                adapter = registry.getAdapter(self, interface, None,
                                              registry.getAdapterClassWithInheritance)
            if adapter is not None:
                if not getattr(adapter, 'temporaryAdapter', False):
                    self.setComponent(interface, adapter)
                return adapter
            return default

    __conform__ = getComponent

    def unsetComponent(self, interfaceClass):
        self.touch()
        del self._adapterCache[interfaceClass]

    def newFile(self):
        return self.store.newSubFileForItem(self)

    def newDirectory(self, name):
        """Create a directory in the user's "address
        space" or return the existing one
        """
        return self.store.newSubDirForItem(self, name)

    def setSubKey(self, item, key, value):
        return self.store.setSubKey(item, key, value)

    def getSubKey(self, item, key):
        return self.store.getSubKey(item, key)

    def delSubKey(self, item, key):
        return self.store.delSubKey(item, key)

    def hasSubKey(self, item, key):
        return self.store.hasSubKey(item, key)

    def nextID(self, allocate=True):
        return self.store.nextSubID(self, allocate)

    def nextSubID(self, item, allocate=True):
        return self.store.nextSubID(item, allocate)

    def getSubID(self, item):
        return self.store.getSubID(item)

    def setSubID(self, item, value):
        return self.store.setSubID(item, value)

    def cacheItem(self, item, path):
        return self.store.cacheItem(item, path)

    def getItemByID(self, itemID):
        return self.store.getSubItem(self, itemID)

    def transact(self, f, *a, **k):
        if not context.get('substore'):
            # XXX FIXME.  Allow hack in DSUPool to locate the substore from
            # context.  Should there be a better, actually-supported way of
            # doing this?
            return context.call({'substore': self},
                                self.store.transact, f, *a, **k)
        else:
            return self.store.transact(f, *a, **k)

    def transback(self, result, f, *a, **k):
        return self.transact(f, result, *a, **k)

    def batch(self, f, *a, **kw):
        return self.store.batch(f, *a, **kw)

    def lockItem(self, item):
        return self.store.lockSubItem(self, item)

    def deleteItem(self, item):
        return self.store.deleteItem(item)

class StorageService(service.Service):
    def __init__(self, store):
        self.setName(reflect.qual(self.__class__))
        self.store = store
        self.store.service = self

    def startService(self):
        service.Service.startService(self)
        def _():
            svc = service.IService(self.store, default=None)
            if svc is not None:
                return svc.startService()
        return self.store.transact(_)

    def _closeLater(self, result):
        d = defer.Deferred()
        from twisted.internet import reactor
        reactor.callLater(0, self._reallyClose, d, result)
        return d

    def _reallyClose(self, d, result):
        self.store.close()
        d.callback(result)

    def stopService(self):
        service.Service.stopService(self)
        def _():
            svc = service.IService(self.store, default=None)
            if svc is not None:
                return svc.stopService()
            return defer.succeed(None)
        return self.store.transact(_).addCallback(self._closeLater)

# subsubstore.store = substore; substore.store = realstore;
# subsubstore.getItemByPath => context.get('store') == subsubstore

ITEM               = object()
INDEX              = object()
INDEX_AND_ITEM     = object()
STORE_ID           = object()
INDEX_AND_STORE_ID = object()

class IPipe(Interface):
    def itemAdded(self, item):
        pass

    def itemRemoved(self, item):
        pass

    def itemChanged(self, item, notificationData):
        pass

class IPool(Interface):
    def suggestIndex(self, index, backwards=False):
        pass

    def suggestView(self, view):
        pass

    def addPipe(self, pipe, ephemeral=False):
        pass

    def removePipe(self, pipe, ephemeral=False):
        pass

    def queryIndex(self, index=UIDIndex, startKey=None, endKey=None, result=ITEM,
                   backwards=False, limit=None, startItemID=None, endItemID=None):
        pass

    def getSequence(self):
        pass

    def getIndexItem(self, index, indexKey=None, default=object()):
        pass

    def nextUID(self, allocate=True):
        pass

    def skipToUID(self, uid):
        pass

    def __contains__(self, item):
        pass

    def changeKeyForItem(self, item, ipref, oldKey, newKey, notify=False):
        pass

    def addItem(self, item, uid=None):
        pass

    def removeItem(self, item):
        pass

    def addIndex(self, index, name=None):
        pass

    def removeIndex(self, index):
        pass


def key2str(whatKey, whatItemID, towardsKey, pfx, obj):
    if whatKey is not None:
        whatStr = pfx + obj.keyToSortableString(whatKey)
        if whatItemID:
            whatStr += SEP + ('%0.10d' % whatItemID)
        elif towardsKey:
            whatStr += MAGIC
    elif towardsKey:
        whatStr = pfx + END
    else:
        whatStr = pfx
    return whatStr


class PoolQuery:
    cursorFactory = None
    cursorMethod = None

    def __init__(self, pool, index=UIDIndex, startKey=None, endKey=None, result=ITEM,
                 backwards=False, limit=None, startItemID=None, endItemID=None):
        """Retrieve information about the contents of this pool.

        An iterator which yields the requested information about the requested
        Items is returned.  This generator I{must} be iterated in a storage
        context.

        @type index: C{type}
        @param index: The index to query, by default atop.store.UIDIndex.

        @type endKey: A type appropriate for the specified index.
        @param endKey: If specified, the last key at which to return a result.

        @param result: If specified as C{ITEM}, return the L{Item} itself
        for each content of the pool returned.  If specified as C{INDEX},
        return the key by which each item is indexed in this pool.  If
        specified as C{STORE_ID}, return the store ID for each content of
        the pool returned.

        @raise NoSuchIndex: If the specified index does not exist for this pool.
        """

        self.startKey = startKey
        self.endKey = endKey
        self.pool = pool
        self.index = index
        try:
            self.iname = self.pool._poolIndexes[index]
        except KeyError:
            raise NoSuchIndex("Pool %r has no index %r" % (self.pool.name, self.index))
        self.result = result
        self.backwards = backwards
        self.limit = limit
        self.startItemID = startItemID
        self.endItemID = endItemID
        self.indexPfx = pool._getSuperStore()._formatIDPath(pool.getItemPath()) + SEP + 'index' + SEP + self.iname + SEP
        self.count = 1

    def __str__(self):
        return '<PoolQuery over %(pool)s/%(index)s [%(startKey)s:%(endKey)s]>' % {
            'pool': self.pool, 'index': self.index.__name__, 
            'startKey': self.startKey, 'endKey': self.endKey}

    def __repr__(self):
        return str(self)

    def resume(self):
        """Bleh.
        """
        warnings.warn("You don't have to call this any more.",
                      DeprecationWarning, stacklevel=2)

    def _nextKey(self, currentFactory):
        if self.cursorFactory is currentFactory:
            # Just keep going.
            return self.cursorMethod()
        else:
            # We're in a new transaction, but have we ever been in a
            # transaction before?
            firstTime = self.cursorMethod is None
        self.startStr = key2str(self.startKey, self.startItemID,
                                self.backwards, self.indexPfx, self.index)
        self.endStr = key2str(self.endKey, self.endItemID,
                              not self.backwards, self.indexPfx, self.index)
        self.cursorFactory = currentFactory
        cursor = currentFactory.cursor()
        rangeKey, rangeValue = cursor.set_range(self.startStr)
        if self.backwards:
            self.oper = operator.lt
            self.cursorMethod = cursor.prev

            # Checking that the range key is correct is necessary because
            # set_range places us at the first key that is equal to or greater
            # than the requested key.  Going backwards, we want to be at the
            # key which is equal to or less than the requested key.

            if rangeKey > self.startStr:
                rangeKey, rangeValue = self.cursorMethod()
        else:
            self.oper = operator.gt
            self.cursorMethod = cursor.next

        if firstTime:
            # If it's our first time, perform a sanity check and return the
            # result that we got from set_range (plus adjustment for
            # backwardsness).
            if self.oper(self.startStr, self.endStr):
                raise Inconsistency(
                    ("%squery starting at %s ending at %s will have no items"
                     ": specify them in order!.") %
                    (self.backwards and 'backwards ' or '',
                     self.startStr.split(SEP),
                     self.endStr.split(SEP)))
            return rangeKey, rangeValue
        else:
            
            # If we've been through this query before, our 'start' string is
            # really the string of the last value we retrieved.  Unless that
            # entry has been removed from the database since last we saw it, we
            # have to skip one more entry down the query before moving on.  If
            # it _has_ been removed, we should stay where we are because this
            # is a new key/value pair.
            
            if rangeKey == self.startStr:
                return self.cursorMethod()
            else:
                return rangeKey, rangeValue

    def __iter__(self):
        return self

    def stop(self):
        self.next = _stopped
        raise StopIteration()

    def next(self):
        cursorFactory = context.get('CursorFactory')
        if cursorFactory is None:
            raise Inconsistency("Cannot iterate outside of a transaction.", self)

        xvalue = self._nextKey(cursorFactory)
        if xvalue is None:
            self.stop()
        skey, svalue = xvalue

        if self.limit is not None:
            if self.count > self.limit:
                self.stop()
        self.count += 1
        if self.oper(skey, self.endStr): # I need an exit.
            self.stop()

        storeID = int(svalue)
        indexString = SEP.join(skey[len(self.indexPfx):].split(SEP)[0:-1])
        indexKeyObj = self.index.sortableStringToKey(indexString, self.pool)
        self.startKey = indexKeyObj
        self.startItemID = storeID

        if self.result == STORE_ID:
            return storeID
        elif self.result == ITEM:
            return self.pool.store.getItemByID(storeID)
        elif self.result == INDEX:
            return indexKeyObj
        elif self.result == INDEX_AND_ITEM:
            item = self.pool.store.getItemByID(storeID)
            return (indexKeyObj, item)
        elif self.result == INDEX_AND_STORE_ID:
            return (indexKeyObj, storeID)
        else:
            raise Inconsistency(
                "Invalid result-mode specified for index query: %s" % (self.result,))

def _stopped():
    raise StopIteration()

def _hSSIDCUTSIM(result, path, method, *a, **kw):
    """hook SubStore Item Deferred Callback Up To Store Item Method

    References cannot be maintained from one Store to Items in another Store
    or SubStore.  This includes such un-thought-of corners as the callbacks
    list on a Deferred referenced from an Item.  Consequently, it is not
    possible to say "storeItem.deferred.addCallback(substoreItem.method)"
    or anything similar.  This function mediates connections between such
    connections, and, given:

        itemA = SomeItem(store)
        sub = SubStore(store)
        itemB = SomeItem(sub)

    should be used something like this:

        itemB.d = d = Deferred()
        d.addCallback(_hSSIDCUTSIM, itemA.store, itemA.storeID, 'cbStuff')
        d.addErrback(_hSSIDCUTSIM, itemA.store, itemA.storeID, 'ebStuff')
    """
    def getItemByID(store, storeID):
        return store.getItemByID(storeID)
    item = reduce(getItemByID, path, context.get('store'))
    return getattr(item, method)(result, *a, **kw)

def _sim(method, *a):
    """A helper for _hSSIDCUTSIM.

    Use it like this:

        d.addCallback(*_sim(foo.cbStuff, *a), **kw)
    """
    item = method.im_self
    name = method.im_func.func_name
    return (_hSSIDCUTSIM, item.getItemPath(), name) + a

class Pool(Item):
    implements(IPool)

    persistenceForgets = ['sequence', 'ephemeralPipes']
    sequence = None

    _indexPrefixCounter = 0

    # These attributes are for use by UI.  I won't change the default sort
    # order because the semantics of queryIndex make sense as they are.

    suggestedIndex = UIDIndex
    suggestedBackwards = True
    suggestedView = 'Table'

    ephemeralPipes = ()

    ## Compatibility for TDB.  As donovan said:
    ## <fzZzy> glyph: it's just all up in the implementation detail's face
    #

    def _getoldstyleindexes(self):
        warnings.warn("Don't use '.indexes' attribute, use index-related methods.",
                      DeprecationWarning, stacklevel=2)
        x = {}
        for index in self._poolIndexes.keys():
            x[index.methodPrefix] = index
        return x

    def _getoldstylecounter(self):
        warnings.warn("Don't use '.indexCounter' attribute, use countIndex.",
                      DeprecationWarning, stacklevel=2)
        x = {}
        for index in self.iterIndexes():
            x[index.methodPrefix] = self.countIndex(index)
        return x

    indexes = property(
        _getoldstyleindexes,
        doc="""
        Don't use this property.  It's just for backwards compatibility.
        """)

    indexCounter = property(
        _getoldstylecounter,
        doc="""
        Don't use this property.  It's just for backwards compatibility.
        """)

    def __init__(self, store, name=None):
        self._poolIndexes = {}
        self.indexRemover = {}
        self.pipes = []
        self.activate()
        Item.__init__(self, store, name)
        self.addIndex(UIDIndex)

    def __contains__(self, item):
        return self in item.poolToUID

    def __str__(self):
        kname = self.__class__.__name__
        indexes = ' '.join([k.__name__ for k in self._poolIndexes.keys()])
        return '<%s %s (Indexes: %s)>' % (kname, self.name, indexes)

    def suggestIndex(self, index, backwards=False):
        index = _deprecatedIndexLookup(index)
        self.suggestedIndex = index
        self.suggestedBackwards = backwards

    def suggestView(self, view):
        self.suggestedView = view

    def addPipe(self, pipe, ephemeral=False):
        """Add a 'pipe', an object that is notified of items arriving in this
        pool, leaving it, and changing while inside it.

        The 'ephemeral' flag is intended for use by UI elements and other
        objects which do not reside in the database.  You will need to keep a
        strong reference to the pool around somewhere if you want ephemeral
        notifications.  (Normally this will not be a big deal because you've
        already got the pool open for reading, writing, or both, somewhere else
        anyway.)

        DO NOT USE THE EPHEMERAL FLAG FOR OBJECTS INSIDE THE DATABASE.  Failure
        of an ephemeral observer will _not_ abort the transaction, and will
        therefore put your database into an inconsistent state if you use it
        incorrectly.
        """
        assert doesImplement(pipe, IPipe)
        if isinstance(pipe, Item):
            pipe = pipe.referenceTo()
        if ephemeral:
            L = self.ephemeralPipes
        else:
            self.touch()
            L = self.pipes
        L.append(pipe)

    def removePipe(self, pipe, ephemeral=False):
        """See addPipe.

        Note that pipes added with the ephemeral flag _must_ be removed with
        the ephemeral flag.
        """
        if isinstance(pipe, Item):
            pipe = pipe.referenceTo()
        if ephemeral:
            L = self.ephemeralPipes
        else:
            self.touch()
            L = self.pipes
        L.remove(pipe)

    def activate(self):
        self.ephemeralPipes = []

    def deleteFromStore(self):
        # Remove from any pools so it cannot be easily accessed
        # while it is being deleted
        for p in self.iterpools():
            p.removeItem(self)
        if self.countIndex(UIDIndex):
            from atop.upgrade import emptyPool
            d = emptyPool(self)
            d.addCallback(*_sim(self._cbEmptied))
            d.addErrback(*_sim(self._ebEmptied))
            return d
        else:
            self._cbEmptied(None)
            return defer.succeed(None)

    def _cbEmptied(self, result):
        def _():
            for idx in self.iterIndexes():
                self.delDataKey(self._formatIndexCountKey(idx))
            Item.deleteFromStore(self)
            return result
        return self.store.transact(_)

    def _ebEmptied(self, failure):
        # XXX - Stick this someplace so we can try again
        log.msg("Error emptying pool %s:" % (self.getItemPath(),))
        log.err(failure)

    def _getSuperStore(self):
        ss = self.store
        while ss.store is not None:
            ss = ss.store
        return ss

    def _formatIndexCountKey(self, index, pfx=SEP + 'count' + SEP):
        try:
            v = self._poolIndexes[index]
        except KeyError:
            v = self.indexRemover[index]
        return pfx + v

    def queryIndex(self, index=UIDIndex, startKey=None, endKey=None, result=ITEM,
                   backwards=False, limit=None, startItemID=None, endItemID=None):
        """Retrieve information about the contents of this pool.

        A generator which yields the requested information about the requested Items is
        returned.  This generator I{must} be iterated in a storage context.

        @type index: C{type}
        @param index: The index to query.

        @type endKey: A type appropriate for the specified index.
        @param endKey: If specified, the last key at which to return a result.

        @param result: If specified as C{ITEM}, return the L{Item} itself
        for each content of the pool returned.  If specified as C{INDEX},
        return the key by which each item is indexed in this pool.  If
        specified as C{STORE_ID}, return the store ID for each content of
        the pool returned.

        @raise NoSuchIndex: If the specified index does not exist for this pool.
        """
        index = _deprecatedIndexLookup(index)
        return PoolQuery(self, index, startKey, endKey, result, backwards, limit, startItemID, endItemID)

    def getSequence(self):
        if self.sequence is None:
            self.sequence = list(self.queryIndex(UIDIndex,result=INDEX))
        return self.sequence

    _noIndexItem = object()

    def getIndexItem(self, index, indexKey=None, default=_noIndexItem):

        """Retrieve an item from an index.  If no indexKey is specified, I will
        return the first item in the index.

        Upon failure, (if there are no items in the index, or the specified key
        is not found), I will raise KeyError, or return the 'default' argument,
        if it is supplied.
        """
        index = _deprecatedIndexLookup(index)
        def getter():
            for k, v in self.queryIndex(index, startKey=indexKey, result=INDEX_AND_ITEM):
                # only iterates once, haha.
                if indexKey is None:
                    return v
                if index.keyToSortableString(k) != index.keyToSortableString(indexKey):
                    if default is Pool._noIndexItem:
                        raise KeyError(indexKey)
                    return default
                return v
            if default is Pool._noIndexItem:
                raise KeyError(indexKey)
            return default
        if context.get('txn') is None:
            # XXX FIXME this should NOT be auto-transacted
            warnings.warn("Calling getIndexItem() outside a transaction is deprecated!",
                          DeprecationWarning, stacklevel=2)
            return self.store.transact(getter)
        else:
            return getter()

    def nextUID(self, allocate=True):
        return self.store.nextSubID(self, allocate)

    def skipToUID(self, uid):
        """Advance the UID counter to the given UID.

        The next UID returned by nextUID will be uid + 1.

        @return: True if the UID is adjusted, False if the given UID
        is lower than the next UID that would have been allocated.
        """
        nextUID = self.nextUID(allocate=False)
        if uid < nextUID:
            return False
        elif uid > nextUID:
            self.store.setSubID(self, uid)
        return True

    def _manipItem(self, item, adding, uid=None):
        """
        Either add or remove an Item from this Pool. This method is used by
        addItem and removeItem.
        """
        if adding:
            announceMethodName = "itemAdded"
            if uid is not None:
                alreadyPresent = self.getIndexItem(UIDIndex, uid, None)
                if alreadyPresent is not None:
                    raise ValueError("Item (%r) with specified UID already exists in %r" % (alreadyPresent, self))
                else:
                    self.skipToUID(uid)
            item.addedToPool(self, uid)
        else:
            assert uid is None
            announceMethodName = "itemRemoved"

        for index in self._poolIndexes.keys():
            if adding and index is UIDIndex:
                continue
            ikeys = item.getIndexKeys(self, index)
            for ikey in ikeys:
                if adding:
                    oldKey = None
                    newKey = ikey
                else:
                    oldKey = ikey
                    newKey = None
                self.changeKeyForItem(item, index, oldKey, newKey)
        if not adding:
            item.removedFromPool(self)
        self._announceItem(item, announceMethodName)

    def changeKeyForItem(self, item, index, oldKey, newKey, notify=False):
        index = _deprecatedIndexLookup(index, 4)
        # assert not isinstance(index, MultiIndex)

        deps = [(index, oldKey, newKey)]
        allIndexes = itertools.chain(self.iterIndexes(),
                                     self.indexRemover.iterkeys())
        for idx in allIndexes:
            if isinstance(idx, MultiIndex) and index in idx.indexes:
                multiIndexes = list(idx.indexes)
                pos = multiIndexes.index(index)
                oldKeys = []
                for multiIndex in multiIndexes:
                    if multiIndex == index:
                        oldKeys.append(oldKey)
                    else:
                        oldKeys.append(item.getIndexKey(self, multiIndex))
                newKeys = oldKeys[:]
                newKeys[pos] = newKey

                if None in oldKeys:
                    oldKeys = None
                if None in newKeys:
                    newKeys = None
                deps.append((idx, oldKeys, newKeys))

        for (index, oldKey, newKey) in deps:
            self._replaceDBKey(item, index, oldKey, newKey)

    def _replaceDBKey(self, item, index, oldKey, newKey):
        sisid = '%0.10d' % item.storeID

        if index not in self._poolIndexes:
            if index in self.indexRemover:
                ipref = self.indexRemover[index]
                newKey = None
            else:
                return # raise KeyError(index)
        else:
            ipref = self._poolIndexes[index]

        if oldKey is not None:
            # delete the old key.
            istr = index.keyToSortableString(oldKey)
            try:
                self.delDataKey(SEP+'index'+SEP+ipref+SEP+istr+SEP+sisid)
            except libdb.DBNotFoundError:
                # Perhaps we were in the process of adding a new index to a
                # pool with items in it already and someone tried to change the
                # value of an item not yet processed; the expected key for the
                # new index will not yet exist, so we will not be able to
                # remove it, but this is not an error.
                indexCountDelta = 0
            else:
                indexCountDelta = -1
        else:
            indexCountDelta = 0

        if newKey is not None:
            # add the new key.
            istr = index.keyToSortableString(newKey)
            dbkey = SEP+'index'+SEP+ipref+SEP+istr+SEP+sisid
            if not self.hasDataKey(dbkey):
                indexCountDelta += 1
                self.setDataKey(dbkey, sisid)
        if indexCountDelta:
            key = self._formatIndexCountKey(index)
            self.setDataKey(key, str(int(self.getDataKey(key)) + indexCountDelta))

    def _announceItem(self, item, announceMethodName, *args):
        for pipe in self.pipes:
            getattr(pipe, announceMethodName)(item, *args)
        for epipe in self.ephemeralPipes:
            # UI elements are non-transactional anyway, so we will not honor
            # their failures.  DO NOT ADD EPHEMERAL PIPES FOR DATABASE OBJECTS.
            try:
                getattr(epipe, announceMethodName)(item, *args)
            except:
                log.err()

    def addItem(self, item, uid=None):
        if item in self:
            warnings.warn("Adding %s to %s, but it is already there." % (item, self),
                          Warning, stacklevel=2)
            return
        self._manipItem(item, True, uid)

    def removeItem(self, item):
        self._manipItem(item, False)

    def addIndex(self, index):
        """Add a new PoolIndex to this pool.  For example,

        pool.addIndex(NameIndex)

        @raise DuplicateIndex: if the specified index is already installed on
        this pool.
        """
        self.touch()
        index = _deprecatedIndexLookup(index)
        if self.hasIndex(index):
            # raise DuplicateIndex()
            warnings.warn("You shouldn't add the same index more than once.",
                          DuplicateIndex, stacklevel=2)
            return
        self._poolIndexes[index] = str(self._indexPrefixCounter)
        self.setDataKey(self._formatIndexCountKey(index), '0')
        self._indexPrefixCounter += 1

    def hasIndex(self, index):
        """Check to see if a particular index is installed on this pool.
        """
        return self._poolIndexes.has_key(index)

    def removeIndex(self, index):
        """Remove an index from this pool.

        @return: an UpgradeQueue which will remove the given index's keys.

        @raise NoSuchIndex: if the index wasn't already installed.
        """
        self.lock()
        index = _deprecatedIndexLookup(index)
        if not self.hasIndex(index):
            raise NoSuchIndex(index)
        self.indexRemover[index] = self._poolIndexes.pop(index)

        from atop.upgrade import deleteIndex
        return deleteIndex(self, index).addCallback(*_sim(self._indexRemoveDone, index))

    def _indexRemoveDone(self, result, index):
        self.delDataKey(self._formatIndexCountKey(index))
        del self.indexRemover[index]
        return result

    def countIndex(self, index):
        """
        @param index: the index whose count we are looking at.

        @return: the number of items present in me on a given index.  I will
        always return zero if I don't have the index.
        """
        index = _deprecatedIndexLookup(index)
        return int(self.getDataKey(self._formatIndexCountKey(index)) or 0)

    def iterIndexes(self):
        return self._poolIndexes.iterkeys()


backwardsCompatImplements(Pool)


class StringMap(Item):
    """Simple mapping of string keys to string values.

    Keys and values are stored directly in the database, allowing for a
    large number of them to be present without unduly increasing the size of
    this item in a serialized object graph.
    """

    def get(self, key, default=None):
        result = self.getDataKey(SEP+key)
        if result is None:
            return default
        return result

    def __getitem__(self, key):
        """StringMap[key] -> value

        If the key is not present, None is returned.
        """
        if not isinstance(key, str):
            raise TypeError("StringMap.__getitem__ requires a string, got %s" % (type(key).__name__,))
        result = StringMap.get(self, key)
        if result is None:
            raise KeyError(key)
        return result

    def __setitem__(self, key, value):
        """StringMap[key] = value
        """
        for o in (key, value):
            if not isinstance(o, str):
                raise TypeError("StringMap.__setitem__ requires strings, got %s" % (type(o).__name__,))
        self.setDataKey(SEP+key, value)

    def __delitem__(self, key):
        if not isinstance(key, str):
            raise TypeError("StringMap.__delitem__ requires a string, got %s" % (type(key).__name__,))
        self.delDataKey(SEP+key)

    def __contains__(self, key):
        return StringMap.get(self, key) is not None

    def has_key(self, key):
        return key in self

    def iterkeys(self):
        return self.queryitems(withValues=False)

    def itervalues(self):
        for k, v in self.queryitems(withValues=True):
            yield v

    def iteritems(self):
        return self.queryitems(withValues=True)

    def keys(self):
        return list(self.iterkeys())

    def values(self):
        return list(self.itervalues())

    def items(self):
        return list(self.iteritems())

    def queryitems(self, prefix='', withValues=True):
        return _StringMapIterator(self, prefix, withValues)

class _StringMapIterator:
    startKey = ''
    cursorMethod = None
    cursorFactory = None

    def __init__(self, map, prefix, withValues=True):
        self.withValues = withValues
        self.map = map
        self.prefix = prefix
        dbpfx = map.store._formatIDPath(map.getItemPath())
        self.allKeysBeginWith = dbpfx + SEP + prefix
        self.lastItem = (self.allKeysBeginWith, '')

    def _nextKey(self):
        firstTime = self.cursorMethod is None
        currentFactory = context.get("CursorFactory")
        if self.cursorFactory is not currentFactory:
            self.cursorFactory = currentFactory
            cursor = currentFactory.cursor()
            self.cursorMethod = cursor.next
            lastKey, lastValue = self.lastItem
            val = newKey, newValue = cursor.set_range(lastKey)
            if newKey == lastKey and not firstTime:
                val = self.cursorMethod()
        else:
            val = self.cursorMethod()
        self.lastItem = val
        return val

    def next(self):
        k, v = self._nextKey()
        if not k.startswith(self.allKeysBeginWith):
            raise StopIteration()
        key = k[len(self.allKeysBeginWith):]
        if self.withValues:
            return key, v
        else:
            return key

    def __iter__(self):
        return self


class TupleMap(StringMap):

    def __contains__(self, key):
        return StringMap.__contains__(self, key)

    def get(self, key, default=None):
        return StringMap.get(self, self._formatTupleKey(key), default)

    def __getitem__(self, key):
        return StringMap.__getitem__(self, self._formatTupleKey(key))

    def __setitem__(self, key, value):
        return StringMap.__setitem__(self, self._formatTupleKey(key), value)

    def __delitem__(self, key):
        return StringMap.__delitem__(self, self._formatTupleKey(key))

    def _formatTupleKey(self, tupleKey):
        if isinstance(tupleKey, basestring):
            # Allow the syntax foo['a'], as it is logically similar to foo['a', 'b']
            tupleKey = (tupleKey,)
        assert not [x for x in tupleKey if SEP in x], \
            "Don't put SEP in tupleKey (or contribute a patch)"
        stringKey = SEP.join(tupleKey)
        return stringKey

    def queryitems(self, prefix=(), withValues=True):
        # prefix=() is special: it really means 'no prefix'
        if prefix:
            sprefix = self._formatTupleKey(prefix)+SEP
        else:
            sprefix = ''
        for x in _StringMapIterator(self, sprefix, withValues):
            if withValues:
                yield tuple(x[0].split(SEP)), x[1]
            else:
                yield tuple(x.split(SEP))

class IPersistable(Interface):

    def touch(self):
        """Notify that this thingy is dirty and in need of persistament.

        This must be called prior to modifying any persistent state of an item
        within a transaction.  It is needed so that atop knows what items must
        be saved to commit the transaction, or reverted if the transaction
        fails.
        """

    def referenceTo(self):
        "produce a sturdy ref to this object"

class Persistability(Item, Adapter):
    implements(IPersistable)
    def __init__(self, original, store=None):
        Item.__init__(self, store or context.get('store'))
        Adapter.__init__(self, original)

backwardsCompatImplements(Persistability)

class StoredAdapter(object, Adapter):
    __metaclass__ = MetaItem

    def __init__(self, original):
        object.__init__(self)
        Adapter.__init__(self, original)
