# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

from twisted.cred.portal import Portal, IRealm
from twisted.cred.checkers import ICredentialsChecker
from twisted.cred.credentials import UsernamePassword
from twisted.python.components import Interface, Adapter, registerAdapter

from atop.test import unittest
from atop.credup import StoreRealm, Avatar, Preauthenticated, DomainIndex, NoSuchDomain

from atop.faraway import IFarIndex
from atop.ifcompat import implements, backwardsCompatImplements

class IUnadaptable(Interface):
    pass

class IAvatarTest(Interface):
    """
    Placeholder interface for tests.
    """

class AdapterTest(Adapter):
    implements(IAvatarTest)

    value = None

    def dummyMethod(self, value):
        self.original.touch()
        self.value = value

backwardsCompatImplements(AdapterTest)

registerAdapter(AdapterTest, Avatar, IAvatarTest)

USERNAME = "userXname"
PASSWORD = "passXword"

def passthrough(a, i):
    pass
def refuseAll(a, i):
    raise Exception("All logins disabled!!!")

class LoginExampleAccess(unittest.TestCase):

    storePowerups = [StoreRealm]

    def dbSetUp(self):
        icc = ICredentialsChecker(self.store)
        icc.addDomain("divmod.org")
        self.avatar = self.createAvatar((), name=USERNAME)
        self.avatar.setInsecurePassword(PASSWORD)
        icc.addAvatarToDomain(self.avatar, "divmod.org")

    def testDefaultLogin(self):
        def portal():
            p = Portal(IRealm(self.store))
            p.registerChecker(ICredentialsChecker(self.store))
            return p
        p = self.store.transact(portal)
        r = unittest.deferredResult(p.login(UsernamePassword(
            USERNAME, PASSWORD), None, IAvatarTest))
        self.assertEquals(len(r), 3)
        self.assertEquals(r[0], IAvatarTest)
        self.assertEquals(r[1], IAvatarTest(self.avatar))
        # Who cares about r[2]?

    def testMultipleInterfaces(self):
        r = self.store.transact(IRealm, self.store)
        ial = r.requestAvatar(self.avatar.storeID, None, IUnadaptable, IAvatarTest)
        self.assertEquals(ial[0], IAvatarTest)
        self.failUnless(isinstance(ial[1], AdapterTest))
        self.assertEquals(ial[1].original, self.avatar)

        # Just make sure it doesn't raise
        ial[2]()

    def dbtestCaseInsensitiveUsernames(self):
        cc = ICredentialsChecker(self.store)
        p = Portal(IRealm(self.store))
        p.registerChecker(cc)

        for f in (str.lower, str.upper, str.title):
            self.assertIdentical(self.avatar, cc.getAvatar(f(USERNAME)))

            d = p.login(Preauthenticated(f(USERNAME)), None, IAvatarTest)
            i, a, l = unittest.deferredResult(d)
            self.assertIdentical(i, IAvatarTest)
            self.assertIdentical(a.original, self.avatar)

    def dbtestDuplicateDomain(self):
        icc = ICredentialsChecker(self.store)
        icc.addDomain("divmod.org")
        domains = icc.domains
        self.assertEquals(
            len(list(domains.queryIndex(DomainIndex, startKey='divmod.org',
                                        endKey='divmod.org'))),
            1)

    def dbtestFarAvatar(self):
        from twisted.internet import reactor
        iffy = IFarIndex(self.avatar).farByName(USERNAME)
        # next line is written in this way to make sure that if I relax the
        # reactor-must-iterate requirement it will still work -gl
        iffy.far(IAvatarTest, "dummyMethod", 7).addBoth(lambda q: (reactor.callLater(0, reactor.crash), q)[-1])
        reactor.run()
        self.assertEquals(IAvatarTest(self.avatar).value, 7)

class DomainHandling(unittest.TestCase):
    def testDomainIndex(self):
        k2ss = DomainIndex.keyToSortableString
        ss2k = DomainIndex.keyToSortableString
        self.assertEquals(k2ss('foo.bar.baz'), 'baz.bar.foo')
        self.assertEquals(ss2k('www.divmod.com'), 'com.divmod.www')

        for d in ('com', 'example.com', 'www.example.com', 'pop.mail.example.com'):
            self.assertEquals(ss2k(k2ss(d)), d)

    def dbtestGetDomain(self):
        dList = []
        icc = ICredentialsChecker(self.store)
        for pfx in ["", "www.", "ns1.", "exarkun."]:
            for mid in ["divmod", "notdivmod"]:
                for tld in [".com", ".org", ".net"]:
                    dList.append((tld, ".example", mid, pfx))
                    name = pfx + mid + ".example" + tld
                    icc.addDomain(name)
                    self.assertEquals(icc.getDomain(name).name, name)

        self.assertRaises(
            NoSuchDomain,
            icc.getDomain, "example.com")

        dList.sort()

        self.assertEquals(
            [d.name for d in icc.getDomains()],

            [''.join(e[::-1]) for e in dList])
