
from twisted.trial import unittest

from atop.notify import NotifyList, NotifyDict, FilteringDict

def makeNotifyTest(self, nodups=False, storeValue=None, loadValue=None):
    data = []
    adds = []
    removes = []
    nl = NotifyList(data, adds.append, removes.append, nodups, storeValue, loadValue)
    if storeValue is None:
        self.data = data
    else:
        self.data = nl
    self.adds = adds
    self.removes = removes
    self.nl = nl

class both:
    def __init__(self, one, two):
        self.one = one
        self.two = two

    def __getattr__(self, attr):
        return both(getattr(self.one, attr),
                    getattr(self.two, attr))

    def __call__(self,*a,**kw):
        r1 = self.one(*a,**kw)
        r2 = self.two(*a,**kw)
        assert r1 == r2
        return r1

class BaseTestsMixin:

    #### Test every notifying method:

    def testAppend(self):
        self.nl.append(1)
        self.assertEquals(self.adds, [1])
        self.assertEquals(self.data, [1])
        self.assertEquals(self.removes, [])

    def testInsert(self):
        data = []
        b = both(self.nl, data)
        b[:] = range(5)
        b.insert(1, 9)
        self.assertEquals(data, self.nl)
        self.assertEquals(self.adds, range(5) + [9])
        self.assertEquals(self.data, data)
        self.assertEquals(self.removes, [])

    def testPop(self):
        self.nl.append(1)
        self.nl.pop(0)
        self.assertEquals(self.adds, [1])
        self.assertEquals(self.data, [])
        self.assertEquals(self.removes, [1])

    def testRemove(self):
        self.nl.append(1)
        self.nl.remove(1)
        self.assertEquals(self.adds, [1])
        self.assertEquals(self.data, [])
        self.assertEquals(self.removes, [1])

    def testExtend(self):
        self.nl.extend(range(10))
        self.assertEquals(self.adds, range(10))
        self.assertEquals(self.data, range(10))

    def testIMul(self):
        self.nl[:] = range(10)
        self.nl *= 5
        self.assertEquals(self.data, range(10) * 5)
        self.assertEquals(self.removes, [])

    def testIAdd(self):
        self.nl += range(10)
        self.assertEquals(self.adds, range(10))
        self.assertEquals(self.data, range(10))
        self.assertEquals(self.removes, [])

    def testSetItem(self):
        self.nl += range(10)
        self.nl[0] = 10
        self.assertEquals(self.adds, range(11))
        self.assertEquals(self.data, [10] + range(1,10))
        self.assertEquals(self.removes, [0])

    def testDelItem(self):
        self.nl += range(10)
        del self.nl[0]
        self.assertEquals(self.data, range(1,10))
        self.assertEquals(self.adds, range(10))
        self.assertEquals(self.removes, [0])

    def testDelSlice(self):
        self.nl.extend(range(10))
        del self.nl[0:5]
        self.assertEquals(self.adds, range(10))
        self.assertEquals(self.removes, range(5))
        self.assertEquals(self.data, range(5, 10))

    def testSetSlice(self):
        self.nl[:] = range(10)
        self.assertEquals(self.adds, range(10))
        self.assertEquals(self.data, range(10))
        self.nl[5:7] = range(15,20)
        L = range(10)
        L[5:7] = range(15, 20)
        self.assertEquals(self.data, L)
        self.assertEquals(self.removes, range(10)[5:7])
        self.assertEquals(self.adds, range(10) + range(15, 20))

    ##### Done, now test nodups behavior:

    def testAddBatchExtend(self):
        self.nl.extend(range(10) * 2)
        self.addBatchAsserts()

    def testAddBatchSetSlice(self):
        self.nl[:] = range(10)
        self.nl[:] = range(10)
        self.addBatchAsserts()
        self.justRemovedDups()

    def testAddBatchAppend(self):
        for x in range(10) * 2:
            self.nl.append(x)
        self.addBatchAsserts()
        del self.nl[0:10]
        self.justRemovedDups()

class NotifyListTest(unittest.TestCase, BaseTestsMixin):
    
    def setUp(self):
        makeNotifyTest(self, False)

    def addBatchAsserts(self):
        self.assertEquals(self.adds, range(10) * 2)

    def justRemovedDups(self):
        self.assertEquals(self.removes, range(10))

class NotifyListTestNodups(unittest.TestCase, BaseTestsMixin):

    def setUp(self):
        makeNotifyTest(self, True)

    def addBatchAsserts(self):
        self.assertEquals(self.adds, range(10))

    def justRemovedDups(self):
        self.assertEquals(self.removes, [])


class NotifyListLoadStore(NotifyListTest):
    def setUp(self):
        makeNotifyTest(self, False, str, int)

class NotifyListLoadStoreNodups(NotifyListTestNodups):
    def setUp(self):
        makeNotifyTest(self, True, str, int)

def starify(f):
    return lambda *a: f(a)
class NotifyDictTest(unittest.TestCase):

    def setUp(self):
        self.data = {}
        self.adds = []
        self.removes = []
        self.nd = NotifyDict(self.data,
                             starify(self.adds.append),
                             starify(self.removes.append))

    def testSetItem(self):
        self.nd[1] = 2
        self.assertEquals(self.nd[1], 2)
        self.assertEquals(self.adds, [(1,2)])
        self.nd[1] = 3
        self.assertEquals(self.adds, [(1,2), (1,3)])
        self.assertEquals(self.removes, [(1,2)])

    def testUpdate(self):
        self.nd.update({1: 2,
                        3: 4})
        self.assertEquals(len(self.adds), 2)
        for kvp in (1, 2), (3, 4):
            self.assertIn(kvp, self.adds)

    def testDelItem(self):
        self.nd[1] = 2
        del self.nd[1]
        self.failIf(self.nd)
        self.assertEquals(self.adds, self.removes)
        self.assertEquals(self.removes, [(1,2)])

    def testClear(self):
        d = dict.fromkeys(range(10), 7)
        j = d.items()
        j.sort()
        self.nd.update(d)
        self.nd.clear()
        self.adds.sort()
        self.removes.sort()
        self.assertEquals(self.adds, self.removes)
        self.assertEquals(self.removes, j)
        
    def testSetDefault(self):
        self.nd.setdefault(1, 2)
        self.assertEquals(self.adds, [(1,2)])
        self.assertEquals(self.removes, [])
        self.nd.setdefault(1, 3)
        self.assertEquals(self.adds, [(1,2)])
        self.assertEquals(self.removes, [])

    def testPop(self):
        self.assertRaises(KeyError, self.nd.pop, 1)
        self.nd[1] = 2
        self.assertEquals(self.nd.pop(1), 2)
        self.assertEquals(self.removes, [(1,2)])

    def testPopItem(self):
        self.nd.update(dict.fromkeys(range(10),7))
        while True:
            try:
                self.nd.popitem()
            except KeyError:
                break
        self.adds.sort()
        self.removes.sort()
        self.assertEquals(self.adds, self.removes)

def add1(x):
    return x + 1

def sub1(x):
    return x - 1

def add2(x):
    return x + 2

def sub2(x):
    return x - 2



class FilteringDictTestCase(unittest.TestCase):
    def testFiltering(self):
        d = {}
        fd = FilteringDict(d,
                           add1, sub1,
                           add2, sub2)
        ud = {1: 2,
              3: 4}
        fd.update(ud)
        self.assertEquals(fd, ud)
        self.assertEquals(d, {0: 0,
                              2: 2})
        self.assertRaises(KeyError, fd.pop, 0)
        self.assertEquals(fd.pop(0, 9), 9)
        self.assertEquals(fd.pop(1), 2)
        self.assertEquals(fd.popitem(), (3, 4))
        self.assertRaises(KeyError, fd.popitem)
