# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

from twisted.python import components

from atop.powerup import Powerup, IPowerStation, Storeup
from atop.store import Store, SubStore

from atop.test import unittest

from os.path import join as opj

class FooUp(Powerup):
    powerupName = 'Foo'
    called = 0
    def setUpPools(self, substore):
        FooUp.called += 1

class BarUp(Powerup):
    powerupName = 'Bar'
    called = 0
    def setUpPools(self, substore):
        BarUp.called += 1

class QuxUp(Powerup):
    powerupName = 'Qux'
    dependencies = [FooUp, BarUp]
    called = 0
    def setUpPools(self, substore):
        QuxUp.called += 1

class PowerupTest(unittest.TestCase):

    def dbSetUp(self):
        self.substore = SubStore(self.store)

    def assertify(self):
        for v in [__name__ + '.FooUp',
                  __name__ + '.BarUp',
                  __name__ + '.QuxUp']:
            self.failUnlessIn(v, IPowerStation(self.substore).powerups.keys())
        self.assertEquals(FooUp.called, 1)
        self.assertEquals(BarUp.called, 1)
        self.assertEquals(QuxUp.called, 1)

    def partDependencies1(self):
        self.assertEquals( IPowerStation(self.substore).powerups , {} )
        IPowerStation(self.substore).installPowerup(FooUp)

        self.failUnlessIn(__name__ + '.FooUp', IPowerStation(self.substore).powerups)

        IPowerStation(self.substore).installPowerup(QuxUp)
        self.assertify()

    def partDependencies2(self):
        self.assertify()

    multitxnDependencyTest = [partDependencies1, partDependencies2]

class QuuxUp(Storeup):
    powerupName = "QuuxUp"
    mangled = 0

from twisted.trial import unittest as trialtest

class StoreupTest(trialtest.TestCase):
    def testInstallAndFind(self):
        p = self.mktemp()
        openit = lambda : Store(opj(p,'db'), opj(p,'files'))
        s = openit()
        def tx1():
            qu1 = IPowerStation(s).installPowerup(QuuxUp)
            qu1.mangled = 1
        s.transact(tx1)
        s.close()
        s = openit()
        def tx2():
            qu2 = IPowerStation(s).installPowerup(QuuxUp)
            self.failUnless(qu2.mangled)
        s.transact(tx2)
        s.close()

class TestUpMixin:
    ghettoSuper = None

    _micounter = 0

    def makeInterface(cls):
        if 'ISup' not in vars(cls):
            cls._micounter += 1
            exec "class I%s%s(components.Interface): pass" % (cls._micounter,
                                                              cls.__name__) in globals()
            cls.ISup = globals()['I' + str(cls._micounter) + cls.__name__]
        return cls.ISup
    makeInterface = classmethod(makeInterface)

    def setUpPools(self, store):
        self.store.setComponent(self.makeInterface(), self)

    def deleteFromStore(self):
        self.store.unsetComponent(self.makeInterface())
        return self.ghettoSuper.deleteFromStore(self)

class TestSup(TestUpMixin, Storeup):
    ghettoSuper = Storeup
    powerupName = None

class TestPup(TestUpMixin, Powerup):
    ghettoSuper = Powerup
    powerupName = None

class TestSupA(TestSup):
    powerupName = 'TestSupA'

class TestSupB(TestSup):
    dependencies = [TestSupA]
    powerupName = 'TestSupB'

class TestSupC(TestSup):
    dependencies = ['atop.test.test_powerup.TestSupB', TestSupA]
    powerupName = 'TestSupC'

class TestSupD(TestSup):
    dependencies = [TestSupC]
    powerupName = 'TestSupD'

class TestSupE(TestSup):
    powerupName = 'TestSupE'

class TestPupA(TestPup):
    powerupName = 'TestPupA'

class TestPupB(TestPup):
    dependencies = [TestPupA]
    storeDependencies = [TestSupB]
    powerupName = 'TestPupB'

class TestPupC(TestPup):
    storeDependencies = [TestSupE]
    powerupName = 'TestPupC'

class TestPupD(TestPup):
    dependencies = [TestPupC, TestPupB]
    powerupName = 'TestPupD'

class Dependencies(unittest.TestCase):
    
    
    def dbtestStoreupInstallDependencies(self):
        IPowerStation(self.store).installPowerup(TestSupD)
        for cls in (TestSupA, TestSupB, TestSupC, TestSupD):
            self.failUnless(isinstance(cls.ISup(self.store), cls),
                            "%r isn't a %r" % (cls.ISup(self.store), cls))

    def dbtestPowerupInstallDependencies(self):
        IPowerStation(self.store).installPowerup(TestPupD)
        for cls in (TestPupA, TestPupB, TestPupC, TestPupD):
            self.failUnless(isinstance(cls.ISup(self.store), cls),
                            "%r isn't a %r" % (cls.ISup(self.store), cls))

    def dbtestPowerupInstallStoreDependencies(self):
        av = self.createAvatar([TestPupD])
        for cls in (TestSupA, TestSupB, TestSupE):
            self.failUnless(isinstance(cls.ISup(self.store), cls),
                            "%r isn't a %r" % (cls.ISup(self.store), cls))

    def dbtestStoreupUninstall(self):
        av = self.createAvatar([TestPupD])
        supB = TestSupB.ISup(self.store)
        unittest.deferredResult(supB.deleteFromStore())
        for cls in (TestSupB, TestSupC, TestSupD):
            try:
                cls.makeInterface()(self.store)
            except components.CannotAdapt:
                pass
            else:
                raise unittest.FailTest("Adapted to %s.ISup!" % (cls.__name__,))
        self.failUnless(isinstance(TestSupE.makeInterface()(self.store), TestSupE))

        for cls in (TestPupB, TestPupD):
            try:
                cls.makeInterface()(av)
            except components.CannotAdapt:
                pass
            else:
                raise unittest.FailTest("Adapted to %s.ISup!" % (cls.__name__,))
        self.failUnless(isinstance(TestPupA.makeInterface()(av), TestPupA))
    dbtestStoreupUninstall.skip = "Fails because Storeups can't be deleted"

    def dbtestPowerupUninstall(self):
        av = self.createAvatar([TestPupD])
        pupA = TestPupA.ISup(av)
        unittest.deferredResult(pupA.deleteFromStore())

        for cls in (TestPupA, TestPupB, TestPupD):
            try:
                cls.makeInterface()(av)
            except components.CannotAdapt:
                pass
            else:
                raise unittest.FailTest("Adapted to %s.ISup!" % (cls.__name__,))
        self.failUnless(isinstance(TestPupC.makeInterface()(av), TestPupC))
