# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

from twisted.application import service

from atop.test import unittest
from atop.sched import IScheduler, SystemScheduler, SubScheduler, SchedulerBase, Schedulable, CallingItem
from atop.powerup import IPowerStation
from atop.store import SubStore, Item

class MySch(Item, Schedulable):
    count = 0
    def run(self):
        self.count += 1

class SchedulerStub:
    def __init__(self, times, calls):
        self.times = times
        self.calls = calls

    def clock(self):
        return self.times[0]

    def callLater(self, n, f, *a, **kw):
        self.calls.append((n, f, a, kw))

class SchedTestMixin:
    storePowerups = [SystemScheduler]

    def dbSetUp(self):
        self.times = self.times[:]
        self.delayedCalls = []
        self.stub = SchedulerStub(self.times, self.delayedCalls)
        self.sched = IScheduler(self.store)
        self.sched.clock = self.stub.clock
        self.sched.callLater = self.stub.callLater
        self.sched.startService()

    def dbTearDown(self):
        del self.times
        return self.sched.stopService()

    def tick(self):
        n, f, a, kw = self.delayedCalls.pop(0)
        f(*a, **kw)

class SchedTest(SchedTestMixin, unittest.TestCase):
    times = [0.0, 1.0, 2.0, 3.0, 10.0, 20.0, 30.0]

    def dbtestTopScheduler(self):
        msc = MySch(self.store)
        msc.schedule(2)

        for x in range(10):
            self.tick()
        self.assertEquals(msc.count, 0)

        for x in range(3):
            self.times.pop(0)
            self.tick()
        self.assertEquals(msc.count, 1)
        self.assertEquals(self.delayedCalls, [])

    def dbtestSubScheduler(self):
        sub = SubStore(self.store)
        IPowerStation(sub).installPowerup(SubScheduler)

        msc = MySch(sub)
        msc.schedule(2.5)

        for x in range(10):
            self.tick()
        self.assertEquals(msc.count, 0)

        for x in range(3):
            self.times.pop(0)
            self.tick()
        self.assertEquals(msc.count, 1)
        self.assertEquals(self.delayedCalls, [])

        self.failIf(
            list(IScheduler(self.store).pool.queryIndex()),
            "Items left in the scheduled pool")

def testCallingItem(xRef, y):
    x = xRef.getItem()
    x.touch()
    x.y = y

class CallingItemTestCase(SchedTestMixin, unittest.TestCase):
    times = [0, 1, 2, 3, 4, 5]

    def dbtestCalling(self):
        x = Item(self.store)
        c = CallingItem(self.store, 3, testCallingItem, x.referenceTo(), 'foo')
        storeID = c.storeID
        for i in range(4):
            self.times.pop(0)
            self.tick()
        self.assertEquals(x.y, 'foo')
        self.assertRaises(KeyError, self.store.getItemByID, storeID)
