# -*- test-case-name: atop.test.test_store -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import gc, os, time
from pprint import pformat
from os.path import join as opj

from atop.ifcompat import implements, backwardsCompatImplements

from atop.store import Store, Item, SubStore, FinalizingCache, StrongCache
from atop.store import Pool, DateIndex, UIDIndex, StringIndex
from atop.store import ITEM, INDEX, STORE_ID, INDEX_AND_ITEM, INDEX_AND_STORE_ID
from atop.store import Inconsistency, NameIndex, StringMap, TupleMap
from atop.store import referenced, StoredAdapter, Persistability, IPersistable
from atop.store import DatabaseAlreadyOpen, NoSuchIndex
from atop.store import indexed, indexedList, indexedPoolDict, indexedPoolDictOfLists

from atop.sched import SystemScheduler, SubScheduler, IScheduler
from atop.powerup import IPowerStation
from atop.util import DSUPool
from atop.test import unittest
from atop import fwd

from twisted.python.components import Interface, Componentized
from twisted.python.reflect import objgrep, isSame
from twisted.python import log
from twisted.trial import unittest as trialtest

class OtherNameIndex(NameIndex):
    methodPrefix = "name_other"

class ExampleException(Exception):
    pass

class ExampleAvatar(SubStore):

    def __init__(self, store, frobnitzes=-99):
        self.frobnitzes = frobnitzes
        SubStore.__init__(self, store)

    def addFrobnitzThenBreak(self):
        self.touch()
        self.frobnitzes += 1
        raise ExampleException('whoops sorry')

    def noReallyAddFrobnitz(self):
        self.frobnitzes += 1
        return 'YES TO HAPPINESS'

class ExampleSubitem(Item):
    def __init__(self, store):
        Item.__init__(self, store)

class StoreMixin:
    pathToDB = None
    def setUp(self):
        if self.pathToDB is None:
            self.pathToDB = self.mktemp()
        self.store = Store(opj(self.pathToDB, "db"), opj(self.pathToDB, "files"))
        # THREE shall be the number of the count
        self.avatar = self.store.transact(ExampleAvatar, self.store, 3)

    def tearDown(self):
        self.store.close()
        try:
            del self.pathToDB
        except AttributeError:
            pass

class CallOnChangeItem(Item):
    called = False

    def changed(self, value):
        self.called = value

    fu = indexed(StringIndex, changed)

class SingleOpenTestCase(unittest.TestCase):
    def testCanOnlyOpenADatabaseOnceWithoutJoining(self):
        dbdir = self.mktemp()
        open = lambda: Store(opj(dbdir, 'db'), opj(dbdir, 'files'))

        for i in range(10):
            s = open()
            self.assertRaises(DatabaseAlreadyOpen, open)
            self.assertRaises(DatabaseAlreadyOpen, open)
            self.assertRaises(DatabaseAlreadyOpen, open)
            s.close()

class CachingTestCase(unittest.TestCase):
    def dbSetUp(self):
        self.avatar = ExampleAvatar(self.store, 3)

    def testCreationCachesObjects(self):
        i = self.store.transact(Item, self.avatar)
        self.failUnless(self.store.cache.has(i.getItemPath()))

    def dbtestCachedBeforeTransactionCommit(self):
        i = Item(self.avatar)
        self.failUnless(self.store.cache.has(i.getItemPath()))

    def dbtestTransactUncallable(self):
        self.assertRaises(TypeError, self.store.transact, None, 1, 2, "3")

    def dbtestSubTransactionUncallable(self):
        def subItem():
            return Item(self.avatar)
        def subException():
            raise TypeError()
        def subTrans():
            i1 = self.avatar.transact(subItem)
            self.avatar.transact(subException)
        self.assertRaises(TypeError, self.avatar.transact, subTrans)

    def testAbortedTransactionClearsCache(self):
        def _():
            self.item = Item(self.avatar)
            self.itemPath = self.item.getItemPath()
            raise ExampleException("Failure")
        self.failUnlessRaises(ExampleException, self.store.transact, _)
        self.failIf(self.store.cache.has(self.itemPath))

    def testSynchronousReferences(self):
        def _():
            self.item = Item(self.avatar)
            self.otheritem = Item(self.avatar)
            self.weirdItem = Item(self.store)
            self.ref = self.item.referenceTo()
            self.otherref = self.otheritem.referenceTo()
            self.weirdref = self.weirdItem.referenceTo()
        self.store.transact(_)
        def _2():
            self.item.touch()
            self.item.ref = self.otheritem.referenceTo()
        self.store.transact(_2)
        del self.item
        del self.otheritem
        gc.collect()
        def _3():
            self.assertIdentical(self.ref.getItem().ref.getItem(),
                                 self.otherref.getItem())
        self.store.transact(_3)
        def _4():
            i = self.ref.getItem()
            i.touch()
            i.brokeref = self.weirdItem.referenceTo()
        self.assertRaises(Inconsistency, self.store.transact, _4)
        def _5():
            i = self.ref.getItem()
            self.avatar.touch()
            self.avatar.goodref = i.referenceTo()
            self.avatar.topref = self.weirdItem.referenceTo()
        self.store.transact(_5)
        del self.weirdItem
        def _6():
            i = self.ref.getItem()
            w = self.weirdref.getItem()
            self.assertIdentical(self.avatar.goodref.getItem(), i)
            self.assertIdentical(self.avatar.topref.getItem(), w)
        self.store.transact(_6)

class UIExampleAccess(unittest.TestCase):
    def dbSetUp(self):
        self.avatar = ExampleAvatar(self.store, 3)

    def testSomethingInteresting(self):
        self.assertRaises(ExampleException, self.avatar.transact, self.avatar.addFrobnitzThenBreak)
        self.assertEquals(self.avatar.frobnitzes, 3) # and the number of the counting shall be THREE
        self.avatar.transact(self.avatar.noReallyAddFrobnitz)
        self.assertEquals(self.avatar.frobnitzes, 4)

    def testSubitemStorage(self):
        subitem = self.avatar.transact(ExampleSubitem, self.avatar)
        self.assertIdentical(self.avatar.getItemByID(subitem.storeID),
                             subitem)
        try:
            nosub = self.store.getItemByID(subitem.storeID)
        except KeyError:
            return
        else:
            # trapping the exception to avoid potential re-numbering problems
            # and potential number-conflicts at multiple namespace levels, but
            # if we got something back from the storage there, it had better
            # not be the same item we put into sub-storage!
            self.assertNotEquals(nosub, subitem)

class XI(NameIndex):
    pass
class YI(NameIndex):
    pass
class ZI(NameIndex):
    pass

class IndexA(NameIndex):
    methodPrefix = "a"
class IndexB(NameIndex):
    methodPrefix = "b"

class ItemAB(Item):
    a = indexed(IndexA)
    b = indexed(IndexB)
    def __init__(self, store, a=None, b=None):
        Item.__init__(self, store)
        self.a = a
        self.b = b

class PoolTestCase(unittest.TestCase):
    def dbSetUp(self):
        self.avatar = ExampleAvatar(self.store, 3)
        self.p = Pool(self.store)
        self.items = []

    storePowerups = [SystemScheduler]

    nAdds = 300
    nRuns = 5
    nItems = nAdds * nRuns

    def testInconsistencyRaised(self):
        self.assertRaises(Inconsistency, list, self.p.queryIndex())

    def testStringification(self):
        self.failUnless(isinstance(IndexA.toString(), str))
        self.assertIdentical(NameIndex.fromString(IndexA.toString()), IndexA)
        self.assertRaises(KeyError, IndexA.fromString, "nonexistent.index")
        o = object()
        self.assertIdentical(NameIndex.fromString("nonexistent.index", o), o)

    def queryIndexAddItems(self):
        # Put some stuff into the pool
        for i in range(self.nAdds):
            i = Item(self.store)
            self.p.addItem(i)
            self.items.append(i)

    def queryIndexAssertState(self):
        query = list(self.p.queryIndex(result=ITEM))
        self.assertEquals(len(query), self.nItems)
        for (orig, fromPool) in zip(self.items, query):
            self.assertIdentical(orig, fromPool)

        query = list(self.p.queryIndex(result=INDEX))
        self.assertEquals(len(query), self.nItems)

        for (ans, fromPool) in zip(range(self.nItems), query):
            self.assertEquals(ans, fromPool)

        query = list(self.p.queryIndex(result=STORE_ID))
        self.assertEquals(len(query), self.nItems)
        for (orig, fromPool) in zip(self.items, query):
            self.assertEquals(orig.storeID, fromPool)

        query = list(self.p.queryIndex(result=INDEX_AND_ITEM))
        self.assertEquals(len(query), self.nItems)
        for (orig, (index, item)) in zip(self.items, query):
            self.assertIdentical(self.p.getIndexItem(UIDIndex, index), orig)
            self.assertIdentical(orig, item)

        query = list(self.p.queryIndex(result=INDEX_AND_STORE_ID))
        self.assertEquals(len(query), self.nItems)
        for (orig, (index, storeID)) in zip(self.items, query):
            self.assertIdentical(self.p.getIndexItem(UIDIndex, index), orig)
            self.assertEquals(orig.storeID, storeID)

        query = list(self.p.queryIndex(limit=10))
        self.assertEquals(len(query), 10)
        itmz = self.items[:10]
        for (orig, fromPool) in zip(itmz, query):
            self.assertEquals(orig, fromPool)
        query = list(self.p.queryIndex(backwards=True))
        itmz = self.items[:]
        itmz.reverse()

        self.assertEquals(len(query), len(itmz))
        for (orig, fromPool) in zip(itmz, query):
            self.assertEquals(orig, fromPool)

        sk = self.nAdds // 5
        ek = self.nAdds // 2
        query = list(self.p.queryIndex(startKey=sk, endKey=ek, result=INDEX))
        self.assertEquals(query[0],  sk)
        self.assertEquals(query[-1], ek)
        self.assertRaises(Inconsistency, lambda : list(self.p.queryIndex(startKey=sk, endKey=ek, result=INDEX, backwards=True)))
        query = list(self.p.queryIndex(startKey=ek, endKey=sk, result=INDEX, backwards=True))
        self.assertEquals(query[0],   ek)
        self.assertEquals(query[-1],  sk)

    multitxnQueryIndex = [queryIndexAddItems] * nRuns + [queryIndexAssertState]

    def switchToDSU(self):
        self.origp = self.p
        self.p = self.store.transact(lambda : DSUPool(list(self.origp.queryIndex()), {}))
    multitxnDSUQueryIndex = [queryIndexAddItems] * nRuns + [switchToDSU, queryIndexAssertState]

    def testStartKey(self):
        nadds = 300
        startKey = 100
        items = []
        def _1():
            for i in range(nadds):
                i = self.store.transact(Item, self.store)
                self.p.addItem(i)
                items.append(i)

        def _2():
            query = list(self.p.queryIndex(startKey=startKey, result=INDEX))
            self.assertEquals(len(query), nadds - startKey)
            self.assertEquals(query, range(100, 300))

        self.store.transact(_1)
        self.store.transact(_2)

    def testDSUStartKey(self):
        self.switchToDSU()
        self.testStartKey()

    def testEndKey(self):
        nadds = 300
        endKey = 200
        items = []
        def _1():
            for i in range(nadds):
                i = self.store.transact(Item, self.store)
                self.p.addItem(i)
                items.append(i)

        def _2():
            query = list(self.p.queryIndex(endKey=endKey, result=INDEX))
            self.assertEquals(len(query), endKey + 1)
            self.assertEquals(query, range(endKey + 1))

        self.store.transact(_1)
        self.store.transact(_2)

    def testDSUEndKey(self):
        self.switchToDSU()
        self.testEndKey()

    def testBothKeys(self):
        nadds = 300
        startKey = 75
        endKey = 225
        items = []
        def _1():
            for i in range(nadds):
                i = self.store.transact(Item, self.store)
                self.p.addItem(i)
                items.append(i)

        def _2():
            query = list(self.p.queryIndex(startKey=startKey, endKey=endKey, result=INDEX))
            self.assertEquals(len(query), endKey - startKey + 1)
            self.assertEquals(query, range(startKey, endKey + 1))

        self.store.transact(_1)
        self.store.transact(_2)

    def testDSUBothKeys(self):
        self.switchToDSU()
        self.testEndKey()

    def dbtestIndexMutation(self):
        i = Item(self.store, name='Fred')
        self.p.addIndex(NameIndex)
        p2 = Pool(self.store)
        p2.addItem(i)
        self.p.addItem(i)
        self.assertEquals(list(self.p.queryIndex(NameIndex, result=INDEX)), ['Fred'])
        i.setIndexKey(NameIndex, 'Bob')
        self.assertEquals(list(self.p.queryIndex(NameIndex, result=INDEX)), ['Bob'])
        i.setIndexKey(NameIndex, None)
        self.assertEquals(list(self.p.queryIndex(NameIndex, result=INDEX)), [])
        i2 = Item(self.store, name='JJ')
        i2.setIndexKey(NameIndex, 'KK')
        self.p.addItem(i2)
        self.assertEquals(list(self.p.queryIndex(NameIndex, result=INDEX)), ['KK'])

    def dbtestAppendToIndex(self):
        i = Item(self.store)
        self.p.addIndex(NameIndex)
        self.p.addItem(i)
        
        i.appendToIndex(NameIndex, 'foo')
        i.appendToIndex(NameIndex, 'bar')
        i.appendToIndex(NameIndex, 'baz')
        
        self.assertEquals(list(self.p.queryIndex()), [i])
        self.assertEquals(list(self.p.queryIndex(NameIndex)), [i, i, i])
        self.assertEquals(list(self.p.queryIndex(NameIndex, result=INDEX)), ['bar', 'baz', 'foo'])
        
        i.removeFromIndex(NameIndex, 'bar')
        i.removeFromIndex(NameIndex, 'foo')
        self.assertEquals(list(self.p.queryIndex()), [i])
        self.assertEquals(list(self.p.queryIndex(NameIndex)), [i])
        self.assertEquals(list(self.p.queryIndex(NameIndex, result=INDEX)), ['baz'])


    def dbtestIndexIntrospection(self):
        self.failIf(self.p.hasIndex(NameIndex))
        self.assertRaises(NoSuchIndex, self.p.removeIndex, NameIndex)
        self.p.addIndex(NameIndex)
        # self.assertRaises(DuplicateIndex, self.p.addIndex, NameIndex)
        ni = [NameIndex, UIDIndex]
        ni.sort()
        pi = list(self.p.iterIndexes())
        pi.sort()
        self.assertEquals(ni, pi)

    def testDSUIndexMutation(self):
        self.switchToDSU()
        self.testEndKey()

    def testMultipleIndexes(self):
        p = self.p
        def _():
            i = Item(self.store, name='Fred')
            i.setIndexKey(DateIndex, time.time())
            p.addIndex(NameIndex)
            p.addIndex(OtherNameIndex)
            p.addIndex(DateIndex)
            p.addItem(i)
            return i

        def _2(i):
            self.assertEquals(list(p.queryIndex(NameIndex)), [i])
            self.assertEquals(list(p.queryIndex(DateIndex)), [i])
            self.assertEquals(list(p.queryIndex(OtherNameIndex)), [])
        i = self.store.transact(_)
        self.store.transact(_2, i)

    def testDSUMultipleIndexes(self):
        self.switchToDSU()
        self.testEndKey()

    def dbtestItemsMovement(self):
        i = Item(self.store, "mover")
        self.p.addItem(i)
        self.assertEquals(list(self.p.queryIndex()), [i])
        self.failUnless(i in self.p)
        self.p.removeItem(i)
        self.assertEquals(list(self.p.queryIndex()), [])
        self.failUnless(i not in self.p)
        self.p.addItem(i)
        self.assertEquals(list(self.p.queryIndex()), [i])

    def testDSUItemsMovement(self):
        self.switchToDSU()
        self.testItemsMovement()

    def resumableQuerySetUp(self):
        self.p = Pool(self.store, "test pool")
        self.i = [Item(self.store, str(i)) for i in range(5)]
        map(self.p.addItem, self.i)
        self.query = self.p.queryIndex()
        self.assertIdentical(self.i.pop(0), self.query.next())

    def getTwoItems(self):
        for x in range(2):
            i1 = self.i.pop(0)
            i2 = self.query.next()
            self.assertIdentical(i1, i2, "%s is not %s" % (i1, i2))

    def getAnItem(self):
        i1 = self.i.pop(0)
        i2 = self.query.next()
        self.assertIdentical(i1, i2, "%s is not %s" % (i1, i2))


    def ensureStopIteration(self):
        self.assertRaises(StopIteration, self.query.next)

    multitxnResumableQueries = [resumableQuerySetUp, getTwoItems,
                                    getAnItem, getAnItem,
                                    ensureStopIteration]

    def resumableQuerySetUp(self):
        p = Pool(self.store)
        i = Item(self.store)
        p.addItem(i)
        return p, i

    def makeQuery(self, pool, item):
        return pool.queryIndex(), item

    def resumeQuery(self, query, item):
        self.assertIdentical(query.next(), item)
        return query,

    def finishQuery(self, query):
        self.assertRaises(StopIteration, query.next)

    multitxnQueryInitiallyResumable = [resumableQuerySetUp, makeQuery,
                                       resumeQuery, finishQuery]


    def dbtestQueryResumableBeforeStarted(self):
        p = Pool(self.store)
        p.addItem(Item(self.store))
        q = p.queryIndex()
        q.next()
        self.assertRaises(StopIteration, q.next)

    def dbtestIndexRemoval(self):
        p = Pool(self.store)
        p.addIndex(NameIndex)
        items = [Item(self.store, str(i)) for i in range(5)]
        for i in items:
            p.addItem(i)

        d = p.removeIndex(NameIndex)
        r = unittest.deferredResult(d)

    def makeItem(self):
        i = Item(self.store)
        # This is a minor hack.  It forces the Scheduled Items pool to allocate
        # its first sub-store ID now, rather than later in the run, which makes
        # the keys before and after the Pool and Item are created and deleted
        # more consistent, and thus easier to test.
        IScheduler(self.store).callItemLater(i)
        return i,

    def getKeyInfo(self, item):
        return item, self.store.db.keys()

    def makePool(self, item, keys):
        # pprint(keys)
        p = Pool(self.store)
        p.addIndex(NameIndex)
        p.addItem(item)
        return p, keys

    def printInfo(self, pool, keys):
        # pprint(self.store.db.keys())
        return pool, keys

    def deletePool(self, pool, keys):
        unittest.deferredResult(pool.deleteFromStore())
        return keys,

    def checkKeyInfo(self, keys):
        newKeys = self.store.db.keys()
        # pprint(newKeys)
        newNotInOld = []
        for k in newKeys:
            if k not in keys:
                newNotInOld.append(k)
        self.failIf(newNotInOld, "Some extra keys %r not in original %r" % (newNotInOld, keys))
        
    multitxnPoolRemoval = [makeItem, getKeyInfo, makePool, printInfo, deletePool, checkKeyInfo]
    del makeItem, getKeyInfo, makePool, printInfo, deletePool, checkKeyInfo

    def resumeableQuerySetUp(self):
        p = Pool(self.store)
        for i in range(3):
            p.addItem(Item(self.store))
        return p,
   
    def startQuery(self, p):
        q = p.queryIndex(result=INDEX_AND_ITEM)
        return p, q, -1
   
    def grabItem(self, p, q, lastIndex):
        if lastIndex >= 0:
            p.removeItem(p.getIndexItem(UIDIndex, lastIndex))
        index, item = q.next()
        self.assertEquals(index, lastIndex + 1)
        return p, q, index
   
    def finished(self, p, q, index):
        pass
   
    multitxnResumeQueryWithRemoval = [resumeableQuerySetUp, startQuery, grabItem, grabItem, grabItem, finished]

    def dbtestIndexCount(self):
        p = Pool(self.store)
        p.addIndex(IndexA)
        p.addIndex(IndexB)

        i = ItemAB(self.store)
        iA = ItemAB(self.store, a='foo')
        iB = ItemAB(self.store, b='bar')
        iAB = ItemAB(self.store, a='baz', b='quux')

        for o in (i, iA, iB, iAB):
            p.addItem(o)

        self.assertEquals(p.countIndex(IndexA), 2)
        self.assertEquals(p.countIndex(IndexB), 2)
        self.assertEquals(p.countIndex(UIDIndex), 4)

        i.a = 'new value'
        self.assertEquals(p.countIndex(IndexA), 3)
        self.assertEquals(p.countIndex(IndexB), 2)
        self.assertEquals(p.countIndex(UIDIndex), 4)

        iB.b = None
        self.assertEquals(p.countIndex(IndexA), 3)
        self.assertEquals(p.countIndex(IndexB), 1)
        self.assertEquals(p.countIndex(UIDIndex), 4)

        p.removeItem(iAB)
        self.assertEquals(p.countIndex(IndexA), 2)
        self.assertEquals(p.countIndex(IndexB), 0)
        self.assertEquals(p.countIndex(UIDIndex), 3)

        p.removeItem(iB)
        self.assertEquals(p.countIndex(IndexA), 2)
        self.assertEquals(p.countIndex(IndexB), 0)
        self.assertEquals(p.countIndex(UIDIndex), 2)

        p.removeItem(iA)
        self.assertEquals(p.countIndex(IndexA), 1)
        self.assertEquals(p.countIndex(IndexB), 0)
        self.assertEquals(p.countIndex(UIDIndex), 1)

        p.removeItem(i)
        self.assertEquals(p.countIndex(IndexA), 0)
        self.assertEquals(p.countIndex(IndexB), 0)
        self.assertEquals(p.countIndex(UIDIndex), 0)
    
    def dbtestIterpools(self):
        i = Item(self.store)
        p = Pool(self.store)
        p.addItem(i)
        self.assertEquals(list(i.iterpools()), [p])
        p.removeItem(i)
        self.assertEquals(list(i.iterpools()), [])
        
    def dbtestBackwardsCompatibilityPoolsAttribute(self):
        i = Item(self.store)
        p = Pool(self.store)
        p.addItem(i)
        self.assertEquals(i.pools, {p.storeID: 0})
        p.removeItem(i)
        self.assertEquals(i.pools, {})


class SecondaryIndices(unittest.TestCase):
    def dbSetUp(self):
        self.p = Pool(self.store)
        self.p.addIndex(XI)
        self.p.addIndex((XI, YI))

        i = Item(self.store, 'I')
        i.setIndexKey(XI, 'A')
        i.setIndexKey(YI, '1')
        self.p.addItem(i)

        j = Item(self.store, 'J')
        j.setIndexKey(XI, 'B')
        j.setIndexKey(YI, '2')
        self.p.addItem(j)

        self.i, self.j = i, j

    def dbtestResultOrder(self):
        i, j = self.i, self.j

        items = list(self.p.queryIndex(XI))
        self.assertEquals(items, [i, j])

        j.setIndexKey(XI, 'A')
        items = list(self.p.queryIndex(XI))
        self.assertEquals(items, [i, j])

        j.setIndexKey(YI, '0') # This is a zero
        items = list(self.p.queryIndex((XI, YI)))
        self.assertEquals(items, [j, i])

    def dbtestResultType(self):
        i, j = self.i, self.j

        values = list(self.p.queryIndex(XI, result=INDEX))
        self.assertEquals(values, ['A', 'B'])

        values = list(self.p.queryIndex((XI, YI), result=INDEX))
        self.assertEquals(values, [('A', '1'), ('B', '2')])

        values = list(self.p.queryIndex((XI, YI), result=INDEX_AND_ITEM))
        self.assertEquals(values, [(('A', '1'), i), (('B', '2'), j)])

        values = list(self.p.queryIndex((XI, YI), result=STORE_ID))
        self.assertEquals(values, [i.storeID, j.storeID])

        values = list(self.p.queryIndex((XI, YI), result=INDEX_AND_STORE_ID))
        self.assertEquals(values, [(('A', '1'), i.storeID), (('B', '2'), j.storeID)])

    def dbtestIndexPresence(self):
        i, j = self.i, self.j

        self.assertEquals(self.p.countIndex(XI), 2)
        self.assertEquals(self.p.countIndex((XI, YI)), 2)

        i.setIndexKey(YI, None)
        self.assertEquals(self.p.countIndex(XI), 2)
        self.assertEquals(self.p.countIndex((XI, YI)), 1)

        items = list(self.p.queryIndex((XI, YI)))
        self.assertEquals(items, [j])

        i.setIndexKey(YI, '3')
        self.assertEquals(self.p.countIndex((XI, YI)), 2)

        items = list(self.p.queryIndex((XI, YI)))
        self.assertEquals(items, [i, j])

        i.setIndexKey(XI, 'B')
        self.assertEquals(self.p.countIndex((XI, YI)), 2)
        items = list(self.p.queryIndex((XI, YI)))
        self.assertEquals(items, [j, i])

        i.setIndexKey(XI, None)
        self.assertEquals(self.p.countIndex((XI, YI)), 1)

        items = list(self.p.queryIndex((XI, YI)))
        self.assertEquals(items, [j])

class ID(unittest.TestCase):
    def dbtestNextID(self):
        x = self.store.nextID()
        self.assertEquals(x + 1, self.store.nextID())
        self.assertEquals(x + 2, self.store.nextID())
        self.assertEquals(x + 3, self.store.nextID())

        self.assertEquals(x + 4, self.store.nextID(allocate=False))
        self.assertEquals(x + 4, self.store.nextID(allocate=False))

    def dbtestNextSubID(self):
        i = Item(self.store)
        x = self.store.nextSubID(i)
        self.assertEquals(x + 1, self.store.nextSubID(i))
        self.assertEquals(x + 2, self.store.nextSubID(i))
        self.assertEquals(x + 3, self.store.nextSubID(i))

        self.assertEquals(x + 4, self.store.nextSubID(i, allocate=False))
        self.assertEquals(x + 4, self.store.nextSubID(i, allocate=False))

    def dbtestSetSubID(self):
        i = Item(self.store)
        for j in xrange(0, 103, 3):
            self.store.setSubID(i, j)
            self.assertEquals(self.store.getSubID(i), j)

class ResurrectionTest(unittest.TestCase):
    def testResurrection(self):
        p = self.mktemp()
        s = Store(opj(p, 'db'), opj(p, 'files'))
        a = s.transact(SubStore, s)
        sid = a.storeID
        i = s.transact(Item, a, name="Foo Foo")
        iid = i.storeID

        s.close()
        del s
        gc.collect()

        s = Store(opj(p, 'db'), opj(p, 'files'))
        a = s.getItemByID(sid)
        self.failUnless(isinstance(a, SubStore))
        i = a.getItemByID(iid)
        self.failUnless(isinstance(i, Item))
        self.assertEquals(i.name, "Foo Foo")
        s.close()

class ItemForwardingComponentTestCase(unittest.TestCase):
    def testEquality(self):
        C = fwd.ItemForwardingComponent
        self.assertEquals(C("fwd"), C("fwd"))
        self.failIfEqual(C("fwd"), "fwd")
        self.failIfEqual(C("fwd"), C("bar"))

    def dbtestInstantiationRestriction(self):
        self.assertRaises(RuntimeError,
                          fwd.ItemForwardingComponent,
                          Item(self.store))

import weakref

class Empty:
    def __finalizer__(self):
        return lambda: None

class Finalizable:
    L = []

    def __init__(self, L=None):
        if L is None:
            L = []
        self.L = L

    def __finalizer__(self):
        return lambda L=self.L, i=id(self): L.append(i)

class ViciouslyDefectiveFinalizable:
    def __init__(self, cache, path, L):
        self.cache = cache
        self.path = path
        self.L = L

    def __finalizer__(self):
        def inner():
            raise NotImplementedError()
        return inner

class ExceptionRaisingFinalizable:
    def __finalizer__(self):
        return lambda: 1 / 0

class FinalizingCacheTestCase(trialtest.TestCase):
    def testNoStrongReferences(self):
        o = Empty()
        r = weakref.ref(o)
        c = FinalizingCache()
        c.cache("path", o)
        del o
        self.assertEquals(r(), None)
        self.failIf(c.has("path"))
        self.assertRaises(KeyError, c.get, "path")

    def testFinalization(self):
        L = []
        o = Finalizable(L)
        i = id(o)
        c = FinalizingCache()
        c.cache("path", o)
        del o
        self.assertEquals(L, [i])
        self.failIf(c.has("path"))
        self.assertRaises(KeyError, c.get, "path")

    def testGarbageCollectorInteraction(self):
        L = []
        o1 = Finalizable(L)
        i1 = id(o1)
        o2 = Finalizable(L)
        i2 = id(o2)

        o1.ref = o2
        o2.ref = o1

        c = FinalizingCache()
        c.cache("path1", o1)
        c.cache("path2", o2)

        del o1, o2
        gc.collect()

        L.sort()
        L2 = [i1, i2]
        L2.sort()
        self.assertEquals(L, L2)
        self.failIf(c.has("path1"))
        self.failIf(c.has("path2"))
        self.assertRaises(KeyError, c.get, "path1")
        self.assertRaises(KeyError, c.get, "path2")

    def testGarbageFinalization(self):
        L = []
        c = FinalizingCache()
        o1 = ViciouslyDefectiveFinalizable(c, "path2", L)
        i1 = id(o1)
        o2 = ViciouslyDefectiveFinalizable(c, "path1", L)
        i2 = id(o2)

        o1.ref = o2
        o2.ref = o1

        c.cache("path1", o1)
        c.cache("path2", o2)

        # gc.collect()
        # print len(gc.get_objects())
        del o1, o2
        gc.collect()
        log.flushErrors(NotImplementedError)
        # print len(gc.get_objects())

        self.failIf([o for o in gc.garbage if getattr(o, '__class__') is ViciouslyDefectiveFinalizable])

    def testFinalizingException(self):
        c = FinalizingCache()
        o = ExceptionRaisingFinalizable()
        c.cache("path", o)
        del o
        self.failIf(c.has("path"))
        self.assertRaises(KeyError, c.get, "path")
        log.flushErrors(ZeroDivisionError)

class Finalizer:
    def __init__(self, f, *a, **kw):
        self.f, self.a, self.kw = f, a, kw

    def __del__(self):
        self.f(*self.a, **self.kw)

class StrongCacheTestCase(trialtest.TestCase):
    def testHoldsReferences(self):
        C = StrongCache()
        C.startService()
        L = []
        C.cache('woop', Finalizer(L.append, 'woop gone'))
        self.failIf(L)
        C.stopService()

    def testTick(self):
        C = StrongCache(3)
        C.startService()
        L = []

        C.cache('woop', Finalizer(L.append, 'woop gone'))

        for i in range(3):
            self.failIf(L)
            C.tick()
        self.assertEquals(L, ["woop gone"])
        C.stopService()

    def testResize(self):
        C = StrongCache(3)
        C.startService()
        L = []

        C.cache('woop', Finalizer(L.append, 'woop gone'))
        C.tick()
        C.cache('poow', Finalizer(L.append, 'poow gone'))
        C.grow(3)

        for i in range(5):
            self.failIf(L)
            C.tick()

        self.assertEquals(L, ["woop gone"])
        C.tick()
        self.assertEquals(L, ["woop gone", "poow gone"])

        C.stopService()

    def testFailureHandling(self):
        from twisted.internet import reactor
        from twisted.python import log

        C = StrongCache()
        C.tick = lambda: 1 / 0
        C.interval = 0.1
        C.startService()
        reactor.iterate(1.0)
        errors = log.flushErrors(ZeroDivisionError)
        self.assertEquals(len(errors), 1)
        errors[0].trap(ZeroDivisionError)

class IMisc(Interface):
    pass

class IWhatever(Interface):
    '''
    '''

class Foo(Item):
    bar = referenced

class Whatever(StoredAdapter):
    implements(IWhatever)
    boz = referenced(IMisc)

backwardsCompatImplements(Whatever)

class Misc(StoredAdapter):
    implements(IMisc)
backwardsCompatImplements(Misc)
class ItemMetaTestCase(unittest.TestCase):
    def testReferenced(self):
        def create():
            return Foo(self.store), Item(self.store)
        a, b = self.store.transact(create)

        def set():
            a.touch()
            a.bar = b
        self.store.transact(set)

        def get():
            return a.bar
        self.assertIdentical(self.store.transact(get), b)

        self.store.cache.uncache(b.getItemPath())

        p = objgrep(a, b, isSame)
        self.failIf(p, "Should be no reference paths (%r)" % (p,))
    try:
        import twisted.trial.assertions
    except ImportError:
        pass
    else:
        testReferenced.todo = 'Trial interferes with the garbage collector something fierce.'

    def testNoneReferenced(self):
        def create():
            return Item(self.store)

        a, b = self.store.transact(create), None

        def set():
            a.touch()
            a.bar = b
        self.store.transact(set)

        def get():
            return a.bar
        self.assertIdentical(self.store.transact(get), b)

    def testAdapterReferenced(self):
        def create():
            a = Componentized()
            a.setComponent(IWhatever, Whatever(a))
            a.setComponent(IPersistable, Persistability(a, self.store))
            b = Componentized()
            b.setComponent(IMisc, Misc(b))
            b.setComponent(IPersistable, Persistability(b, self.store))
            return a, b
        a, b = self.store.transact(create)
        def set():
            IPersistable(a).touch()
            IWhatever(a).boz = b

        self.store.transact(set)

        def get():
            return IWhatever(a).boz
        self.assertIdentical(self.store.transact(get), IMisc(b))

class ItemRemoval(unittest.TestCase):
    def createItem(self):
        self.item = Item(self.store, "This Item")
        self.itemID = self.item.storeID

    def deleteItem(self):
        self.item.deleteFromStore()

    def determineExistence(self):
        self.assertRaises(KeyError, self.store.getItemByID, self.itemID)

    multitxnMultipleTransactionDeletion = (createItem, deleteItem, determineExistence)
    del createItem, deleteItem, determineExistence

    def dbtestRemovedFromPools(self):
        i = Item(self.store, "That Item")
        iid = i.storeID
        p = Pool(self.store, "Stuff")
        p.addItem(i)

        i.deleteFromStore()

        self.assertEquals(list(p.queryIndex()), [])

class SubStoreRemoval(unittest.TestCase):
    storePowerups = [SystemScheduler]

    def scheduleItem(self):
        # We need to do this to force the scheduled pool to grow a next id key.
        i = Item(self.store)
        IScheduler(self.store).callItemLater(i)

    def peekStore(self):
        # XXX - Abstraction violation.  Is there a better way to
        # ensure that everything gets ripped out of the database?
        self.initialKeys = self.store.db.keys()

    def createStuff(self):
        # Create a SubStore, give it some Items
        self.substore = s = SubStore(self.store)
        IPowerStation(self.substore).installPowerup(SubScheduler)
        p = Pool(s)
        p.addItem(Item(s))

    def checkSanity(self):
        self.failUnless(len(self.store.db.keys()) > len(self.initialKeys))

    def destruction(self):
        # Destroy the SubStore
        # XXX - Maybe assert something about the return value
        #       but I don't know what that might be yet.
        unittest.deferredResult(self.substore.deleteFromStore())

    def checkStore(self):
        a = self.store.db.keys()
        b = self.initialKeys
        self.assertEquals(a, b,
                          '%s != %s' % (pformat(a), pformat(b)))

    multitxnAnnihilation = [scheduleItem, peekStore,
                            createStuff, checkSanity,
                            destruction, checkStore]
    del scheduleItem, peekStore, createStuff, checkSanity, destruction, checkStore

from atop.store import IComponentForwarder

class RedirectFrom:
    def __init__(self, original):
        self.original = original

class WhateverFlyweight:
    implements(IWhatever, IComponentForwarder)

    def forwardComponent(self, interface):
        assert interface is IWhatever
        return RedirectFrom(self)

backwardsCompatImplements(WhateverFlyweight)

class WhateverItem(Item, WhateverFlyweight):
    implements(IComponentForwarder, IWhatever)

backwardsCompatImplements(WhateverItem)

class ComponentTestCase(unittest.TestCase):
    def dbtestMultipleInterfaces(self):
        item = Item(self.store)
        self.store.setComponent(IWhatever, item)
        self.store.setComponent(IMisc, item)

        self.assertIdentical(IWhatever(self.store), item)
        self.assertIdentical(IMisc(self.store), item)

    def dbtestSimpleForward(self):
        subst = SubStore(self.store)
        flyw = WhateverFlyweight()
        subst.setComponent(IWhatever, flyw)
        i1 = IWhatever(subst)
        self.failUnless( isinstance(i1, RedirectFrom) )
        self.failUnlessEqual(i1.original, flyw)
        itm = WhateverItem(subst)
        subst.setComponent(IWhatever, itm)
        i2 = IWhatever(subst)
        i1 = IWhatever(subst)
        self.failUnless( isinstance(i2, RedirectFrom) )
        self.failUnlessEqual(i2.original, itm)

class TestException(Exception):
    pass

class CyclicStoreReferenceConsistency(unittest.TestCase):
    def testReversion(self):
        def _1():
            self.substore = SubStore(self.store)
            item = Item(self.substore)
            self.substore.setComponent(IWhatever, item)
        self.store.transact(_1)

        def _2():
            self.substore.touch()
            raise TestException("Revert this transaction")
        self.assertRaises(TestException, self.store.transact, _2)

        def _3():
            self.assertIdentical(IWhatever(self.substore).store, self.substore)
        self.store.transact(_3)

class Nicknames(NameIndex):
    pass

class Evil(Item):
    nicknames = indexedList(Nicknames)

class Myth(Item):
    langnames = indexedPoolDict(Nicknames)

class ExampleUser(Item):
    localpartsInDomain = indexedPoolDictOfLists(Nicknames)

class ComplexIndexTest(unittest.TestCase):

    def dbtestMyths(self):
        roman = Pool(self.store)
        roman.addIndex(Nicknames)
        greek = Pool(self.store)
        greek.addIndex(Nicknames)
        zeus = Myth(self.store)
        zeus.langnames[greek] = 'zeus'
        zeus.langnames[roman] = 'jupiter'
        hera = Myth(self.store)
        hera.langnames[greek] = 'hera'
        hera.langnames[roman] = 'juno'
        mercury = Myth(self.store)
        mercury.langnames.update({
            roman: 'mercury',
            greek: 'hermes'})
        athena = Myth(self.store)
        athena.langnames[greek] = 'athena'
        athena.langnames.update({roman: 'minerva'})
        apollo = Myth(self.store)
        apollo.langnames[greek] = 'apollo'
        for x in roman, greek: x.addItem(apollo)
        apollo.langnames[roman] = 'apollo'

        for p in roman, greek:
            for myt in zeus, hera, mercury, athena:
                p.addItem(myt)

        lrqi = list(roman.queryIndex(Nicknames, result=INDEX_AND_ITEM))
        lgqi = list(greek.queryIndex(Nicknames, result=INDEX_AND_ITEM))

        self.assertEquals(lgqi,
                          [('apollo', apollo),
                           ('athena', athena),
                           ('hera', hera),
                           ('hermes', mercury),
                           ('zeus', zeus),
                           ])

        self.assertEquals(lrqi,
                          [('apollo', apollo),
                           ('juno', hera),
                           ('jupiter', zeus),
                           ('mercury', mercury),
                           ('minerva', athena)
                           ])

    def dbtestEvils(self):
        evils = Pool(self.store)
        evils.addIndex(Nicknames)
        devil = Evil(self.store)
        n = devil.nicknames
        n.append('devil')
        evils.addItem(devil)
        n.extend(['satan', 'lucifer'])
        lqi = list(evils.queryIndex(Nicknames, result=INDEX_AND_ITEM))

        self.assertEquals(lqi,
                          [('devil', devil),
                           ('lucifer', devil),
                           ('satan', devil)])

    def dbtestMultiDomains(self):
        dotcom = Pool(self.store)
        dotorg = Pool(self.store)

        bob = ExampleUser(self.store)

        for p in dotcom, dotorg:
            p.addIndex(Nicknames)
            p.addItem(bob)

        bob.localpartsInDomain[dotorg].append('bob')
        bob.localpartsInDomain[dotcom].extend([
            'bob', 'support', 'billing'])


class Paths(unittest.TestCase):
    def dbtestNewSubFileForItem(self):
        i = Item(self.store)
        f = self.store.newSubFileForItem(i)
        f.write('Test bytes')
        p = f.close("test_file")
        self.assertEquals(file(p.path).read(), 'Test bytes')

    def dbtestNewSubDirForItem(self):
        i = Item(self.store)
        f = self.store.newSubDirForItem(i, "test_dir")
        self.failUnless(os.path.exists(f.path))

    def dbtestNewFile(self):
        a = self.createAvatar([])
        f = a.newFile()
        f.write('Bytes of test')
        p = f.close('file_test')
        self.assertEquals(file(p.path).read(), 'Bytes of test')

    def dbtestNewDirectory(self):
        a = self.createAvatar([])
        d = a.newDirectory('directory_test')
        self.failUnless(os.path.exists(d.path))

    def serializePath(self):
        i = Item(self.store)
        i.d = self.store.newSubDirForItem(i, "persistability")
        return i.storeID,

    def unserializePath(self, storeID):
        self.store.getItemByID(storeID)

    multitxnPersistability = [serializePath, unserializePath]
    del serializePath, unserializePath

    def testRelocatablePaths(self):
        storeRoot = self.mktemp()
        s = Store(os.path.join(storeRoot, 'db'), os.path.join(storeRoot, 'files'))
        def setUp():
            a = SubStore(s)
            i = Item(a)
            f = a.newFile()
            f.write('relocation test')
            i.p = f.close('test_relocation')
            return a.storeID, i.storeID
        subStoreID, itemID = s.transact(setUp)
        s.close()

        newRoot = self.mktemp()
        os.rename(storeRoot, newRoot)
        s = Store(os.path.join(newRoot, 'db'), os.path.join(newRoot, 'files'))
        def retest():
            i = s.getItemByID(subStoreID).getItemByID(itemID)
            self.assertEquals(file(i.p.path).read(), 'relocation test')
        s.transact(retest)
        s.close()


class IndexedTestCase(unittest.TestCase):
    def testCallOnChange(self):
        def test():
            item = CallOnChangeItem(self.store)
            self.assertEquals(item.called, False)
            item.fu = 'bar'
            self.assertEquals(item.called, 'bar')
        self.store.transact(test)


class TupleMapTest(unittest.TestCase):
    def dbtestBasics(self):
        m = TupleMap(self.store)
        m['a', 'b', 'c'] = 'def'
        m['a'] = 'nothing'
        m['a', 'c'] = 'xxx'
        self.assertEquals(list(m.queryitems(prefix=['a'], withValues=False)),
                          [('b','c'), ('c',)])
        self.assertEqual(list(m.queryitems(withValues=False)),
                         [('a',),
                          ('a','b','c'),
                          ('a','c')])

class StringMapTest(unittest.TestCase):
    def dbtestSimpleInterface(self):
        m = StringMap(self.store)
        self.assertRaises(TypeError, m.__setitem__, 5, 'hello')
        self.assertRaises(TypeError, m.__setitem__, 'hello', 5)
        self.assertRaises(TypeError, m.__setitem__, 5, 5)

        self.assertRaises(KeyError, m.__getitem__, 'hello')
        self.assertEquals(m.get('hello', 'foo'), 'foo')
        m['hello'] = 'world'
        self.assertEquals(m['hello'], 'world')
        self.assertEquals(m.get('hello'), 'world')
        self.assertEquals(m.get('hello', 'foo'), 'world')
        del m['hello']
        self.assertRaises(KeyError, m.__getitem__, 'hello')
        self.assertEquals(m.get('hello', 'foo'), 'foo')

    def dbtestDictMethods(self):
        m = StringMap(self.store)

        letters = 'abcdefghijklmnopqrstuvwxyz'
        for k in letters:
            m[k] = 'value for: ' + k

        expectedKeys = list(letters)
        expectedValues = ['value for: ' + k for k in letters]
        self.assertEquals(m.keys(), expectedKeys)
        self.assertEquals(m.values(), expectedValues)
        self.assertEquals(m.items(), zip(expectedKeys, expectedValues))

        for k in letters:
            self.assertEquals(list(m.queryitems(k)), [('', 'value for: ' + k)])

    def setupMapAndQuery(self):
        m = StringMap(self.store)

        m['abb'] = 'garbage'
        m['abc'] = 'ERRONEOUS VALUE'
        m['abc/'] = '123'
        m['abc/def'] = '456'
        m['abd'] = 'more garbage'

        q = m.queryitems('abc/')
        return q,

    def getFirstItem(self, query):
        self.assertEquals(query.next(), ('', '123'))
        return query,

    def getSecondItem(self, query):
        self.assertEquals(query.next(), ('def', '456'))
        return query,

    def ensureFinished(self, query):
        self.assertRaises(StopIteration, query.next)

    multitxnIterMethods = [setupMapAndQuery, getFirstItem, getSecondItem,
                           ensureFinished]

from atop.storable import Storable, StoredAttribute, UnstoredAttribute, StoredTupleMapping

class FunStorable(Storable):
    name = StoredAttribute()
    frotz = UnstoredAttribute()
    tups = StoredTupleMapping()

class Storables(unittest.TestCase):

    def dbtestTupleMapAttr(self):
        fs = FunStorable()
        fs.addToStore(self.store)
        m = fs.tups
        m['a', 'b', 'c'] = 'def'
        m['a'] = 'nothing'
        m['a', 'c'] = 'xxx'
        self.assertEquals(list(m.queryitems(prefix=['a'], withValues=False)),
                          [('b','c'), ('c',)])

    def instantiation(self):
        s = FunStorable()
        s.addToStore(self.store)
        s.name = "Happy Fun Thing"
        return s,

    def checkName(self, s):
        self.assertEquals(s.name, "Happy Fun Thing")
        return s,

    def serializeDeserialize(self, s):
        newS = self.store.getItemByID(s.storeID)
        self.assertEquals(newS.name, "Happy Fun Thing")
        self.assertIdentical(s, newS)
        return s.storeID,

    def deserialization(self, storeID):
        newNewS = self.store.getItemByID(storeID)
        self.assertEquals(newNewS.name, "Happy Fun Thing")

    def dbtestPersistence(self):
        s, = self.instantiation()
        s, = self.checkName(s)
        path, = self.serializeDeserialize(s)
        self.deserialization(path)

    multitxnPersistence2 = [instantiation, checkName,
                            serializeDeserialize, deserialization]

    def createStorableWithUnstoredAttributes(self):
        s = FunStorable()
        s.addToStore(self.store)
        s.frotz = 79
        return s.storeID,

    def loadStorableWithoutUnstoredAttributes(self, storeID):
        s = self.store.getItemByID(storeID)
        self.assertRaises(AttributeError, getattr, s, 'frotz')

    multitxnStorableWithUnstoredAttributes = [
        createStorableWithUnstoredAttributes,
        loadStorableWithoutUnstoredAttributes]

from atop.store import HierarchyIndex, HierarchyBrowser

class HierarchicalItem(Item):
    path = indexed(HierarchyIndex)

    def __init__(self, store, name, parent):
        Item.__init__(self, store, name)
        if parent is None:
            self.path = (self.storeID,)
        else:
            self.path = parent.path + (self.storeID,)

class HierarchicalTestCase(unittest.TestCase):
    def dbSetUp(self):
        self.p = Pool(self.store)
        self.p.addIndex(HierarchyIndex)

        H = lambda n, p=None: HierarchicalItem(self.store, n, p)

        # a
        # a/b
        # a/b/c
        # a/b/d
        # a/e
        # a/e/f
        # a/e/g
        # a/h
        # a/h/i
        # a/j
        # a/k
        # w
        # w/x
        # w/x/y
        # w/z

        # Map of hierarchical items to the name of their parent.
        config = {
            'a': None,
            'b': 'a',
            'c': 'b',
            'd': 'b',
            'e': 'a',
            'f': 'e',
            'g': 'e',
            'h': 'a',
            'i': 'h',
            'j': 'a',
            'k': 'a',
            'w': None,
            'x': 'w',
            'y': 'x',
            'z': 'w'}

        items = self.items = {}
        itemNames = config.keys()
        itemNames.sort()
        for k in itemNames:
            v = config[k]
            if v is None:
                parent = None
            else:
                parent = items[v]
            items[k] = H(k, parent)

        for item in items.values():
            self.p.addItem(item)

    def dbtestIndex(self):
        def get(name):
            return self.items[name]

        itemNames = self.items.keys()
        itemNames.sort()
        for startPos in range(len(itemNames)):
            for endPos in range(startPos, len(itemNames)):
                startItemName = itemNames[startPos]
                endItemName = itemNames[endPos]

                results = list(
                    self.p.queryIndex(HierarchyIndex,
                                      startKey=get(startItemName).path,
                                      endKey=get(endItemName).path))

                expected = map(get, itemNames[startPos:endPos + 1])

                self.assertEquals(results, expected)

    def dbtestBrowser(self):
        browser = HierarchyBrowser(self.p)

        def vector(names):
            return [(self.items[o].path, self.items[o].storeID) for o in names]

        self.assertEquals(
            list(browser.queryAtDepth((), 1)),
            vector('aw'))

        self.assertEquals(
            list(browser.queryAtDepth(self.items['a'].path, 1)),
            vector('abehjk'))

        self.assertEquals(
            list(browser.queryAtDepth(self.items['b'].path, 1)),
            vector('bcd'))

        self.assertEquals(
            list(browser.queryAtDepth(self.items['a'].path, 2)),
            vector('abcdefghijk'))

