try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO
import time, datetime, operator

from twisted.trial import unittest, util
from twisted.internet import defer, reactor
from twisted.python import components

from atop import tpython
from atop.ifcompat import implements, backwardsCompatImplements

class IFakeTransport(components.Interface):
    pass

class FakeTransport:
    implements(IFakeTransport)

    connLost = False
    prodReg = None
    unregProd = None
    def __init__(self):
        self.data = []
    def loseConnection(self):
        self.connLost = True
    def write(self, data):
        self.data.append(data)
    def writeSequence(self, iovec):
        self.data.extend(iovec)
    def getPeer(self):
        return "PEER"
    def registerProducer(self, producer, streaming):
        self.prodReg = (producer, streaming)
    def unregisterProducer(self):
        self.unregProd = True

backwardsCompatImplements(FakeTransport)

class FakeProtocol:
    connMade = None
    connLost = None
    def __init__(self):
        self.data = []
    def dataReceived(self, data):
        self.data.append(data)
    def makeConnection(self, transport):
        self.connMade = transport

class DebugShimTestCase(unittest.TestCase):
    def setUp(self):
        self.s = StringIO()
        self.p = FakeProtocol()
        self.dst = tpython.DebugShimTransport(self.p, self.s)
        self.o = FakeTransport()
        self.dst.makeConnection(self.o)

    def testMakeConnection(self):
        self.assertEquals(self.s.getvalue(), "*\n")

    def testDataReceived(self):
        self.dst.dataReceived("Foo")
        self.dst.dataReceived("Bar")
        self.assertEquals(self.s.getvalue(), "*\nC 1: 'Foo'\nC 1: 'Bar'\n")
        self.assertEquals(self.p.data, ["Foo", "Bar"])

    def testLoseConnection(self):
        self.dst.loseConnection()
        self.failUnless(self.o.connLost)
        self.assertEquals(self.s.getvalue(), "*\nS 4: *\n")

    def testWrite(self):
        self.dst.write("foo")
        self.dst.write("bar")
        self.dst.write("baz")
        self.assertEquals(self.o.data, ["foo", "bar", "baz"])
        self.assertEquals(self.s.getvalue(), "*\nS 7: 'foo'\nS 7: 'bar'\nS 7: 'baz'\n")

    def testGetPeer(self):
        self.assertEquals(self.dst.getPeer(), "PEER")

    def testProducer(self):
        self.dst.registerProducer("foo", "bar")
        self.dst.unregisterProducer()
        self.assertEquals(self.o.prodReg, ("foo", "bar"))
        self.failUnless(self.o.unregProd)

    def testImplements(self):
        self.failUnless(components.implements(self.p.connMade, IFakeTransport))

class TPythonTestException(Exception):
    pass

class IteratorTest(unittest.TestCase):
    def testIterateInReactor(self):
        i = iter(range(10))
        d = tpython.iterateInReactor(i, 0)
        self.assertIdentical(unittest.deferredResult(d), i)

    def testFailureInIterateInReactor(self):
        def failer():
            yield 10
            yield 20
            raise TPythonTestException("test_tpython Test Exception")
        d = tpython.iterateInReactor(failer(), 0)
        e = unittest.deferredError(d)
        e.trap(TPythonTestException)

class HumanIntervalTestCase(unittest.TestCase):
    def testSeconds(self):
        self.assertEquals(tpython.humanInterval(0), "0 seconds")
        self.assertEquals(tpython.humanInterval(1), "1 second")
        self.assertEquals(tpython.humanInterval(30), "30 seconds")
        self.assertEquals(tpython.humanInterval(59), "59 seconds")

    def testMinutes(self):
        self.assertEquals(tpython.humanInterval(60), "1 minute")
        self.assertEquals(tpython.humanInterval(61), "1 minute, 1 second")
        self.assertEquals(tpython.humanInterval(119), "1 minute, 59 seconds")
        self.assertEquals(tpython.humanInterval(599), "9 minutes, 59 seconds")

    def testHours(self):
        self.assertEquals(tpython.humanInterval(60 * 60), "1 hour")
        self.assertEquals(tpython.humanInterval(60 * 61 + 1), "1 hour, 1 minute, 1 second")
        self.assertEquals(tpython.humanInterval(60 * 60 * 3), "3 hours")

    def testLots(self):
        self.assertEquals(tpython.humanInterval(604799), "6 days, 23 hours, 59 minutes, 59 seconds")

class LockingTestCase(unittest.TestCase):
    def testBasics(self):
        lockf = self.mktemp()
        lock = tpython.FilesystemLock(lockf)
        self.failUnless(lock.lock())
        self.failUnless(lock.clean)
        lock.unlock()
        self.failUnless(lock.lock())
        self.failUnless(lock.clean)
        lock.unlock()

    def testProtection(self):
        lockf = self.mktemp()
        lock = tpython.FilesystemLock(lockf)
        self.failUnless(lock.lock())
        self.failUnless(lock.clean)
        self.failIf(lock.lock())
        lock.unlock()

    def testBigLoop(self):
        lockf = self.mktemp()
        lock = tpython.FilesystemLock(lockf)
        self.failUnless(lock.lock())
        for i in xrange(500):
            self.failIf(lock.lock())
        lock.unlock()

    # A multiprocess test would be good here, for the sake of
    # completeness.  However, it is fairly safe to rely on the
    # filesystem to provide the semantics we require.

class AsyncMap(unittest.TestCase):
    def testSync(self):
        self.expectedAssertions = 5

        args = range(5, 10), range(7, 13), range(50)
        argCopy = zip(*args)
        def f(a, b, c):
            self.assertEquals((a, b, c), argCopy.pop(0))
        unittest.deferredResult(tpython.amap(f, *args))

    def testAsync(self):
        self.expectedAssertions = 3

        args = ['a', 'b', 'c']
        argCopy = args[:]
        def f(a):
            d = defer.Deferred()
            reactor.callLater(0, d.callback, None)
            self.assertEquals(a, argCopy.pop(0))
        unittest.deferredResult(tpython.amap(f, args))

class TestUtils(unittest.TestCase):

    def testDeferLaterPositive(self):
        self.assertEquals(util.wait(tpython.deferLater(0.01, lambda : 1)),
                          1)

    def testDeferLaterNegative(self):
        self.assertRaises(ZeroDivisionError, util.wait, tpython.deferLater(0.01, lambda : 1/0))


class TestTime(unittest.TestCase):
    class MST(datetime.tzinfo):
        def tzname(self, dt):
            return 'MST'
        def utcoffset(self, dt):
            return datetime.timedelta(hours = -7)
        def dst(self, dt):
            return datetime.timedelta(0)

    class CET(datetime.tzinfo):
        def tzname(self, dt):
            return 'MST'
        def utcoffset(self, dt):
            return datetime.timedelta(hours = 1)
        def dst(self, dt):
            return datetime.timedelta(0)

    reference = datetime.datetime(2004, 12, 6, 14, 15, 16)
    awareReference = datetime.datetime(2004, 12, 6, 14, 15, 16, tzinfo=tpython.FixedOffset(0, 0))

    def _checkReference(self, timeInstance, reference=None):
        """Check timeInstance against self.reference."""
        self.assertEquals(timeInstance._time, reference or self.reference)

    def _createReference(self, reference=None):
        """Return a refernce instance."""
        return tpython.Time.fromDatetime(reference or self.reference)

    def testPytzWeirdness(self):
        """pytz weirdness; RT ticket #2755"""
        try:
            import pytz
        except ImportError:
            raise unittest.SkipTest, 'pytz could not be imported'
        tz = pytz.timezone('America/Detroit')
        time = tpython.Time.fromRFC2822('Wed, 06 Apr 2005 23:12:27 -0400')
        dtime = time.asDatetime(tz)
        self.assertEquals(dtime.hour, 23)
        self.assertEquals(dtime.minute, 12)

    def testCmp(self):
        now = time.gmtime()
        self.assertEquals(tpython.Time.fromStructTime(now), tpython.Time.fromStructTime(now))
        self.assertNotEquals(tpython.Time.fromStructTime(now), tpython.Time.fromStructTime(time.localtime()))
        self.assertNotEquals(tpython.Time.fromStructTime(now), 13)

        aTime = tpython.Time.fromStructTime(now)
        for op in 'lt', 'le', 'gt', 'ge':
            self.assertRaises(TypeError, getattr(operator, op), aTime, now)

    def testFromNow(self):
        diff = datetime.datetime.utcnow() - tpython.Time()._time
        if diff < datetime.timedelta():
            diff = -diff
        self.failUnless(diff.days == 0 and diff.seconds <= 5, 'Time created now is %r away from now' % (diff,))

    def testFromHumanly(self):
        def test(input, expected, tzinfo=None):
            time = tpython.Time.fromHumanly(
                input,
                tzinfo,
                self._createReference())

            self.assertEquals(
                time.asISO8601TimeAndDate(tzinfo=tzinfo),
                expected)

            return time

        def testMalformed(input):
            self.assertRaises(ValueError, tpython.Time.fromHumanly, input)

        def testDay(input, expected, tzinfo=None):
            time = test(input, expected, tzinfo)
            self.assert_(time.isAllDay(tzinfo=tzinfo))

        def testMinute(input, expected, tzinfo=None):
            time = test(input, expected, tzinfo)
            self.assertEquals(time.resolution, datetime.timedelta(minutes=1))

        # 'now' is Monday, 2004-12-06 14:15:16
        testDay('yesterday',       '2004-12-05T00:00:00+00:00')
        testDay(' ToDaY ',         '2004-12-06T00:00:00+00:00')
        testDay('   TuESDaY ',     '2004-12-07T00:00:00+00:00')
        testDay(' ToMoRroW ',      '2004-12-07T00:00:00+00:00')
        testDay('wednesday',       '2004-12-08T00:00:00+00:00')
        testDay('This wednesday',  '2004-12-08T00:00:00+00:00')
        testDay('neXt wednesday',  '2004-12-08T00:00:00+00:00')
        testDay('thursday',        '2004-12-09T00:00:00+00:00')
        testDay('friday',          '2004-12-10T00:00:00+00:00')
        testDay('saturday',        '2004-12-11T00:00:00+00:00')
        testDay('sunday',          '2004-12-12T00:00:00+00:00')
        testDay('sunday',          '2004-12-12T00:00:00-07:00', self.MST())
        testDay('monday',          '2004-12-13T00:00:00+00:00')

        testMinute('15:00',        '2004-12-06T15:00:00+00:00')
        testMinute(' 14:00  ',     '2004-12-07T14:00:00+00:00')
        testMinute(' 2:00  pm ',   '2004-12-07T14:00:00+00:00')
        testMinute(' 02:00  pm ',  '2004-12-07T14:00:00+00:00')
        testMinute(' noon ',       '2004-12-07T12:00:00+00:00')
        testMinute('midnight',     '2004-12-07T00:00:00+00:00')

        testMalformed('24:01')
        testMalformed('24:00')  # this one might be considered valid by some people, but it's just dumb.
        testMalformed('13:00pm')
        testMalformed('13:00am')

        # these are perfectly reasonable cases, but are totally broken. Good enough for demo work.
        testMalformed('13:00 tomorrow')
        testMalformed('13:00 next thursday')
        testMalformed('last monday')

    def testFromISO8601DateAndTime(self):
        self._checkReference( tpython.Time.fromISO8601TimeAndDate('2004-12-06T14:15:16') )
        self._checkReference( tpython.Time.fromISO8601TimeAndDate('20041206T141516') )
        self._checkReference( tpython.Time.fromISO8601TimeAndDate('20041206T091516-0500') )
        self._checkReference( tpython.Time.fromISO8601TimeAndDate('2004-12-06T07:15:16', self.MST()) )
        self._checkReference( tpython.Time.fromISO8601TimeAndDate('2004-12-06T14:15:16Z', self.MST()) )
        self._checkReference( tpython.Time.fromISO8601TimeAndDate('2004-12-06T14:15:16-0000', self.MST()) )
        self._checkReference( tpython.Time.fromISO8601TimeAndDate('2004-12-06T14:15:16-0000') )
        self._checkReference( tpython.Time.fromISO8601TimeAndDate('2004-W50-1T14:15:16') )
        self._checkReference( tpython.Time.fromISO8601TimeAndDate('2004-341T14:15:16') )

        self.assertRaises( ValueError, tpython.Time.fromISO8601TimeAndDate, '2005-W53' )
        self.assertRaises( ValueError, tpython.Time.fromISO8601TimeAndDate, '2004-367' )
        try:
            tpython.Time.fromISO8601TimeAndDate('2004-366')
        except ValueError:
            raise unittest.FailTest, 'leap years should have 366 days'

        try:
            tpython.Time.fromISO8601TimeAndDate('2004-123T14-0600')
            tpython.Time.fromISO8601TimeAndDate('2004-123T14:13-0600')
            tpython.Time.fromISO8601TimeAndDate('2004-123T14:13:51-0600')
        except ValueError:
            raise unittest.FailTest, 'timezone should be allowed if time with *any* resolution is specified'

        self.assertEquals( tpython.Time.fromISO8601TimeAndDate('2005').resolution, datetime.timedelta(days=365) )
        self.assertEquals( tpython.Time.fromISO8601TimeAndDate('2004').resolution, datetime.timedelta(days=366) )
        self.assertEquals( tpython.Time.fromISO8601TimeAndDate('2004-02').resolution, datetime.timedelta(days=29) )
        self.assertEquals( tpython.Time.fromISO8601TimeAndDate('2004-02-29').resolution, datetime.timedelta(days=1) )
        self.assertEquals( tpython.Time.fromISO8601TimeAndDate('2004-02-29T13').resolution, datetime.timedelta(hours=1) )
        self.assertEquals( tpython.Time.fromISO8601TimeAndDate('2004-02-29T13:10').resolution, datetime.timedelta(minutes=1) )
        self.assertEquals( tpython.Time.fromISO8601TimeAndDate('2004-02-29T13:10:05').resolution, datetime.timedelta(seconds=1) )
        self.assertEquals( tpython.Time.fromISO8601TimeAndDate('2004-02-29T13:10:05.010').resolution, datetime.timedelta(microseconds=1000) )
        self.assertEquals( tpython.Time.fromISO8601TimeAndDate('2004-02-29T13:10:05.010000').resolution, datetime.timedelta(microseconds=1) )
        self.assertEquals( tpython.Time.fromISO8601TimeAndDate('2004-02-29T13:10:05.010000123').resolution, datetime.timedelta(microseconds=1) )
        self.assertEquals( tpython.Time.fromISO8601TimeAndDate('2004-W11').resolution, datetime.timedelta(days=7) )
        self.assertEquals( tpython.Time.fromISO8601TimeAndDate('2004-W11-3').resolution, datetime.timedelta(days=1) )
        self.assertEquals( tpython.Time.fromISO8601TimeAndDate('2004-W11-3T14:16:21').resolution, datetime.timedelta(seconds=1) )
        self.assertEquals( tpython.Time.fromISO8601TimeAndDate('2004-123').resolution, datetime.timedelta(days=1) )
        self.assertEquals( tpython.Time.fromISO8601TimeAndDate('2004-123T14:16:21').resolution, datetime.timedelta(seconds=1) )

    def testFromStructTime(self):
        self._checkReference( tpython.Time.fromStructTime((2004, 12, 6, 14, 15, 16, 0, 0, 0)) )
        self._checkReference( tpython.Time.fromStructTime((2004, 12, 6, 7, 15, 16, 0, 0, 0), self.MST()) )
        self._checkReference( tpython.Time.fromStructTime((2004, 12, 6, 15, 15, 16, 0, 0, 0), self.CET()) )
        self._checkReference( tpython.Time.fromStructTime(time.struct_time((2004, 12, 6, 7, 15, 16, 0, 0, 0)), self.MST()) )

    def testFromDatetime(self):
        self._checkReference( tpython.Time.fromDatetime(datetime.datetime(2004, 12, 6, 14, 15, 16)) )
        self._checkReference( tpython.Time.fromDatetime(datetime.datetime(2004, 12, 6, 7, 15, 16, tzinfo=self.MST())) )
        self._checkReference( tpython.Time.fromDatetime(datetime.datetime(2004, 12, 6, 15, 15, 16, tzinfo=self.CET())) )

    def testFromPOSIXTimestamp(self):
        self._checkReference( tpython.Time.fromPOSIXTimestamp(tpython.mkgmtime((2004, 12, 6, 14, 15, 16, 0, 0, 0))) )

    def testFromRFC2822(self):
        self._checkReference( tpython.Time.fromRFC2822('Mon, 6 Dec 2004 14:15:16 -0000') )
        self._checkReference( tpython.Time.fromRFC2822('Mon, 6 Dec 2004 9:15:16 -0500') )
        self._checkReference( tpython.Time.fromRFC2822('6   Dec   2004   9:15:16   -0500') )
        self._checkReference( tpython.Time.fromRFC2822('Mon,6 Dec 2004 9:15:16 -0500') )
        self._checkReference( tpython.Time.fromRFC2822('Mon,6 Dec 2004 9:15 -0500'), datetime.datetime(2004, 12, 6, 14, 15) )
        self._checkReference( tpython.Time.fromRFC2822('Mon,6 Dec 2004 9:15:16 EST') )
        self._checkReference( tpython.Time.fromRFC2822('Monday,6 December 2004 9:15:16 EST') )
        self.assertRaises( ValueError, tpython.Time.fromRFC2822, 'some invalid date' )

    def testObsoleteRFC2822(self):
        self._checkReference( tpython.Time.fromRFC2822('Monday,6 December (i hate this month) 2004 9:15:16 R') )

    testObsoleteRFC2822.todo = '''\
    email.Utils implementation does not handle obseleted military style
    timezones, nor does it handled obseleted comments in the header'''

    def testAsPOSIXTimestamp(self):
        self.assertEquals( self._createReference().asPOSIXTimestamp(), 1102342516 )

    def testAsRFC2822(self):
        self.assertEquals( self._createReference().asRFC2822(), 'Mon, 6 Dec 2004 14:15:16 -0000' )
        self.assertEquals( self._createReference().asRFC2822(self.MST()), 'Mon, 6 Dec 2004 07:15:16 -0700' )
        self.assertEquals( self._createReference().asRFC2822(self.CET()), 'Mon, 6 Dec 2004 15:15:16 +0100' )

    def testAsISO8601TimeAndDate(self):
        self.assertEquals(
            self._createReference().asISO8601TimeAndDate(),
            '2004-12-06T14:15:16+00:00' )
        self.assertEquals(
            self._createReference(reference=datetime.datetime(2004, 12, 6, 14, 15, 16, 43210)).asISO8601TimeAndDate(),
            '2004-12-06T14:15:16.04321+00:00' )
        self.assertEquals(
            self._createReference().asISO8601TimeAndDate(tzinfo=self.MST()),
            '2004-12-06T07:15:16-07:00' )
        self.assertEquals(
            self._createReference().asISO8601TimeAndDate(tzinfo=self.CET()),
            '2004-12-06T15:15:16+01:00' )
        self.assertEquals(
            self._createReference().asISO8601TimeAndDate(includeTimezone=False),
            '2004-12-06T14:15:16' )
        self.assertEquals(
            self._createReference(reference=datetime.datetime(2004, 12, 6, 14, 15, 16, 43210)).asISO8601TimeAndDate(includeTimezone=False),
            '2004-12-06T14:15:16.04321' )
        self.assertEquals(
            self._createReference().asISO8601TimeAndDate(tzinfo=self.MST(), includeTimezone=False),
            '2004-12-06T07:15:16' )
        self.assertEquals(
            self._createReference().asISO8601TimeAndDate(tzinfo=self.CET(), includeTimezone=False),
            '2004-12-06T15:15:16' )

    def testAsStructTime(self):
        self.assertEquals( self._createReference().asStructTime(), (2004, 12, 06, 14, 15, 16, 0, 341, 0) )
        self.assertEquals( self._createReference().asStructTime(tzinfo=self.MST()), (2004, 12, 06, 7, 15, 16, 0, 341, 0) )
        self.assertEquals( self._createReference().asStructTime(tzinfo=self.CET()), (2004, 12, 06, 15, 15, 16, 0, 341, 0) )

    def testAsDatetime(self):
        datetime.datetime(2004, 12, 6, 14, 15, 16)
        self.assertEquals( self._createReference().asDatetime(), self.reference )
        self.assertEquals( self._createReference().asDatetime(tzinfo=self.MST()), self.awareReference )
        self.assertEquals( self._createReference().asDatetime(tzinfo=self.CET()), self.awareReference )

    def testAsHumanly(self):
        self.assertEquals(
            self._createReference().asHumanly(now=tpython.Time.fromStructTime((2004, 12, 6, 14, 15, 16, 0, 0, 0))),
            '02:15 pm' )
        self.assertEquals(
            self._createReference().asHumanly(tzinfo=self.MST(), now=tpython.Time.fromStructTime((2004, 12, 6, 14, 15, 16, 0, 0, 0))),
            '07:15 am' )
        self.assertEquals(
            self._createReference().asHumanly(tzinfo=self.CET(), now=tpython.Time.fromStructTime((2004, 12, 6, 14, 15, 16, 0, 0, 0))),
            '03:15 pm' )

        self.assertEquals(
            self._createReference().asHumanly(now=tpython.Time.fromStructTime((2005, 12, 6, 14, 15, 16, 0, 0, 0))),
            '6 Dec 04, 02:15 pm' )
        self.assertEquals(
            self._createReference().asHumanly(tzinfo=self.MST(), now=tpython.Time.fromStructTime((2005, 12, 6, 14, 15, 16, 0, 0, 0))),
            '6 Dec 04, 07:15 am' )
        self.assertEquals(
            self._createReference().asHumanly(tzinfo=self.CET(), now=tpython.Time.fromStructTime((2005, 12, 6, 14, 15, 16, 0, 0, 0))),
            '6 Dec 04, 03:15 pm' )

        allDay = tpython.Time.fromISO8601TimeAndDate('2005-123')
        self.assertEquals(allDay.asHumanly(now=allDay), 'all day')

    def testInverse(self):
        for style in [
        'POSIXTimestamp',
        'Datetime',
        'RFC2822',
        'StructTime',
        'ISO8601TimeAndDate']:
            parse = getattr(tpython.Time, 'from'+style)
            format = getattr(tpython.Time, 'as'+style)
            self.assertEquals( self._createReference(), parse(format(self._createReference())), '%s() is not the inverse of %s()' % (style, style))

    def testEvalRepr(self):
        import datetime         # needed for eval
        now = tpython.Time()
        self.assertEquals( now, eval(repr(now)) )

    def testContainment(self):
        makeTime = tpython.Time.fromISO8601TimeAndDate

        self.assertIn(makeTime('2004-05'), makeTime('2004'))
        self.assertNotIn(makeTime('2005-01'), makeTime('2004'))

    def testArithmetic(self):
        self._checkReference( tpython.Time.fromISO8601TimeAndDate('2004-12-03T14:15:16') + datetime.timedelta(days=3) )
        self._checkReference( tpython.Time.fromISO8601TimeAndDate('2004-12-09T14:15:16') - datetime.timedelta(days=3) )

    def testOneDay(self):
        day = self._createReference().oneDay()
        self.assertEquals(day._time, datetime.datetime(2004, 12, 6, 0, 0, 0))
        self.assertEquals(day.resolution, datetime.timedelta(days=1))

        day = self._createReference().oneDay(self.MST())
        self.assertEquals(day._time, datetime.datetime(2004, 12, 6, 7, 0, 0))
        self.assertEquals(day.resolution, datetime.timedelta(days=1))

    def testIsAllDay(self):
        self.failIf(self._createReference().isAllDay())
        self.failUnless(tpython.Time.fromISO8601TimeAndDate('2005-123').isAllDay())


class TestNamedProperty(unittest.TestCase):

    def testNamedProperty(self):
        class TestProperty(tpython.NamedProperty):
            value = 0
            def getter(self, oself, name, value):
                return (name, value + 1)
            def setter(self, oself, name, value):
                self.value = value
                return value
        prop = TestProperty()
        class X(object):
            y = prop

        x = X()
        x.y = 1
        self.assertEquals(x.y, ('y', 2))
        self.assertEquals(prop.value, 1)
        x.y = 700
        self.assertEquals(x.y, ('y', 701))
        self.assertEquals(prop.value, 700)
        self.assertEquals(prop.name, 'y')


from atop.tpython import dsu
from string import letters
class TestDSU(unittest.TestCase):
    def testSorting(self):
        L = list(letters)
        L.sort(lambda x, y: cmp(x.lower(), y.lower()))
        L2 = list(letters)
        dsu(L2, lambda x: x.lower())
        self.assertEquals(L, L2)
