
from twisted.trial import unittest

from atop.util import across, multiQueryIndex
from atop import store

class TestAcross(unittest.TestCase):
    def testSimple(self):
        simple = ['abc', 'def']
        result = list('abcdef')
        self.assertEquals(
            list(across(simple)), result)

    def testInterleave(self):
        interleave = ['ace', 'bdf']
        result = list('abcdef')
        self.assertEquals(
            list(across(interleave)), result)

    def testUniquify(self):
        nonunique = ['aabbccdd', 'aabbccddee']
        result = list('abcde')
        self.assertEquals(
            list(across(nonunique)), result)

    def testNonuniqify(self):
        nonunique = ['abc', 'abc']
        result = list('aabbcc')
        self.assertEquals(
            list(across(nonunique, uniquify=False)),
            result)

    def testDecorator(self):
        reverser = lambda i, x: x.lower()
        input = ['abc', 'ABC']
        result = list('abc')
        self.assertEquals(
            map(
                lambda x: x.lower(),
                list(across(input, decorator=reverser))),
            result)

    def testConflictResolvesInIteratorOrderAlways(self):
        """
        Test that conflicts between decorated items are resolved by iterator order.

        across() was developed primarily for the use case of Q2Q sharing.  In
        the sharing use case, we sort a set of roles in priority order, then
        return objects which are equivalent in terms of persistent identity
        (e.g. URL) but differ in terms of associated authority.  The roles are
        therefore sorted in priority order, with the 'best' role coming first
        in the list of iterables passed to across().  Items declared equivalent
        by their decorators should, therefore, be uniquified using the order of
        iterators as the tiebreaker.

        The iterators are always passed in the same order, regardless of
        whether the iterators themselves are in forwards or backwards order.
        Therefore this behavior is preserved independent of the reverse= flag
        passed to across().
        """

        inputs = [list('abc'), list('ABC')]
        equivalence = lambda i, o: o.lower()
        self.assertEquals(list(across(inputs, decorator=equivalence)),
                          ['a','b','c'])

        inputs = [list('cba'), list('CBA')]
        equivalence = lambda i, o: o.lower()
        self.assertEquals(list(across(inputs, decorator=equivalence, reverse=True)),
                          ['c','b','a'])

from atop.test import unittest
class TestMultiQuery(unittest.TestCase):
    def dbtestCorrectness(self):
        x = store.Item(self.store, "x test item")
        y = store.Item(self.store, "y test item")

        p = store.Pool(self.store, "MultiQuery Pool")
        q = store.Pool(self.store, "Other MultiQuery Pool")

        p.addIndex(store.NameIndex)
        q.addIndex(store.NameIndex)

        p.addItem(x, uid=1)
        q.addItem(y, uid=1)

        self.assertEquals(
            [x],
            list(multiQueryIndex([p, q], store.UIDIndex)))
