# -*- test-case-name: "atop.test.test_versioning" -*-

import new, inspect

from twisted.trial import unittest
from twisted.python.reflect import qual
from twisted.python import components, context

from atop.ifcompat import implements, backwardsCompatImplements
from atop import store, versioning
from atop.test import unittest as atopunit

class IStableInterface(components.Interface):
    pass

class A(versioning.Versionable):
    implements(IStableInterface)

    def __init__(self):
        self.list = ['1', '2', '3', '4']
        self.dict = {'a': 'b', 'c': 'd'}
        self.int = 7
        self.float = 42.5
        self.str = 'foo'

    def twiddle(self, x, y):
        self.twiddled = (x, y)

    def touch(self):
        # These are kinda-sorta pretending to be Items.
        pass

backwardsCompatImplements(A)

class SubA(A):
    pass

class B(versioning.Versionable):
    pass

def makeUpgrader(*delete, **set):
    def upgrade(inert):
        for (k, v) in set.items():
            inert.state[k] = v
        for k in delete:
            del inert.state[k]
    code, globals = upgrade.func_code, upgrade.func_globals,
    name = 'deleting (' + ' '.join(delete) + ') and setting ' + repr(set)
    argdefs = inspect.getargspec(upgrade)
    closure = upgrade.func_closure
    return new.function(code, globals, name, argdefs, closure)

class VersioningMixin:
    def setUp(self):
        self.registry = versioning.theUpgradeRegistry.clone()
        self.registerUpgrader = self.registry.registerUpgrader
        self.loadWithVersions = self.registry.loadWithVersions
        self.saveWithVersions = self.registry.saveWithVersions


class InertTestCase(unittest.TestCase, VersioningMixin):
    def setUp(self):
        # Down with Trial.  Down with Twisted.  Down with Python.
        # Down with me, mostly.  -exarkun
        for cls in self.__class__.__bases__:
            cls.setUp(self)

    def testPersistenceForgets(self):
        orig = A()

        A.persistenceForgets = ('list', 'int', 'float')
        try:
            pkl = self.saveWithVersions(orig)
        finally:
            del A.persistenceForgets

        new = self.loadWithVersions(pkl)
        self.assertEquals(new.dict, {'a': 'b', 'c': 'd'})
        self.assertEquals(new.str, 'foo')
        self.failIf(hasattr(new, 'list'))
        self.failIf(hasattr(new, 'int'))
        self.failIf(hasattr(new, 'float'))

    def testContextRemembers(self):
        orig = A()
        pkl = self.saveWithVersions(orig)

        A.contextRemembers = [('contextName', 'attrName')]

        # Normally classes will not have their contextRemembers/
        # persistenceForgets attributes twiddled externally like
        # this.  In fact, I don't think we will support it at all!
        # But it is handy for this unit test to do.  Manually bumping
        # the version state validity will prevent the caching mechanism
        # from missing the change.
        self.registry.versionStateValidity += 1
        try:
            new = context.call({'contextName': 'contextValue'},
                               self.loadWithVersions, pkl)
            for attr in ('list', 'dict', 'int', 'float', 'str'):
                self.assertEquals(getattr(orig, attr), getattr(new, attr))
            self.assertEquals(new.attrName, 'contextValue')
        finally:
            del A.contextRemembers

    def testRegistryVersionValidity(self):
        x = self.registry.versionStateValidity
        self.registerUpgrader(qual(A), 1, {}, lambda i: None)
        self.assertNotEqual(x, self.registry.versionStateValidity)
        self.assertEquals(self.registry.versionStateValidity, self.registry.versionStateValidity)

    def testHighestVersion(self):
        self.assertEquals(self.registry.highestVersion(qual(A)), 0)
        self.registerUpgrader(qual(A), 1, {}, lambda i: None)
        self.assertEquals(self.registry.highestVersion(qual(A)), 1)
        self.assertEquals(self.registry.highestVersion(qual(B)), 0)
        self.registerUpgrader(qual(B), 1, {}, lambda i: None)
        self.assertEquals(self.registry.highestVersion(qual(B)), 1)
        self.registerUpgrader(qual(B), 2, {}, lambda i: None)
        self.assertEquals(self.registry.highestVersion(qual(B)), 2)
        self.assertEquals(self.registry.highestVersion(qual(A)), 1)

    def testSameVersion(self):
        orig = A()
        pkl = self.saveWithVersions(orig)
        new = self.loadWithVersions(pkl)

        for attr in ('list', 'dict', 'int', 'float', 'str'):
            self.assertEqual(getattr(orig, attr), getattr(new, attr))

    def testSingleUpgrade(self):
        orig = A()
        pkl = self.saveWithVersions(orig)

        aTo1 = makeUpgrader('str', list=['2', '1'], float=6*9)
        self.registerUpgrader(qual(A), 1, {}, aTo1)

        new = self.loadWithVersions(pkl)
        self.assertEquals(new.list, ['2', '1'])
        self.assertEquals(new.dict, {'a': 'b', 'c': 'd'})
        self.assertEquals(new.int, 7)
        self.assertEquals(new.float, 6 * 9)
        self.failIf(hasattr(new, 'str'))

    def testMultipleUpgrades(self):
        orig = A()
        pkl = self.saveWithVersions(orig)

        aTo1 = makeUpgrader('int', str='hello')
        aTo2 = makeUpgrader('float', list=['4', '3', '2', '1'])
        aTo3 = makeUpgrader(int=42, str='hello, world')
        self.registerUpgrader(qual(A), 1, {}, aTo1)
        self.registerUpgrader(qual(A), 2, {}, aTo2)
        self.registerUpgrader(qual(A), 3, {}, aTo3)

        new = self.loadWithVersions(pkl)
        self.assertEquals(new.list, ['4', '3', '2', '1'])
        self.assertEquals(new.dict, {'a': 'b', 'c': 'd'})
        self.assertEquals(new.int, 42)
        self.failIf(hasattr(new, 'float'))
        self.assertEquals(new.str, 'hello, world')

    def testUpgradeDependencies(self):
        orig = SubA()
        pkl = self.saveWithVersions(orig)

        aTo1 = makeUpgrader(key='value')

        def subTo1(inert):
            inert.state['subkey'] = inert.state['key']

        self.registerUpgrader(qual(A), 1, {}, aTo1)
        self.registerUpgrader(qual(SubA), 1, {qual(A): 1}, subTo1)

        new = self.loadWithVersions(pkl)
        self.assertEquals(new.key, 'value')
        self.assertEquals(new.subkey, 'value')

    def testMultipleUpgradeOrdering(self):
        orig = A()
        orig.b = B()
        pkl = self.saveWithVersions(orig)

        upgrades = []
        def makeOrderedUpgrader(k):
            def orderedUpgrader(inert):
                inert.state[k] = len(upgrades)
                upgrades.append(inert)
            return orderedUpgrader

        for cls in (A, B):
            for n in (1, 2):
                u = makeOrderedUpgrader(cls.__name__ + str(n))
                self.registerUpgrader(qual(cls), n, {}, u)

        new = self.loadWithVersions(pkl)
        self.assertEquals(new.A1, 2)
        self.assertEquals(new.A2, 3)
        self.assertEquals(new.b.B1, 0)
        self.assertEquals(new.b.B2, 1)
        self.assertEquals(len(upgrades), 4)
        self.assertIdentical(upgrades[0], upgrades[1])
        self.assertEquals(upgrades[1].name, qual(B))
        self.assertIdentical(upgrades[2], upgrades[3])
        self.assertEquals(upgrades[3].name, qual(A))
    testMultipleUpgradeOrdering.skip = "This kind of dependency is not supported"

    def testCyclicReferences(self):
        origA = A()
        origB = B()
        origA.b, origB.a = origB, origA
        pkl = self.saveWithVersions(origA)

        aTo1 = makeUpgrader(upgradedA=True)
        bTo1 = makeUpgrader(upgradedB=True)

        self.registerUpgrader(qual(A), 1, {}, aTo1)
        self.registerUpgrader(qual(B), 1, {}, bTo1)

        newA = self.loadWithVersions(pkl)
        self.failUnless(newA.upgradedA)
        self.failUnless(isinstance(newA.b, B))
        self.failUnless(newA.b.upgradedB)
        self.assertIdentical(newA.b.a, newA)

    def testUpgradeHooks(self):
        orig = A()
        pkl = self.saveWithVersions(orig)

        def aTo1(inert):
            inert.callAfterUpgrade(IStableInterface, 'twiddle', 'x', y='y')
        self.registerUpgrader(qual(A), 1, {}, aTo1)

        new = self.loadWithVersions(pkl)
        self.assertEquals(new.twiddled, ('x', 'y'))

class VersionedItem(store.Item):
    pass

class UpgradeTestCase(atopunit.TestCase, VersioningMixin):
    def setUp(self):
        # Down with Trial.  Down with Twisted.  Down with Python.
        # Down with me, mostly.  -exarkun
        for cls in self.__class__.__bases__:
            cls.setUp(self)

    def dbSetUp(self):
        self.meths = versioning.loadWithVersions, versioning.saveWithVersions
        versioning.loadWithVersions, versioning.saveWithVersions = self.loadWithVersions, self.saveWithVersions

    def dbTearDown(self):
        versioning.loadWithVersions, versioning.saveWithVersions = self.meths
        del self.meths

    def setUpgrader(self, upgradeToVersion1):
        self.registerUpgrader(qual(VersionedItem), 1, {}, upgradeToVersion1)

    def createItems(self):
        self.pool = store.Pool(self.store, "Test Pool")
        self.pool.addItem(VersionedItem(self.store, "Test Item"))

    def jackVersioned(self):
        def upgradeToVersion1(inert):
            created = store.Pool(self.store, "New Test Pool")
            pf = inert.state['poolRef'] = versioning.Inert(qual(store.ItemReference))
            pf.state = { 'storeID': created.storeID,
                         'iface': None }
            inert.state['jacked'] = True
        self.setUpgrader(upgradeToVersion1)

    def reloadItem(self):
        item = list(self.pool.queryIndex())[0]
        self.failUnless(item.jacked)
        self.assertEquals(item.poolRef.getItem().name, "New Test Pool")

    multitxnSimpleUpgrade = [createItems, jackVersioned, reloadItem]
    del createItems, jackVersioned, reloadItem

    def testRevertDoesNotUpgradeItems(self):
        def createItem():
            i = VersionedItem(self.store)
            return i
        storeID = self.store.transact(createItem).storeID

        def upgradeToVersion1(inert):
            inert.state['upgraded'] = True
        self.setUpgrader(upgradeToVersion1)

        class TestException(Exception):
            pass

        def loadItemAndFail():
            i = self.store.getItemByID(storeID)
            self.failUnless(i.upgraded)
            raise TestException("Revert this transaction")
        self.assertRaises(TestException, self.store.transact, loadItemAndFail)


    def testUpgradeTheSubclass(self):
        oldInstance = SubA()
        oldPickle = self.saveWithVersions(oldInstance)
        didTheUpgraderRun = []

        def upgradeTo1(inertA):
            didTheUpgraderRun.append(True)

        self.registerUpgrader('atop.test.test_versioning.A', 1, {}, upgradeTo1)
        self.loadWithVersions(oldPickle)
        self.failUnless(didTheUpgraderRun, 'The upgrader did not run')

    def testClassChanging(self):
        oldInstance = A()
        oldPickle = self.saveWithVersions(oldInstance)
        
        def upgradeTo1(inertA):
            inertA.name = qual(SubA)
        
        self.registerUpgrader(qual(A), 1, {}, upgradeTo1)
        newInstance = self.loadWithVersions(oldPickle)
        self.failUnless(isinstance(newInstance, SubA))
    
    def testClassReferenceChanging(self):
        oldInstance = A()
        oldInstance.classReference = SubA
        oldPickle = self.saveWithVersions(oldInstance)
        
        def upgradeTo1(inertA):
            inertA.state['classReference'].inertName = qual(B)
        
        self.registerUpgrader(qual(A), 1, {}, upgradeTo1)
        newInstance = self.loadWithVersions(oldPickle)
        self.assertIdentical(newInstance.classReference, B)
    testClassReferenceChanging.todo = "_NamedTypeProxy.inertName does not support mutation yet"
