# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import types
from os.path import join as opj

from twisted.trial import unittest
from twisted.python import reflect, log

from atop import store, sched, credup, powerup, service

def remove(sequence, omit):
    r = []
    for e in sequence:
        try:
            r.append(remove(e, omit))
        except:
            if e not in omit:
                r.append(e)
    return tuple(r)

_unittestSpecials = ['todo', 'skip']

class MetaTestClass(type):
    def makeTransactor(klass, f, name):
        def tf(self):
            try:
                return self.store.transact(f, self)
            except:
                log.err()
        tf = types.FunctionType(tf.func_code, tf.func_globals, name, tf.func_defaults, tf.func_closure)
        for attr in _unittestSpecials:
            hasattr(f, attr) and setattr(tf, attr, getattr(f, attr))
        return tf
    makeTransactor = classmethod(makeTransactor)

    def makeMultiTransactor(klass, fns, name):
        def tf(self):
            args = ()
            for f in fns:
                if getattr(f, 'transact', True):
                    args = self.store.transact(f, self, *args) or ()
                else:
                    args = f(self, *args) or ()
            if args:
                raise unittest.FailTest("Arguments returned from last function in multi-transaction test")

        f = fns[0]
        tf = types.FunctionType(tf.func_code, tf.func_globals, name, tf.func_defaults, tf.func_closure)
        for attr in _unittestSpecials:
            hasattr(f, attr) and setattr(tf, attr, getattr(f, attr))
        return tf
    makeMultiTransactor = classmethod(makeMultiTransactor)

    def __new__(klass, name, bases, attrs):
        for (key, value) in attrs.items():
            if key.startswith('dbtest') and isinstance(value, types.FunctionType):
                del attrs[key]
                attrs[key[2:]] = klass.makeTransactor(value, key[2:])
            elif key.startswith('multitxn') and isinstance(value, (list, tuple)):
                del attrs[key]
                attrs['test' + key[8:]] = klass.makeMultiTransactor(value, 'test' + key[8:])
        if name != 'TestCase':
            return types.ClassType.__new__(types.ClassType, name, remove(bases, [object, TestCase]) + (_TestCase,), attrs)
        return type.__new__(klass, name, bases, attrs)

class TestCase(object):
    """Base class for test cases which rely heavily on the database.

    Methods beginning with the prefix "dbtest" will be run as normal
    tests, but in a database transaction.

    Tuples and lists beginning with "multitxn" will have the functions they
    contain run as a unit test, each function being run in a separate transaction.
    A function in a multitxn list may have the attribute "transact" set to False
    to avoid being invoked in a transaction.

    @cvar dbPath: The path relative to L{storePath} beneath which the database
    environment is created.

    @cvar mailPath: The path relative to L{storePath} beneath which message files
    are stored.

    @cvar storePowerups: A tuple of the powerups to be installed when the store is
    created.

    @ivar storePath: The path beneath the database was created.

    @ivar store: A database instance unique to the running test method.
    
    @ivar closeDbOnTearDown: If set to True, close the store when the test method finishes.
    
    @ivar avatarDomain: If set to a non-None value, indicate the name of the domain all avatars
    created with L{createAvatar} should be created within.
    """

    __metaclass__ = MetaTestClass

class _TestCase(unittest.TestCase):
    dbPath = "db"
    mailPath = "mail"

    schedService = None

    storePowerups = (credup.StoreRealm, service.Serviceup)

    AvatarFactory = credup.Avatar
    avatarDomain = None

    closeDbOnTearDown = True

    def setUp(self):
        self.storePath = self.mktemp()
        self.store = store.Store(opj(self.storePath, self.dbPath), opj(self.storePath, self.mailPath))
        def _():
            self.__storageService = svc = store.StorageService(self.store)
            self.installPowerups(self.store, self.collectPowerups('storePowerups'))
            if self.avatarDomain:
                credup.ICredentialsChecker(self.store).addDomain(self.avatarDomain)
            self.dbSetUp()
            svc.startService()
        self.store.transact(_)

    def tearDown(self):
        self.store.transact(self.dbTearDown)
        unittest.deferredResult(self.__storageService.stopService())

    def dbSetUp(self):
        pass

    def dbTearDown(self):
        pass

    def installPowerups(self, store, powerups):
        ps = powerup.IPowerStation(store)
        for p in powerups:
            ps.installPowerup(p)

    def collectPowerups(self, name):
        d = dict.fromkeys(getattr(self, name, ()))
        for k in reflect.allYourBase(self.__class__):
            d.update(dict.fromkeys(getattr(k, name, ())))
        return d.keys()

    def createAvatar(self, powerups, *a, **kw):
        def _():
            avatar = self.AvatarFactory(self.store, *a, **kw)
            ps = powerup.IPowerStation(avatar)
            if self.avatarDomain:
                credup.ICredentialsChecker(self.store).addAvatar(avatar)
            for p in powerups:
                ps.installPowerup(p)
            return avatar
        return self.store.transact(_)

deferredResult = unittest.deferredResult
deferredError = unittest.deferredError
FailTest = unittest.FailTest
