# -*- test-case-name: atop.test.test_tpython -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.
# objects that should be in tpython

import types, errno, time, sys, os.path, imp, sets, datetime, re

from atop.ifcompat import *

from twisted.internet import defer
from twisted.internet import reactor
from twisted.internet.protocol import Protocol
from twisted.python import reflect, log, failure

def spawnProcess(processProtocol, executable, args=(), env={},
                 path=None, uid=None, gid=None, usePTY=0,
                 packages=()):
    """Launch a process with a particular Python environment.

    All arguments as to reactor.spawnProcess(), except for the
    addition of an optional packages iterable.  This should be
    of strings naming packages the subprocess is to be able to
    import.
    """

    env = env.copy()

    pythonpath = []
    for pkg in packages:
        p = os.path.split(imp.find_module(pkg)[1])[0]
        if p.startswith(os.path.join(sys.prefix, 'lib')):
            continue
        pythonpath.append(p)
    pythonpath = list(sets.Set(pythonpath))
    pythonpath.extend(env.get('PYTHONPATH', '').split(os.pathsep))
    env['PYTHONPATH'] = os.pathsep.join(pythonpath)

    return reactor.spawnProcess(processProtocol, executable, args,
                                env, path, uid, gid, usePTY)

def spawnPythonProcess(processProtocol, args=(), env={},
                       path=None, uid=None, gid=None, usePTY=0,
                       packages=()):
    """Launch a Python process

    All arguments as to spawnProcess(), except the executable
    argument is omitted.
    """
    return spawnProcess(processProtocol, sys.executable,
                        args, env, path, uid, gid, usePTY,
                        packages)

def fork(d):
    """
    Sometimes it is necessary to hand a Deferred back to user code whose result
    will be used later on by intermediary framework code, and while the user
    may validly want to transform the result as part of a result chain,
    framework code must depend on the idempotence of user callbacks with
    respect to the original Deferred.  In those cases, you may return
    C{fork(yourDeferred)} rather than C{yourDeferred} to user code, and be
    assured that there will be no side-effects on yourDeferred when users add
    callbacks.

    @param d: Any Deferred that needs to preserve its value while having
    callbacks added.
    @type d: L{defer.Deferred}

    @return: A new Deferred that will fire with the result of the input
    Deferred, but which will start a new result chain: i.e.  C{addCallback} on
    the result will have no side-effects on the input C{d}.

    @rtype: L{defer.Deferred}
    """
    d2 = defer.Deferred()
    d.addCallbacks(
        lambda r: (d2.callback(r), r)[1],
        lambda f: (d2.errback(f), f)[1])
    return d2

def mkgmtime(gmtt):
    'convert GMT time-tuple to local time'
    if time.daylight and gmtt[-1]:
        zone = time.altzone
    else:
        zone = time.timezone
    return time.mktime(gmtt) - zone

def gmtForYearMonthDay(year, month, day):
    return time.gmtime(                          # 3 to a gmt tuple...
        time.mktime(                             # 2 to a localtime float...
            (year, month, day, 0, 0, 0, 0, 0, 0) # 1 go from a localtime tuple...
        )
    )

def makeLoggingMethod(name, f):
    def loggingMethod(*a, **kw):
        log.msg("%s.%s(*%r, **%r)" % (name, f.func_name, a, kw))
        return f(*a, **kw)
    return loggingMethod

class MethodCallLoggingType(type):
    def __new__(cls, name, bases, attrs):
        for (k, v) in attrs.items():
            if isinstance(v, types.FunctionType):
                attrs[k] = makeLoggingMethod(name, v)
        return super(MethodCallLoggingType, cls).__new__(cls, name, bases, attrs)

class DebugShimTransport(Protocol):
    counter = 0

    def __init__(self, p, f=sys.stdout, lengthLimit=None):
        self.p = p
        self.f = f
        DebugShimTransport.counter += 1
        self.counter = DebugShimTransport.counter
        self.lengthLimit = lengthLimit

    def _mungeData(self, data):
        if self.lengthLimit and len(data) > self.lengthLimit:
            data = data[:self.lengthLimit - 12] + '<... elided>'
        return data

    # Protocol
    def makeConnection(self, t):
        t.write = self.write
        t.loseConnection = self.loseConnection
        self.t = t
        self.f.write('*\n')
        self.f.flush()
        self.p.makeConnection(t)

    def dataReceived(self, data):
        self.f.write( 'C %d: %r\n' % (self.counter, self._mungeData(data)))
        self.p.dataReceived(data)
        self.f.flush()

    def connectionLost(self, reason):
        self.f.write( 'C %d: %r\n' % (self.counter, reason) )
        self.f.flush()
        self.p.connectionLost(reason)

    def factory():
        def get(self):
            return self.p.factory
        def set(self, value):
            self.p.factory = value
        def delete(self, value):
            del self.p.factory
        return get, set, delete
    factory = property(*factory())

    def orig(__self, __name, *a, **kw):
        wrapper = __self.t.__dict__.pop(__name)
        try:
            origmeth = getattr(__self.t, __name)
        finally:
            __self.t.__dict__[__name] = wrapper
        return origmeth(*a, **kw)

    # Transport
    def getHost(self):
        return self.t.getHost()

    def getPeer(self):
        return self.t.getPeer()

    def write(self, data):
        self.f.write( 'S %d: %r\n' % (self.counter, self._mungeData(data)))
        self.f.flush()
        self.orig('write', data)

    disconnecting = False

    def loseConnection(self):
        self.f.write( 'S %d: *\n' % self.counter)
        self.f.flush()
        self.orig('loseConnection')
        self.disconnecting = True

    def registerProducer(self, producer, streaming):
        self.t.registerProducer(producer, streaming)

    def unregisterProducer(self):
        self.t.unregisterProducer()


def iterateInReactor(i, delay=0.0001, holder=[None]):
    d = defer.Deferred()
    def callLaterer():
        try:
            maybe = i.next()
            if isinstance(maybe, defer.Deferred):
                def rslts(x):
                    #print 'rslts', x
                    holder[0] = reactor.callLater(delay, callLaterer)
                    return x
                maybe.addBoth(rslts)
                return
        except StopIteration:
            d.callback(i)
        except:
            d.errback()
        else:
            holder[0] = reactor.callLater(delay, callLaterer)
    callLaterer()
    return d


PERIODS = (
    ('week', 60 * 60 * 24 * 7),
    ('day', 60 * 60 * 24),
    ('hour', 60 * 60),
    ('minute', 60),
    ('second', 1),
)

def humanInterval(seconds):
    s = []
    for (p, d) in PERIODS:
        n = int(seconds / d)
        seconds %= d
        if n:
            s.append('%d %s%s' % (n, p, n > 1 and 's' or ''))
    if s:
        return ', '.join(s)
    return '0 seconds'


class _counter:
    persec = 0
    def accum(self):
        self.persec += 1
        reactor.callLater(0, self.accum)
    def display(self):
        log.msg("FPS: %d" % self.persec)
        self.persec = 0
        reactor.callLater(1, self.display)

def framecounter():
    c = _counter()
    c.display()
    c.accum()

from atop.filepile import readlink, symlink, rmlink
class FilesystemLock:
    """A mutex.

    This relies on the filesystem property that creating
    a symlink is an atomic operation and that it will
    fail if the symlink already exists.  Deleting the
    symlink will release the lock.

    @ivar name: The name of the file associated with this lock.
    @ivar clean: Indicates whether this lock was released cleanly by its
    last owner.  Only meaningful after C{lock} has been called and returns
    True.
    """

    clean = None
    locked = False

    def __init__(self, name):
        self.name = name

    def lock(self):
        """Acquire this lock.

        @rtype: C{bool}
        @return: True if the lock is acquired, false otherwise.

        @raise: Any exception os.symlink() may raise, other than
        EEXIST.
        """
        try:
            pid = readlink(self.name)
        except (OSError, IOError), e:
            if e.errno != errno.ENOENT:
                raise
            self.clean = True
        else:
            if not hasattr(os, 'kill'):
                return False
            try:
                os.kill(int(pid), 0)
            except (OSError, IOError), e:
                if e.errno != errno.ESRCH:
                    raise
                rmlink(self.name)
                self.clean = False
            else:
                return False

        symlink(str(os.getpid()), self.name)
        self.locked = True
        return True

    def unlock(self):
        """Release this lock.

        This deletes the directory with the given name.

        @raise: Any exception os.readlink() may raise, or
        ValueError if the lock is not owned by this process.
        """
        pid = readlink(self.name)
        if int(pid) != os.getpid():
            raise ValueError("Lock %r not owned by this process" % (self.name,))
        rmlink(self.name)
        self.locked = False

import shutil
import hotshot
import tempfile

from twisted.python import components

class IProfiler(components.Interface):
    def start(self):
        """Begin recording profiling information.
        """

    def stop(self):
        """Stop recording profiling information.
        """

    def save(self, fileObj):
        """Save recorded profiling information to the given file.
        """

from twisted.application import service

class HotshotProfilerService(service.Service):
    implements(IProfiler)

    running = False

    def __init__(self, filename=None):
        self.setName("atop.tpython.HotshotProfilerService")
        if filename is None:
            foo, self.fn = tempfile.mkstemp('hotshot')
        else:
            self.fn = filename

    def start(self):
        self.profile = hotshot.Profile(self.fn)
        self.profile.start()
        self.running = True

    def stop(self):
        self.profile.stop()
        self.running = False

    def save(self, fileObj):
        shutil.copyfileobj(file(self.fn, 'rb'), fileObj)

backwardsCompatImplements(HotshotProfilerService)


class PickleableSequenceIterator:
    def __init__(self, sequence):
        self.sequence = sequence
        self.index = 0

    def __iter__(self):
        return self

    def next(self):
        try:
            e = self.sequence[self.index]
        except IndexError:
            self.sequence = ()
            self.index = 0
            raise StopIteration
        else:
            self.index += 1
            return e

def amap(f, *i):
    """Asynchronous map

    Applies f to each element of the sequence i, waiting on the Deferred
    returned by f before proceeding to the next element.
    """
    return _amapHelper(None, f, map(PickleableSequenceIterator, i))

def _amapHelper(result, f, iterables):
    try:
        args = [i.next() for i in iterables]
    except StopIteration:
        return defer.succeed(None)
    else:
        return defer.maybeDeferred(f, *args).addCallback(_amapHelper, f, iterables)

class CannotFindFunction(ValueError):
    pass


class Tracer(object):
    skip = object()

    installed = False

    def install(self):
        self.installed = True
        sys.settrace(self.trace)

    def uninstall(self):
        self.installed = False
        sys.settrace(None)

    def toggle(self):
        if self.installed:
            self.uninstall()
        else:
            self.install()

    def trace(self, frame, event, arg):
        r = getattr(self, 'trace_' + event.upper())(frame, arg)
        if r is self.skip:
            return None
        elif r is None:
            return self.trace
        else:
            return r

    def trace_CALL(self, frame, arg):
        pass

    def trace_LINE(self, frame, arg):
        pass

    def trace_RETURN(self, frame, arg):
        pass

    def trace_EXCEPTION(self, frame, arg):
        pass

def extractArgs(frame):
    co = frame.f_code
    dict = frame.f_locals
    n = co.co_argcount
    if co.co_flags & 4: n = n+1
    if co.co_flags & 8: n = n+1
    result = {}
    for i in range(n):
        name = co.co_varnames[i]
        result[name] = dict.get(name, "*** undefined ***")
    return result

def formatArgs(args):
    return ', '.join(['='.join((k, reflect.safe_repr(v))) for (k, v) in args.iteritems()])

class Spewer(Tracer):
    callDepth = 0

    def trace_CALL(self, frame, arg):
        self.callDepth += 1

        frameSelf = frame.f_locals.get('self')
        if frameSelf is not None:
            if hasattr(frameSelf, '__class__'):
                k = reflect.qual(frameSelf.__class__)
            else:
                k = reflect.qual(type(frameSelf))
            k = k + '.'
        else:
            k = ''

        log.msg("%s%s%s(%s)" % (
            self.callDepth * ' ',
            k,
            frame.f_code.co_name,
            formatArgs(extractArgs(frame))))

    def trace_RETURN(self, frame, arg):
        if arg is not None:
            log.msg("%s<= %s" % (
                self.callDepth * ' ',
                reflect.safe_repr(arg),))
        self.callDepth = max(0, self.callDepth - 1)

    def trace_EXCEPTION(self, frame, arg):
        log.msg("%s^- %s" % (
            self.callDepth * ' ',
            reflect.safe_repr(arg),))
        self.callDepth = max(0, self.callDepth - 1)

import signal
class SignalService(service.Service):
    def __init__(self, sigmap):
        self.sigmap = sigmap

    def startService(self):
        service.Service.startService(self)
        self.oldsigmap = {}
        for sig, handler in self.sigmap.items():
            self.oldsigmap[sig] = signal.signal(sig, handler)

    def stopService(self):
        for sig, handler in self.oldsigmap.items():
            signal.signal(sig, handler)
        del self.oldsigmap
        service.Service.stopService(self)


def _callInD(d, f, a, k):
    try:
        r = f(*a,**k)
    except:
        r = failure.Failure()
    d.callback(r)

def deferLater(s, f, *a, **k):
    """Returns a Deferred which fires later, as the result of a timed event.
    """
    d = defer.Deferred()
    reactor.callLater(s, _callInD, d, f, a, k)
    return d


class FixedOffset(datetime.tzinfo):
    _zeroOffset = datetime.timedelta()

    def __init__(self, hours, minutes):
        self.offset = datetime.timedelta(minutes = hours * 60 + minutes)

    def utcoffset(self, dt):
        return self.offset

    def tzname(self, dt):
        return Time._timedeltaToHrMin(self.offset)

    def dst(self, tz):
        return self._zeroOffset


class Time(object):
    """An object representing a well defined instant in time.
    
    A Time object unambiguously addresses some time, independent of timezones,
    contorted base-60 counting schemes, leap seconds, and the effects of
    general relativity. It provides methods for returning a representation of
    this time in various ways that a human or a programmer might find more
    useful in various applications.

    Every Time instance has an attribute 'resolution'. This can be ignored, or
    the instance can be considered to address a span of time. This resolution
    is determined by the value used to initalize the instance, or the
    resolution of the internal representation, whichever is greater. It is
    mostly useful when using input formats that allow the specification of
    whole days or weeks. For example, ISO 8601 allows one to state a time as,
    "2005-W03", meaning "the third week of 2005". In this case the resolution
    is set to one week. Other formats are considered to express only an instant
    in time, such as a POSIX timestamp, because the resolution of the time is
    limited only by the hardware's representation of a real number.

    Date arithmetic and boolean operations operate on instants in time, not
    periods. In this case, the start of the period is used as the value, and
    the result has a resolution of 0.

    For containment tests with the 'in' operator, the period addressed by the
    instance is used.

    The methods beginning with 'from' are constructors to create instances from
    various formats. Some of them are textual formats, and others are other
    time types commonly found in Python code.

    Likewise, methods beginning with 'as' return the represented time in
    various formats. Some of these methods should try to reflect the resolution
    of the instance. However, they don't yet.

    For formats with both a constructor and a formatter, d == fromFu(d.asFu())

    @type resolution: datetime.timedelta
    @ivar resolution: the length of the period to which this instance could
    refer. For example, "Today, 13:38" could refer to any time between 13:38
    until but not including 13:39. In this case resolution would be
    timedelta(minutes=1).
    """

    # the instance variable _time is the internal representation of time. It
    # is a naive datetime object which is always UTC. A UTC tzinfo would be
    # great, if one existed, and anyway it complicates pickling.

    rfc2822Weekdays = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']
    rfc2822Months = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']

    resolution = datetime.timedelta.resolution

    #
    # Methods to create new instances
    #

    def __init__(self):
        """Return a new Time instance representing the time now.
        
        See also the fromFu methods to create new instances from other types of
        initializers.
        """
        self._time = datetime.datetime.utcnow()

    def _timedeltaToHrMin(offset):
        """Return a (sign, hour, minute) triple coresponding to the offset described by timedelta.

        sign is "+" or "-". In the case of 0 offset, sign is "+".
        """
        minutes = round((offset.days * 3600000000 * 24 + offset.seconds * 1000000 + offset.microseconds) / 60000000.0)
        if minutes < 0:
            sign = '-'
            minutes = -minutes
        else:
            sign = '+'
        return (sign, minutes // 60, minutes % 60)

    _timedeltaToHrMin = staticmethod(_timedeltaToHrMin)


    def _fromWeekday(klass, match, tzinfo, now):
        weekday = klass.weekdays.index(match.group('weekday').lower())
        dtnow = now.asDatetime(tzinfo).replace(hour=0, minute=0, second=0, microsecond=0)
        daysInFuture = (weekday - dtnow.weekday()) % len(klass.weekdays)
        if daysInFuture == 0:
            daysInFuture = 7
        self = klass.fromDatetime(dtnow + datetime.timedelta(days=daysInFuture))
        assert self.asDatetime(tzinfo).weekday() == weekday
        self.resolution = datetime.timedelta(days=1)
        return self


    def _fromTodayOrTomorrow(klass, match, tzinfo, now):
        dtnow = now.asDatetime(tzinfo).replace(hour=0, minute=0, second=0, microsecond=0)
        when = match.group(0).lower()
        if when == 'tomorrow':
            dtnow += datetime.timedelta(days=1)
        elif when == 'yesterday':
            dtnow -= datetime.timedelta(days=1)
        else:
            assert when == 'today'
        self = klass.fromDatetime(dtnow)
        self.resolution = datetime.timedelta(days=1)
        return self


    def _fromTime(klass, match, tzinfo, now):
        minute = int(match.group('minute'))
        hour = int(match.group('hour'))
        ampm = (match.group('ampm') or '').lower()
        if ampm:
            if not 1 <= hour <= 12:
                raise ValueError, 'hour %i is not in 1..12' % (hour,)
            if hour == 12 and ampm == 'am':
                hour = 0
            elif ampm == 'pm':
                hour += 12
        if not 0 <= hour <= 23:
            raise ValueError, 'hour %i is not in 0..23' % (hour,)

        dtnow = now.asDatetime(tzinfo).replace(second=0, microsecond=0)
        dtthen = dtnow.replace(hour=hour, minute=minute)
        if dtthen < dtnow:
            dtthen += datetime.timedelta(days=1)

        self = klass.fromDatetime(dtthen)
        self.resolution = datetime.timedelta(minutes=1)
        return self


    def _fromNoonOrMidnight(klass, match, tzinfo, now):
        when = match.group(0).lower()
        if when == 'noon':
            hour = 12
        else:
            assert when == 'midnight'
            hour = 0
        dtnow = now.asDatetime(tzinfo).replace(minute=0, second=0, microsecond=0)
        dtthen = dtnow.replace(hour=hour)
        if dtthen < dtnow:
            dtthen += datetime.timedelta(days=1)

        self = klass.fromDatetime(dtthen)
        self.resolution = datetime.timedelta(minutes=1)
        return self


    weekdays = ['monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday', 'sunday']

    humanlyPatterns = [
        (re.compile(r"""
            \b
            ((next|this)\s+)?
            (?P<weekday>
                monday
                | tuesday
                | wednesday
                | thursday
                | friday
                | saturday
                | sunday
            )
            \b
            """, re.IGNORECASE | re.VERBOSE),
            _fromWeekday),
        (re.compile(r"\b(today|tomorrow|yesterday)\b", re.IGNORECASE),
            _fromTodayOrTomorrow),
        (re.compile(r"""
            \b
            (?P<hour>\d{1,2}):(?P<minute>\d{2})
            (\s*(?P<ampm>am|pm))?
            \b
            """, re.IGNORECASE | re.VERBOSE),
            _fromTime),
        (re.compile(r"\b(noon|midnight)\b", re.IGNORECASE),
            _fromNoonOrMidnight),
    ]

    _fromWeekday = classmethod(_fromWeekday)
    _fromTodayOrTomorrow = classmethod(_fromTodayOrTomorrow)
    _fromTime = classmethod(_fromTime)
    _fromNoonOrMidnight = classmethod(_fromNoonOrMidnight)


    def fromHumanly(klass, humanStr, tzinfo=None, now=None):
        """Return a new Time instance from a string a human might type.

        @param humanStr: the string to be parsed.

        @param tzinfo: A tzinfo instance indicating the timezone to assume if
        none is specified in humanStr. If None, assume UTC.

        @param now: A Time instance to be considered "now" for when
        interpreting relative dates like "tomorrow". If None, use the real now.

        Total crap now, it just supports weekdays, "today" and "tomorrow" for
        now. This is pretty insufficient and useless, but good enough for some
        demo functionality, or something.
        """
        humanStr = humanStr.strip()
        if now is None:
            now = Time()
        if tzinfo is None:
            tzinfo = FixedOffset(0, 0)

        for pattern, creator in klass.humanlyPatterns:
            match = pattern.match(humanStr)
            if not match \
            or match.span()[1] != len(humanStr):
                continue
            try:
                return creator(klass, match, tzinfo, now)
            except ValueError:
                continue
        raise ValueError, 'could not parse date: %r' % (humanStr,)

    fromHumanly = classmethod(fromHumanly)


    iso8601pattern = re.compile(r"""
        ^ (?P<year> \d{4})
        (
            # a year may optionally be followed by one of:
            # - a month
            # - a week
            # - a specific day, and an optional time
            #     a specific day is one of:
            #     - a month and day
            #     - week and weekday
            #     - a day of the year
            (
                -? (?P<month1> \d{2})
                |
                -? W (?P<week1> \d{2})
                |
                (
                    -? (?P<month2> \d{2})
                    -? (?P<day> \d{2})
                    |
                    -? W (?P<week2> \d{2})
                    -? (?P<weekday> \d)
                    |
                    -? (?P<dayofyear> \d{3})
                )
                (
                    T (?P<hour> \d{2})
                    (
                        :? (?P<minute> \d{2})
                        (
                            :? (?P<second> \d{2})
                            (
                                [\.,] (?P<fractionalsec> \d+)
                            )?
                        )?
                    )?
                    (
                        (?P<zulu> Z)
                        |
                        (?P<tzhour> [+\-]\d{2})
                        (
                            :? (?P<tzmin> \d{2})
                        )?
                    )?
                )?
            )?
        )? $""", re.VERBOSE)


    def fromISO8601TimeAndDate(klass, iso8601string, tzinfo=None):
        """Return a new Time instance from a string formated as in ISO 8601.

        If the given string contains no timezone, it is assumed to be in the
        timezone specified by the parameter `tzinfo`, or UTC if tzinfo is None.
        An input string with an explicit timezone will always override tzinfo.

        If the given iso8601string does not contain all parts of the time, they
        will default to 0 in the timezone given by `tzinfo`.

        WARNING: this function is incomplete. ISO is dumb and their standards
        are not free. Only a subset of all valid ISO 8601 dates are parsed,
        because I can't find a formal description of the format. However,
        common ones should work.
        """

        def calculateTimezone():
            if groups['zulu'] == 'Z':
                return FixedOffset(0, 0)
            else:
                tzhour = groups.pop('tzhour')
                tzmin = groups.pop('tzmin')
                if tzhour is not None:
                    return FixedOffset(int(tzhour), int(tzmin or 0))
            return tzinfo or FixedOffset(0, 0)

        def coerceGroups():
            groups['month'] = groups['month1'] or groups['month2']
            groups['week'] = groups['week1'] or groups['week2']

            defaultTo0 = ['hour', 'minute', 'second'] # don't include fractionalsec, because it's not an integer.
            defaultTo1 = ['month', 'day', 'week', 'weekday', 'dayofyear']
            if groups['fractionalsec'] is None:
                groups['fractionalsec'] = '0'
            for key in defaultTo0:
                if groups[key] is None:
                    groups[key] = 0
            for key in defaultTo1:
                if groups[key] is None:
                    groups[key] = 1
            groups['fractionalsec'] = float('.'+groups['fractionalsec'])
            for key in defaultTo0 + defaultTo1 + ['year']:
                groups[key] = int(groups[key])

            for group, min, max in [
            ('week', 1, 53),            # some years have only 52
            ('weekday', 1, 7),
            ('month', 1, 12),
            ('day', 1, 31),
            ('hour', 0, 24),
            ('minute', 0, 59),
            ('second', 0, 61),          # it's possible to have *two* leapseconds. I'll kill myself when it happens, though.
            ('dayofyear', 1, 366)]:     # don't forget leap years
                if not min <= groups[group] <= max:
                    raise ValueError, '%s must be in %i..%i' % (group, min, max)

        def determineResolution():
            if match.group('fractionalsec') is not None:
                return max(datetime.timedelta.resolution,
                    datetime.timedelta( microseconds = 1 * 10 ** -len(match.group('fractionalsec')) * 1000000 ) )

            for testGroup, resolution in [
            ('second', datetime.timedelta(seconds=1)),
            ('minute', datetime.timedelta(minutes=1)),
            ('hour', datetime.timedelta(hours=1)),
            ('weekday', datetime.timedelta(days=1)),
            ('dayofyear', datetime.timedelta(days=1)),
            ('day', datetime.timedelta(days=1)),
            ('week1', datetime.timedelta(weeks=1)),
            ('week2', datetime.timedelta(weeks=1))]:
                if match.group(testGroup) is not None:
                    return resolution

            if match.group('month1') is not None \
            or match.group('month2') is not None:
                if self._time.month == 12:
                    return datetime.timedelta(days=31)
                nextMonth = self._time.replace(month=self._time.month+1)
                return nextMonth - self._time
            else:
                nextYear = self._time.replace(year=self._time.year+1)
                return nextYear - self._time

        def calculateDtime(tzinfo):
            """Calculate a datetime for the start of the addressed period."""

            if match.group('week1') is not None \
            or match.group('week2') is not None:
                if not 0 < groups['week'] <= 53:
                    raise ValueError, 'week must be in 1..53 (was %i)' % (groups['week'],)
                dtime = datetime.datetime(
                    groups['year'],
                    1,
                    4,
                    groups['hour'],
                    groups['minute'],
                    groups['second'],
                    int(round(groups['fractionalsec'] * 1000000)),
                    tzinfo=tzinfo
                )
                dtime -= datetime.timedelta(days = dtime.weekday())
                dtime += datetime.timedelta(days = (groups['week']-1) * 7 + groups['weekday'] - 1)
                if dtime.isocalendar() != (groups['year'], groups['week'], groups['weekday']):
                    # actually the problem could be an error in my logic, but
                    # nothing should cause this but requesting week 53 of a
                    # year with 52 weeks.
                    raise ValueError, 'year %04i has no week %02i' % (groups['year'], groups['week'])
                return dtime

            if match.group('dayofyear') is not None:
                dtime = datetime.datetime(
                    groups['year'],
                    1,
                    1,
                    groups['hour'],
                    groups['minute'],
                    groups['second'],
                    int(round(groups['fractionalsec'] * 1000000)),
                    tzinfo=tzinfo
                )
                dtime += datetime.timedelta(days=groups['dayofyear']-1)
                if dtime.year != groups['year']:
                    raise ValueError, 'year %04i has no day of year %03i' % (groups['year'], groups['dayofyear'])
                return dtime
            
            else:
                return datetime.datetime(
                    groups['year'],
                    groups['month'],
                    groups['day'],
                    groups['hour'],
                    groups['minute'],
                    groups['second'],
                    int(round(groups['fractionalsec'] * 1000000)),
                    tzinfo=tzinfo
                )


        match = klass.iso8601pattern.match(iso8601string)
        if match is None:
            raise ValueError, '%r could not be parsed as an ISO 8601 date and time' % (iso8601string,)

        groups = match.groupdict()
        coerceGroups()
        self = klass.fromDatetime(calculateDtime(calculateTimezone()))
        self.resolution = determineResolution()
        return self

    fromISO8601TimeAndDate = classmethod(fromISO8601TimeAndDate)

    def fromStructTime(klass, structTime, tzinfo=None):
        """Return a new Time instance from a time.struct_time.

        If tzinfo is None, structTime is in UTC. Otherwise, tzinfo is a
        datetime.tzinfo instance coresponding to the timezone in which
        structTime is.

        Many of the functions in the standard time module return these things.
        This will also work with a plain 9-tuple, for parity with the time
        module. The last three elements, or tm_wday, tm_yday, and tm_isdst are
        ignored.
        """
        dtime = datetime.datetime(tzinfo=tzinfo, *structTime[:6])
        self = klass.fromDatetime(dtime)
        self.resolution = datetime.timedelta(seconds=1)
        return self

    fromStructTime = classmethod(fromStructTime)

    def fromDatetime(klass, dtime):
        """Return a new Time instance from a datetime.datetime instance.

        If the datetime instance does not have an associated timezone, it is
        assumed to be UTC.
        """
        self = klass.__new__(klass)
        if dtime.tzinfo is not None:
            self._time = dtime.astimezone(FixedOffset(0, 0)).replace(tzinfo=None)
        else:
            self._time = dtime
        self.resolution = datetime.timedelta.resolution
        return self

    fromDatetime = classmethod(fromDatetime)

    def fromPOSIXTimestamp(klass, secs):
        """Return a new Time instance from seconds since the POSIX epoch.

        The POSIX epoch is midnight Jan 1, 1970 UTC. According to POSIX, leap
        seconds don't exist, so one UTC day is exactly 86400 seconds, even if
        it wasn't.

        `secs` can be an integer or a float.
        """
        self = klass.fromDatetime(datetime.datetime.utcfromtimestamp(secs))
        self.resolution = datetime.timedelta()
        return self

    fromPOSIXTimestamp = classmethod(fromPOSIXTimestamp)

    def fromRFC2822(klass, rfc822string):
        """Return a new Time instance from a string formated as described in RFC 2822.

        RFC 2822 specifies the format of email messages.

        Some of the obsoleted elements of the specification are not parsed
        correctly, and will raise ValueError.
        """
        import email.Utils
        date = email.Utils.parsedate_tz(rfc822string)
        if date is None:
            raise ValueError, 'could not parse RFC 2822 date %r' % (rfc822string,)
        self = klass.fromStructTime(time.gmtime(email.Utils.mktime_tz(date)))
        self.resolution = datetime.timedelta(seconds=1)
        return self

    fromRFC2822 = classmethod(fromRFC2822)

    #
    # Methods to produce various formats
    #

    def asPOSIXTimestamp(self):
        """Return this time as a timestamp as specified by POSIX.

        This timestamp is the count of the number of seconds since Midnight,
        Jan 1 1970 UTC, ignoring leap seconds.
        """
        return 60*60*24 * (self._time.toordinal()-719163) \
             + self._time.hour * 60*60 \
             + self._time.minute * 60 \
             + self._time.second \
             + self._time.microsecond * 0.000001

    def asDatetime(self, tzinfo=None):
        """Return this time as a datetime.datetime instance.

        If tzinfo is None, returns a naive UTC datetime instance. Otherwise it
        is a datetime.tzinfo instance, and the returned datetime instance is
        aware.
        """
        if tzinfo is not None:
            return self._time.replace(tzinfo=FixedOffset(0, 0)).astimezone(tzinfo)
        else:
            return self._time

    def asRFC2822(self, tzinfo=None, includeDayOfWeek=True):
        """Return this Time formatted as specified in RFC 2822.

        RFC 2822 specifies the format of email messages.

        RFC 2822 says times in email addresses should reflect the local
        timezone. If tzinfo is a datetime.tzinfo instance, the returned
        formatted string will reflect that timezone. Otherwise, the timezone
        will be '-0000', which RFC 2822 defines as UTC, but with an unknown
        local timezone.

        RFC 2822 states that the weekday is optional. The parameter
        includeDayOfWeek indicates whether or not to include it.
        """
        dtime = self.asDatetime(tzinfo)

        offset = dtime.utcoffset()
        if offset is None:
            rfcoffset = '-0000'
        else:
            rfcoffset = '%s%02i%02i' % self._timedeltaToHrMin(offset)

        rfcstring = ''
        if includeDayOfWeek:
            rfcstring += self.rfc2822Weekdays[dtime.weekday()] + ', '

        rfcstring += '%i %s %4i %02i:%02i:%02i %s' % (
            dtime.day,
            self.rfc2822Months[dtime.month - 1],
            dtime.year,
            dtime.hour,
            dtime.minute,
            dtime.second,
            rfcoffset)

        return rfcstring

    def asISO8601TimeAndDate(self, includeDelimiters=True, tzinfo=None, includeTimezone=True):
        """Return this time formatted as specified by ISO 8861.
        
        ISO 8601 allows optional dashes to delimit dates and colons to delimit
        times. The parameter includeDelimiters (default True) defines the
        inclusion of these delimiters in the output.

        If tzinfo is a datetime.tzinfo instance, the output time will be in the
        timezone given. If it is None (the default), then the timezone string
        will not be included in the output, and the time will be in UTC.

        The includeTimezone parameter coresponds to the inclusion of an
        explicit timezone. The default is True.
        """
        dtime = self.asDatetime(tzinfo)

        if includeDelimiters:
            dateSep = '-'
            timeSep = ':'
        else:
            dateSep = timeSep = ''

        date = '%04i%s%02i%s%02i' % (dtime.year, dateSep, dtime.month, dateSep, dtime.day)
        time = '%02i%s%02i%s%02i' % (dtime.hour, timeSep, dtime.minute, timeSep, dtime.second)
        dateandtime = '%sT%s' % (date, time)
        if dtime.microsecond:
            dateandtime += ('.%06i' % (dtime.microsecond,)).rstrip('0')

        if includeTimezone:
            if tzinfo is None:
                dateandtime += '+00%s00' % (timeSep,)
            else:
                sign, hour, min = self._timedeltaToHrMin(dtime.utcoffset())
                dateandtime += '%s%02i%s%02i' % (sign, hour, timeSep, min)

        return dateandtime

    def asStructTime(self, tzinfo=None):
        """Return this time represented as a time.struct_time.
        
        tzinfo is a datetime.tzinfo instance coresponding to the desired
        timezone of the output. If is is the default None, UTC is assumed.
        """
        dtime = self.asDatetime(tzinfo)
        if tzinfo is None:
            return dtime.utctimetuple()
        else:
            return dtime.timetuple()

    def asHumanly(self, tzinfo=None, now=None):
        """Return this time as a short string, tailored to the current time.

        Parts of the date that can be assumed are omitted. Consequently, the
        output string depends on the current time. This is the format used for
        displaying dates in most user visible places in the quotient web UI.

        By default, the current time is determined by the system clock. The
        current time used for formatting the time can be changed by providing a
        Time instance as the parameter 'now'.
        """
        if now is None:
            now = Time().asDatetime(tzinfo)
        else:
            now = now.asDatetime(tzinfo)
        dtime = self.asDatetime(tzinfo)

        if dtime.date() == now.date():
            if self.isAllDay():
                return 'all day'
            return dtime.strftime('%I:%M %p').lower()
        elif dtime.date().year == now.date().year:
            res = dtime.strftime('%b, %I:%M %p').lower().capitalize()
            res = str(dtime.date().day) + ' ' + res
            return res
        else:
            res = dtime.strftime('%b %y, %I:%M %p').lower().capitalize()
            res = str(dtime.date().day) + ' ' + res
            return res

    #
    # methods to return related times
    #

    def oneDay(self, tzinfo=None):
        """Return a Time instance representing the day of the start of self.
        
        The returned new instance will be set to midnight of the day containing
        the first instant of self in the specified timezone, and have a
        resolution of datetime.timedelta(days=1).
        """
        day = self.__class__.fromDatetime(self.asDatetime(tzinfo).replace(hour=0, minute=0, second=0, microsecond=0))
        day.resolution = datetime.timedelta(days=1)
        return day

    def isAllDay(self, tzinfo=None):
        """Return True iff this instance represents exactly all day.
        
        This is somewhat dumb currently, because if a user inputs a day
        somewhere, we interpret them a bit to literally, because the resulting
        Time instance addresses exactly that day in the timezone specified.
        What this ultimately means in the UI is that if a user sets the
        timezone to EST, then creates an event for "2005-123", it will be
        rendered as "all day" only as long as the user's timezone remains EST.
        """
        dtime = self.asDatetime(tzinfo)

        return dtime.hour == dtime.minute == dtime.second == 0 \
        and self.resolution == datetime.timedelta(days=1)

    #
    # other magic methods
    #

    def __cmp__(self, other):
        if not isinstance(other, Time):
            raise TypeError("Cannot meaningfully compare %r with %r" % (self, other))
        return cmp(self._time, other._time)

    def __eq__(self, other):
        if isinstance(other, Time):
            return cmp(self._time, other._time) == 0
        return False

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return 'tpython.Time.fromDatetime(%r)' % (self._time,)

    __str__ = asISO8601TimeAndDate

    def __contains__(self, other):
        """Test if another Time instance is entirely within the period addressed by this one."""
        if not isinstance(other, Time):
            raise TypeError, '%r is not a Time instance; can not test for containment' % (other,)
        if other._time < self._time:
            return False
        if self._time + self.resolution < other._time + other.resolution:
            return False
        return True

    def __add__(self, addend):
        if not isinstance(addend, datetime.timedelta):
            raise TypeError, 'expected a datetime.timedelta instance'
        return Time.fromDatetime(self._time + addend)

    def __sub__(self, subtrahend):
        if not isinstance(subtrahend, datetime.timedelta):
            raise TypeError, 'expected a datetime.timedelta instance'
        return Time.fromDatetime(self._time - subtrahend)


from twisted.python.reflect import allYourBase

class NamedProperty(object):
    """A property-alike which knows the name by which it goes.

    For example:

        | class MagicHat(NamedProperty):
        |     def getter(self, oself, value):
        |         print 'MagicHat getting', self.getName()
        |         return 'hat'
        |     def setter(self, oself, value):
        |         print 'MagicHat setting', self.getName(), 'to', value
        |         return value
        |
        | class Foo(object):
        |     bar = MagicHat()
        |
        | f = Foo()
        | f.bar = \"zoom\"
        | x = f.bar
        | print x

    will result in the output:

        MagicHat setting bar to zoom
        MagicHat getting bar
        hat

    """

    def __init__(self, prefix='_named_', underlying=None):
        self.name = None
        self.prefix = prefix
        self.underlying = underlying

    def getName(self, oself):
        if self.name is not None:
            return self.name
        classObj = oself.__class__
        for base in allYourBase(classObj) + [classObj]:
            for k, v in base.__dict__.items():
                if v is self:
                    self.name = k
                    self.classObj = base
                    if self.underlying is None:
                        self.underlying = self.prefix + self.name
                    return self.name
        raise ValueError()

    def __get__(self, oself, type=None):
        if oself is None:
            if oself is None:
                return self
        name = self.getName(oself)
        return self.getter(oself, name, getattr(oself, self.underlying))

    def __set__(self, oself, value):
        name = self.getName(oself)
        return setattr(oself, self.underlying, self.setter(oself, name, value))

    def getter(self, oself, name, value):
        print '%s.%s ---> %s' % (oself, name, value)
        return value

    def setter(self, oself, name, value):
        print '%s.%s <--- %s' % (oself, name, value)
        return value

def dsu(L, F):
    """Zero-copy implementation of Decorate/Sort/Undecorate design pattern.

    @param L: A list to be sorted.

    @param F: A decorator function, which takes one argument, an item in the
    list, and produces a sortable value.
    """

    for n, o in enumerate(L):
        L[n] = F(o), n, o
    L.sort()
    for n2, (Fo, n, o) in enumerate(L):
        L[n2] = o
    return L

##
## Monkey patches to improve error reporting of hanging LoopingCalls
## This can be removed when support for Twisted 1.3 is dropped
##
from twisted.internet.base import DelayedCall
from twisted.python import reflect

def __str__(self):
    if hasattr(self.func, 'func_name'):
        func = self.func.func_name
        if hasattr(self.func, 'im_class'):
            func = self.func.im_class.__name__ + '.' + func
    else:
        func = reflect.safe_repr(self.func)

    L = ["<DelayedCall %s [%ss] called=%s cancelled=%s %s%s" % (
            id(self), self.time - seconds(), self.called, self.cancelled, func,
            reflect.safe_repr(self.args))]

    if self.debug:
        L.append("\n\ntraceback at creation: \n\n%s" % ('    '.join(self.creator)))
    L.append('>')

    return "".join(L)
DelayedCall.__str__ = __str__
del __str__

from twisted.internet.task import LoopingCall
from twisted.python.runtime import seconds
from twisted.internet import defer

def start(self, interval, now=True):
    """Start running function every interval seconds.

    @param interval: The number of seconds between calls.  May be less than
    one.  Precision will depend on the underlying platform, the available
    hardware, and the load on the system.

    @param now: If True, run this call right now.  Otherwise, wait until the
    interval has elapsed before beginning.

    @return: A Deferred whose callback will be invoked with C{self} when
    C{self.stop} is called, or whose errback will be invoked if the function
    raises an exception.
    """
    assert not self.running
    if interval < 0:
        raise ValueError, "interval must be >= 0"
    self.running = True
    d = self.deferred = defer.Deferred()
    self.starttime = seconds()
    self.count = 0
    self.interval = interval
    if now:
        self()
    else:
        self._reschedule()
    return d

def __call__(self):
    self.call = None
    try:
        self.f(*self.a, **self.kw)
    except:
        self.running = False
        d, self.deferred = self.deferred, None
        d.errback()
    else:
        self._reschedule()

def _reschedule(self):
    if self.interval == 0:
        self.call = reactor.callLater(0, self)
        return

    fromNow = self.starttime - seconds()

    while self.running:
        self.count += 1
        fromStart = self.count * self.interval
        delay = fromNow + fromStart
        if delay > 0:
            self.call = reactor.callLater(delay, self)
            return

def __repr__(self):
    if hasattr(self.f, 'func_name'):
        func = self.f.func_name
        if hasattr(self.f, 'im_class'):
            func = self.f.im_class.__name__ + '.' + func
    else:
        func = reflect.safe_repr(self.f)

    return 'LoopingCall<%r>(%s, *%s, **%s)' % (
        self.interval, func, reflect.safe_repr(self.a),
        reflect.safe_repr(self.kw))

LoopingCall.start = start
LoopingCall.__call__ = __call__
LoopingCall._reschedule = _reschedule
LoopingCall.__repr__ = __repr__

del start, __call__, _reschedule, __repr__
