# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

"""Definitions of Atop upgraders.

Functions defined here should be registered as upgraders with the
registerUpgrader function.  For example::

    | def fooTo3(fooInert):
    |     fooInert.state['x'] = 'y'
    | registerUpgrader('atop.foo.Foo', 3, {'atop.foo.Bar': 2}, fooTo3)
"""

from atop.versioning import registerUpgrader, named, _NamedTypeProxy
from atop import store

def attrToSingleIndexValue(inert, attrName, indexObj):
    """Utility to help remove index_* methods.

    Example:

    | class C:
    |     def index_fu(self):
    |         return self.fu

    becomes:

    | class FuIndex(StringIndex)
    | class C:
    |     fu = indexed(FuIndex)

    and the upgrader calls:

    | attrToSingleIndexValue(inertC, 'fu', FuIndex)

    @param inert: the inert instance being upgraded
    @param attrName: name of the attribute being replaced
    @param indexObj: the index class associated with the attribute
    """
    value = inert.state.pop(attrName)
    inert.state['_indexes'][indexObj] = value


def itemTo1(inertItem):
    # don't use attrToSingleIndexValue here because changes to store.Item were
    # deployed but the upgrader was not. There are two cases that must be handled:
    #
    # 1) Item created before change, had index_received at creation. Item does
    # not have a value for DateIndex in _indexes dict, but does have a
    # dateCreated attribute.
    #
    # 2) Item created after change. Item has no dateCreated attribute, but does
    # have an entry in _indexes
    #
    # Case 2 shouldn't exist because those changes were accidentally deployed.
    # The solution here is to do nothing in case 2, but assert that DateIndex
    # is in _indexes. If an object was created in case 1, then unpickled with
    # the changed code, then DateIndexed changed, badness could result, but
    # it's very unlikely that could happen and the consequences are just an
    # incorrect date index, so don't worry about it here.

    try:
        value = inertItem.state.pop('dateCreated')
    except KeyError:
        assert named('atop.store.DateIndex') in inertItem.state['_indexes']
    else:
        inertItem.state['_indexes'][named('atop.store.DateIndex')] = value

registerUpgrader('atop.store.Item', 1, {}, itemTo1)


def itemTo2(inertItem):
    """This upgrade changes dateCreated from a 9-tuple to a tpython.Time instance."""
    from atop import tpython
    dateIndex = named('atop.store.DateIndex')
    inertItem.state['_indexes'][dateIndex] = tpython.Time.fromStructTime(inertItem.state['_indexes'][dateIndex].newargs[0])

registerUpgrader('atop.store.Item', 2, {}, itemTo2)

##
## atop.sched upgraders
##

from twisted.python.components import Interface, registerAdapter
from twisted.application.service import IServiceCollection
from twisted.persisted.styles import Ephemeral

from atop import sched

class ISystemSchedulerUpgrade(Interface):
    def fixServiceParent(self):
        pass

class SystemSchedulerUpgradeAdapter(Ephemeral):
    def __init__(self, sysSched):
        self.sysSched = sysSched

    def fixServiceParent(self):
        parent = IServiceCollection(self.sysSched.store)
        parent.addService(self.sysSched)

registerAdapter(SystemSchedulerUpgradeAdapter,
                sched.SystemScheduler,
                ISystemSchedulerUpgrade)

def systemSchedulerTo1(schedInert):
    schedInert.callAfterUpgrade(ISystemSchedulerUpgrade, 'fixServiceParent')

registerUpgrader('atop.sched.SystemScheduler', 1, {}, systemSchedulerTo1)

##
## SubStore upgraders
##

def subStoreTo1(subStoreInert):
    oldAdapterCache = subStoreInert.state['_adapterCache']
    newAdapterCache = subStoreInert.state['_adapterCache'] = {}
    for k, v in oldAdapterCache.items():
        if isinstance(v, _NamedTypeProxy): # and v.inertName.endswith('._NoImplementor'):
            continue
        newAdapterCache[named(k)] = v

registerUpgrader('atop.store.SubStore', 1, {}, subStoreTo1)

##
## Domain upgraders
##

from atop import credup

class IDomainUpgrade(Interface):
    def resetName(self):
        pass

class DomainUpgradeAdapter(Ephemeral):
    def __init__(self, domain):
        self.domain = domain

    def resetName(self):
        self.domain.name = self.domain.name

registerAdapter(DomainUpgradeAdapter, credup.Domain, IDomainUpgrade)

def domainTo1(inertDomain):
    """
    NameIndex was replaced with DomainIndex, a case insensitive flavor
    which also sorts hierarchically.
    """
    didx = named('atop.credup.DomainIndex')
    nidx = named('atop.store.NameIndex')
    inertDomain.state['_indexes'][didx] = inertDomain.state['_indexes'].pop(nidx)
    inertDomain.callAfterUpgrade(IDomainUpgrade, 'resetName')
    inertDomain.callAfterUpgrade(store.IPool, 'addIndex', credup.UsernameIndex)

registerUpgrader('atop.credup.Domain', 1, {}, domainTo1)

##
## CredentialsChecker upgraders
##

class ICheckerUpgrade(Interface):
    def recreateDomainPool(self):
        """
        Create a new C{DomainPool} to store domains in and move all of
        the domains in the old C{Pool} used for this purpose into it.
        """

class CheckerUpgradeAdapter(Ephemeral):
    def __init__(self, checker):
        self.checker = checker

    def recreateDomainPool(self):
        oldPool = self.checker.domains
        self.checker.domains = credup.DomainPool(self.checker.store, 'domains')
        for d in oldPool.queryIndex():
            self.checker.domains.addItem(d)
            oldPool.removeItem(d)
        oldPool.deleteFromStore()

registerAdapter(CheckerUpgradeAdapter, credup.StoreCredentialsChecker, ICheckerUpgrade)

def storeCredentialsCheckerTo1(inertChecker):
    inertChecker.state['_domainsRef'] = inertChecker.state.pop('domainPoolRef')

    inertChecker.callAfterUpgrade(ICheckerUpgrade, 'recreateDomainPool')

registerUpgrader(
    'atop.credup.StoreCredentialsChecker', 1, {}, storeCredentialsCheckerTo1)
