# -*- test-case-name: atop.test.test_upgrade -*-

from twisted.python import log
from twisted.python import failure
from twisted.internet import defer

from atop.store import Item, Pool, IPipe, referenced, INDEX, UIDIndex, _deprecatedIndexLookup, _sim
from atop import sched

class DeleteMe(Item):
    """I am the simplest kind of upgrade.  Use this class when you have made a
    mistake and made something persistent which you should never have stored
    anywhere, in memory or otherwise.  Use it like this:

        Your old code looks like this::

            # dumb.py
            from atop.store import Item
            class Stupid(Item):
                # do stupid stuff

        Your new code should look like this
            # dumb.py
            from atop.store import DeleteMe
            Stupid = DeleteMe

    If you are removing a subclass of Pool, make sure to use DeleteMePool
    instead.  There is currently no version of DeleteMe which works with
    Substores.
    """

    def activate(self):
        self.deleteFromStore()

class DeleteMePool(Pool):
    """A version of DeleteMe that works with pools.
    """
    def activate(self):
        self.deleteFromStore()

class QueueErrorPool(Pool):
    """Pool containing FailedTasks.

    Items whose process method raises an exception will end
    up in this pool, wrapped FailedTask Items.
    """

class QueuedTaskError(Exception):
    """Exception indicating one or more Items failed processing.

    @ivar queue: The Queue which had failing Items.
    """
    def __init__(self, pool):
        Exception.__init__(self, pool)
        self.queue = pool

class FailedTask(Item):
    """A wrapper around Items whose process method raises an exception.

    @ivar item: The Item which failed.
    @ivar reason: A failure created from the exception.
    """
    item = referenced

    def __init__(self, store, item, reason):
        Item.__init__(self, store)
        self.item = item
        self.reason = reason

class UpgradeBase(Item, sched.Schedulable):
    # Whether or not start() has been called on this instance yet
    startedUpgrade = False

    # Whether or not every item has been examined
    finishedUpgrade = False

    # Whether or not we have fired our deferred and deleted ourself
    finishedCompletely = False

    # A pool containing error information relating to this upgrade
    errors = referenced

    # A reminder as to whether or not any errors have occurred during this upgrade
    failures = False

    # A deferred we will fire when we complete the upgrade
    deferred = None

    # Indicates we are waiting to fire before proceeding with the upgrade
    waitingOnDeferred = False

    # The last integer percent of completeness we had
    lastPercent = 0

    # The callable we will invoke on each Item we traverse
    upgradeFactory = None

    def pickDelay(self, howMany):
        return 0

    def start(self):
        assert self.startedUpgrade == False, "Can only start an Upgrader once"
        self.touch()
        self.startedUpgrade = True
        self.deferred = defer.Deferred()
        self.reset(0)
        return self.deferred

    def calculatePercent(self):
        raise NotImplementedError()

    def nextItem(self):
        raise NotImplementedError()

    def _oneUpgrade(self):
        raise NotImplementedError()

    def run(self):
        self.touch()
        count = self.store.batch(self._oneUpgrade)
        if self.finishedUpgrade:
            self.finish()
        else:
            newPercent = self.calculatePercent()
            if int(newPercent) != self.lastPercent:
                self.lastPercent = int(newPercent)
                log.msg("%s %d%% complete." % (self, self.lastPercent))
            if not self.waitingOnDeferred:
                self.reset(self.pickDelay(count))

    def finish(self):
        self.touch()
        self.unschedule()
        if self.finishedCompletely:
            # XXX What the heck?
            log.msg("Warning: double-finished %s" % (self,))
            return
        d, self.deferred = self.deferred, None
        self.finishedCompletely = True
        log.msg("Finished %s" % (self,))
        if self.failures:
            d.errback(QueuedTaskError(self))
        else:
            upgradeFactory = self.upgradeFactory
            self.errors.deleteFromStore()
            self.deleteFromStore()
            d.callback(upgradeFactory)

    def fail(self, item, failure):
        self.failures = True
        log.err(failure)
        failure.cleanFailure()
        self.errors.addItem(FailedTask(self.store, item, failure))


class UpgradeQueue(UpgradeBase):
    pool = referenced

    def __init__(self, store, name, pool, upgradeFactory, *upgradeArgs, **upgradeKwArgs):
        self.pool = pool
        self.upgradeFactory = upgradeFactory
        self.errors = QueueErrorPool(store)
        self.upgradeArgs = upgradeArgs
        self.upgradeKwArgs = upgradeKwArgs

        UpgradeBase.__init__(self, store, name)

        self.currentUID = None
        try:
            self.maxUID = pool.queryIndex(backwards=True, result=INDEX).next()
        except StopIteration:
            self.maxUID = -1

    def _oneUpgrade(self):
        item = self.nextItem()
        if item is not None and item.getIndexKey(self.pool, UIDIndex) <= self.maxUID:
            try:
                self.upgradeFactory(self.pool, item, *self.upgradeArgs, **self.upgradeKwArgs)
            except:
                self.fail(item, failure.Failure())
        else:
            self.finishedUpgrade = True
            return True

    def calculatePercent(self):
        return 100.0 * self.currentUID / (self.maxUID + 1)

    def nextItem(self):
        q = self.pool.queryIndex(startKey=self.currentUID)
        try:
            item = q.next()
        except StopIteration:
            return None
        self.currentUID = item.getIndexKey(self.pool, UIDIndex) + 1
        return item

class BackwardsUpgradeQueue(UpgradeBase):
    pool = referenced

    def __init__(self, store, name, pool, upgradeFactory, *upgradeArgs, **upgradeKwArgs):
        self.pool = pool
        self.upgradeFactory = upgradeFactory
        self.errors = QueueErrorPool(store)
        self.upgradeArgs = upgradeArgs
        self.upgradeKwArgs = upgradeKwArgs

        UpgradeBase.__init__(self, store, name)

        try:
            self.firstUID = self.currentUID = pool.queryIndex(backwards=True, result=INDEX).next()
        except StopIteration:
            self.firstUID = self.currentUID = -1

    def _oneUpgrade(self):
        item = self.nextItem()
        if item is not None:
            self.currentUID = item.getIndexKey(self.pool, UIDIndex) - 1
            try:
                self.upgradeFactory(self.pool, item, *self.upgradeArgs, **self.upgradeKwArgs)
            except:
                self.fail(item, failure.Failure())
        else:
            self.finishedUpgrade = True
            return True

    def calculatePercent(self):
        return 100.0 - (100.0 * self.currentUID / self.firstUID)

    def nextItem(self):
        q = self.pool.queryIndex(startKey=self.currentUID, backwards=True)
        try:
            return q.next()
        except StopIteration:
            return None


class ItemUpgradeFactoryUpgradeQueue(UpgradeQueue):
    upgradeFactory = referenced

default = object()
class NewItemIndex:
    def __init__(self, index):
        self.index = _deprecatedIndexLookup(index)

    def __call__(self, pool, item):
        newIndexKey = item.getIndexKey(pool, self.index, default)
        if newIndexKey is not default:
            pool.changeKeyForItem(item, self.index, None, newIndexKey, notify=False)

    def __str__(self):
        return "%s(%s)" % (self.__class__.__name__, self.index,)

def upgradeIndex(pool, index):
    """Add keys to an index newly added to a Pool.

    This should be called immediately after C{pool.addIndex(index)}
    and before any new items are added to the pool.
    """
    index = _deprecatedIndexLookup(index)
    name = '(Adding index %r to pool %r [%r] for user %r)' % (
        index.__name__, pool.name, pool.getItemPath(),
        pool.store.name)
    return UpgradeQueue(pool.store, name, pool,
                        NewItemIndex(index)).start()

class DelItemIndex(NewItemIndex):
    def __call__(self, pool, item):
        pool.changeKeyForItem(item, self.index,
                              item.getIndexKey(pool,
                                               self.index),
                              None)

def deleteIndex(pool, index):
    name = '(Deleting index %r from pool %r [%r] for user %r)' % (
        index.__name__, pool.name, pool.getItemPath(),
        pool.store.name)
    return UpgradeQueue(pool.store, name, pool,
                        DelItemIndex(index)).start()

def noop(pool, item):
    pass

def versionedUpgrade(pool):
    """Gradually load all Items from a Pool.

    This is a helper for L{twisted.persisted.styles.Versioned}.
    Object upgrade methods may have side effects which it is
    desirable to see sooner rather than later.  This function
    causes all Items in a pool to be deserialized, forcing their
    upgrade methods to run, so that those side effects may be
    seen.
    """
    name = '(Version upgrade of items in pool %r [%r] for user %r)' % (
        pool.name, pool.getItemPath(), pool.store.name)
    return UpgradeQueue(pool.store, name, pool, noop).start()

class IndexCount(Item):
    __implements__ = IPipe
    pool = referenced

    currentUID = 0

    def __init__(self, pool):
        Item.__init__(self, pool.store, '(%s)' % pool.name)
        self.pool = pool
        self.pool.addPipe(self.fwd(IPipe))
        self.counts = dict.fromkeys(pool._poolIndexes, 0)

    def deleteFromStore(self):
        self.pool.removePipe(self.fwd(IPipe))
        return Item.deleteFromStore(self)

    def __call__(self, pool, item):
        self.touch()
        self.currentUID = item.poolToUID[pool] + 1
        for index in self.counts:
            if item.getIndexKey(self.pool, index) is not None:
                self.counts[index] += 1

    def itemAdded(self, item):
        # We iterate in UID order.  Nothing will ever be added
        # that we won't proceed to iterate over later.
        pass

    def itemRemoved(self, item):
        if item.getIndexKey(self.pool, UIDIndex) < self.currentUID:
            # We counted it already.  Reverse that.
            for index in self.counts:
                if item.getIndexKey(self.pool, index) is not None:
                    self.counts[index] -= 1

    def itemChanged(self, item, changeData):
        pass

def cbTransferCounts(counter):
    pool = counter.pool
    def _():
        pool.removePipe(counter.fwd(IPipe))
        for index in counter.counts:
            k = pool._formatIndexCountKey(index)
            pool.setDataKey(k, str(counter.counts[index]))
    pool.store.transact(_)
    return counter

def ebTransferCounts(failure):
    cbTransferCounts(failure.queue)
    log.err(failure)
    return failure

def recountIndexes(pool):
    """Helper for L{store.Pool.upgradeToVersion1}.

    Counts how many Items belong to each index on the given
    pool.
    """
    name = '(Recounting indexes on pool %r [%r] for user %r)' % (
        pool.name, pool.getItemPath(), pool.store.name)
    ic = IndexCount(pool)
    d = ItemUpgradeFactoryUpgradeQueue(pool.store, name,
                                       pool, ic).start()
    d.addCallbacks(cbTransferCounts, ebTransferCounts)
    return d

def _poolOperation(name, pool, op, backwards, *a, **kw):
    if backwards:
        f = BackwardsUpgradeQueue
    else:
        f = UpgradeQueue
    return f(pool.store, name, pool, op, *a, **kw).start()

def removeItems(pool, item):
    pool.removeItem(item)

def emptyPool(pool, backwards=False):
    """Removes all items from the given Pool.

    If backwards is True, start at the end and work towards
    the beginning.  Items added to the Pool after this function
    is called will be ignored.
    """
    name = '(Removing items from pool %r [%r]%s for user %r)' % (
        pool.name, pool.getItemPath(),
        backwards and ' backwards' or '',
        pool.store.name)
    return _poolOperation(name, pool, removeItems, backwards)

def deleteItems(pool, item):
    return item.deleteFromStore()

def demolishPool(pool, backwards=False):
    """Delete all items in the given Pool.

    If backwards is True, start at the end and work towards
    the beginning.  Items added to the Pool after this function
    is called will be ignored.
    """
    name = '(Deleting contents of pool %r [%r]%s for user %r)' % (
        pool.name, pool.getItemPath(),
        backwards and ' backwards' or '',
        pool.store.name)
    return _poolOperation(name, pool, deleteItems, backwards)

def moveItems(oldPool, item, newPoolRef):
    newPoolRef.getItem().addItem(item)
    oldPool.removeItem(item)

def relocatePoolContents(sourcePool, targetPool, backwards=False):
    name = '(Moving contents of pool %r [%r] into pool %r [%r]%s for user %r)' % (
        sourcePool.name, sourcePool.getItemPath(),
        targetPool.name, targetPool.getItemPath(),
        backwards and ' (backwards)' or '',
        sourcePool.store.name)
    return _poolOperation(name, sourcePool, moveItems,
                          backwards, targetPool.referenceTo())

class SubStoreItem(Item):
    def __init__(self, item):
        Item.__init__(self, item.store.store)
        self.subItemID = item.storeID

    def getItem(self):
        return self.store.getItemByID(self.subItemID)

class SubStoreDeletion(UpgradeBase):
    def __init__(self, substore):
        name = '(Deleting user %r [%r])' % (substore.name, substore.getItemPath())
        UpgradeBase.__init__(self, substore.store, name=name)
        self.errors = QueueErrorPool(self.store)

        # The store ID of the SubStore we are demolishing
        self.subStoreID = substore.storeID

        # The store ID of the Item we will next delete
        self.currentStoreID = 0

        # The store ID of the last Item that must be deleted
        self.maxStoreID = substore.nextID(allocate=False) - 1

        # Items we will save 'til the end to delete
        scheduler = sched.IScheduler(substore)
        self.skip = [scheduler.storeID, scheduler.pool.storeID, 2]

    def _oneUpgrade(self):
        item = self.nextItem()
        if item is not None:
            try:
                d = item.deleteFromStore()
            except:
                self.fail(SubStoreItem(item), failure.Failure())
            else:
                if d is not None:
                    self.waitingOnDeferred = True
                    d.addCallback(*_sim(self._cbDelete))
                    d.addErrback(*_sim(self._ebDelete, item.storeID))
                    return True
        else:
            self.finishedUpgrade = True
            return True


    def _cbDelete(self, result):
        def _():
            self.touch()
            self.waitingOnDeferred = False
            self.reset(0)
        self.store.transact(_)


    def _ebDelete(self, failure, itemID):
        def _():
            self.touch()
            substore = self.store.getItemByID(self.subStoreID)
            item = substore.getItemByID(itemID)
            self.waitingOnDeferred = False
            self.fail(SubStoreItem(item), failure)
            self.reset(0)
        self.store.transact(_)


    def calculatePercent(self):
        return 100.0 * self.currentStoreID / self.maxStoreID


    def nextItem(self):
        substore = self.store.getItemByID(self.subStoreID)
        while self.currentStoreID <= self.maxStoreID:
            current = self.currentStoreID
            self.currentStoreID += 1
            try:
                item = substore.getItemByID(current)
            except KeyError:
                pass
            except:
                log.msg("Error loading %d from %s:" % (current, substore.getItemPath()))
                log.err()
            else:
                if item.storeID not in self.skip:
                    return item

    def start(self):
        return UpgradeBase.start(self
            ).addCallback(self._cbDeleteFinished, self.subStoreID, self.skip
            )

    def _cbDeleteFinished(self, result, subStoreID, skip):
        substore = self.store.getItemByID(subStoreID)
        for extra in skip:
            try:
                substore.getItemByID(extra).deleteFromStore()
            except KeyError:
                pass
        Item.deleteFromStore(substore)
        return result
