# -*- test-case-name: atop.test.test_util -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import os, glob, sets, sys
import os.path
from os.path import join as opj
from bsddb import db as libdb

from atop.store import IInvertedIndexStorage, IPool
from atop.store import Inconsistency, UIDIndex, _deprecatedIndexLookup, Store
from atop.store import ITEM, STORE_ID, INDEX, INDEX_AND_ITEM, INDEX_AND_STORE_ID
from atop.store import DatabaseEnvironment
from atop import credup
from atop.ifcompat import implements, backwardsCompatImplements

from twisted.internet import defer
from twisted.python import context, log

class BackupError(Exception):
    pass

class BackupDirectoryExists(BackupError):
    pass

class CannotCreateBackupDirectory(BackupError):
    pass

def _copyFile(iname, oname, pageSize):
    # XXX - Should these be opened with O_DIRECT?
    destDirName = os.path.dirname(oname)
    if not os.path.isdir(destDirName):
        os.makedirs(destDirName)
    fIn = file(iname, 'rb')
    fOut = file(oname, 'wb')

    for chunk in iter(lambda: fIn.read(pageSize), ''):
        fOut.write(chunk)
        fOut.flush()

    fIn.close()
    fOut.close()

def _copyTree(src, dst):
    a, b = os.path.split(src)
    os.makedirs(opj(dst, b))
    for dirpath, dirnames, filenames in os.walk(src):
        destPath = dirpath[len(a)+1:]
        for fname in filenames:
            i = opj(dirpath, fname)
            o = opj(dst, destPath, fname)
            _copyFile(i, o, 2 ** 16)

def _makeVerbose(obj, files=[]):
    # Only our super special modified bsddb module
    # exports the functions this function depends on.
    if not hasattr(obj, 'set_verbose'):
        return

    for verb in (libdb.DB_VERB_CHKPOINT, libdb.DB_VERB_DEADLOCK,
                 libdb.DB_VERB_RECOVERY, libdb.DB_VERB_REPLICATION,
                 libdb.DB_VERB_WAITSFOR):
        obj.set_verbose(verb, 1)
    files.append(file('errfile', 'w'))
    obj.set_errfile(files[-1])

##
## Functions for hot-copying a bsddb
##

def _cleanup(db, env, lock):
    for (o, m) in ((db, 0), (env, 0), (lock, 'unlock')):
        try:
            getattr(o, m or 'close')()
        except:
            log.err()

def tidy(datadir):
    """Recover a corrupt database.

    This is intended for use on databases and associated
    transaction logs hot copied from a running server (as
    done, for example, by L{backupDatabase}.
    """
    dbenv = DatabaseEnvironment()
    dbenv.normalRecoveryFlags = dbenv.fatalRecoveryFlags
    env, db, lock = dbenv.open(datadir)
    _cleanup(db, env, lock)

# XXX
# Nothing calls this yet
def verify(datadir):
    """Invoke the underlying database verification system.

    This must only be used on environments which are not in
    use by any other process.
    """
    dbenv = DatabaseEnvironment()
    dbenv.envOpenFlags = libdb.DB_INIT_MPOOL | libdb.DB_CREATE
    env, db, lock = dbenv.open(datadir)
    try:
        db.verify(opj(datadir, 'db'))
    finally:
        _cleanup(db, env, lock)

#def determinePageSize(datadir):
#    # Join the environment to discover the block size
#    env, db = envdbjoin(datadir)
#    try:
#        pageSize = db.stat()['pagesize']
#    finally:
#        db.close()
#        env.close()
#    return pageSize

# The following definition of `determinePageSize' allows users with no
# write access to a database make a backup of it.  If you use this function,
# you must ensure the page size being returned is correct, or the backup
# will contain errors.
#
def determinePageSize(datadir):
    return 1024 * 4

def backupDatabase(datadir, backupdir):
    """Create a hot copy of an open database.

    @param datadir: The path to the database file.

    @param backupdir: The path to which to save the backups.
    This directory should not already exist.
    """
    if os.path.exists(backupdir):
        raise BackupDirectoryExists(backupdir)
    try:
        os.makedirs(backupdir)
    except OSError, e:
        raise CannotCreateBackupDirectory(e)

    pageSize = determinePageSize(datadir)

    # Copy the database file itself, pageSize bytes at a time
    _copyFile(opj(datadir, 'db'), opj(backupdir, 'db'), pageSize)

    # Next copy all the logfiles present
    for log in glob.glob(opj(datadir, 'log.*')):
        origDir, origFile = os.path.split(log)
        destPath = opj(backupdir, origFile)
        _copyFile(log, destPath, pageSize)

##
## Functions for collecting non-db data for users
##

def topFileFor(avatar):
    return avatar.nextID(allocate=False)

def extractAllocations(st):
    r = {}
    cc = credup.ICredentialsChecker(st)
    for av in cc.getAvatars():
        r[av.storeID] = topFileFor(av)
    return r

def formatAllocations(allocs):
    k = allocs.keys()
    k.sort()
    L = []
    for id in k:
        L.append('%d %d' % (id, allocs[id]))
    return '\n'.join(L)

def collectRequiredFiles(filepath, avatarID, topItemID):
    results = []
    basepath = os.path.join(filepath, str(avatarID))
    for dirpath, dirnames, filenames in os.walk(basepath):
        for fname in filenames:
            try:
                fileNum = int(fname)
            except ValueError:
                pass
            else:
                if fileNum < topItemID:
                    results.append(os.path.join(dirpath, fname))
    return results

def backupEmail(datadir, filedir, backupdir):
    """Copy mail files to another directory.
    """
    st = Store(datadir, filedir)
    try:
        allocs = st.transact(extractAllocations, st)
    finally:
        st.close()

    madeDirs = sets.Set()
    fdirparts = filedir.split(os.sep)
    for (avatarID, topItemID) in allocs.iteritems():
        for fname in collectRequiredFiles(filedir, avatarID, topItemID):
            fparts = fname.split(os.sep)
            uniqPart = fparts[len(os.path.commonprefix([fdirparts, fparts])):]
            dest = os.path.join(backupdir, *uniqPart)
            destDir = os.path.dirname(dest)
            if destDir not in madeDirs:
                madeDirs.add(destDir)
                try:
                    os.makedirs(destDir)
                except:
                    pass
            _copyFile(fname, dest, 2 ** 16)


def _backupMisc(dirname):
    def backup(datadir, filedir, backupdir):
        for avatarID in os.listdir(filedir):
            try:
                avatarID = int(avatarID)
            except ValueError:
                continue

            srcPath = opj(filedir, str(avatarID), dirname)
            dstPath = opj(backupdir, str(avatarID))
            _copyTree(srcPath, dstPath)
    return backup

def _createEmptyIndex(datadir, filedir, backupdir):
    for avatarID in os.listdir(filedir):
        try:
            avatarID = int(avatarID)
        except ValueError:
            continue
        if os.path.exists(opj(filedir, str(avatarID), 'index')):
            dstPath = opj(backupdir, str(avatarID), 'index')
            os.mkdir(dstPath)
        
            

def backupFiles(datadir, filedir, backupdir):
    """Copy all non-db files in use by a database to another directory.

    @param datadir: The path to the database file.
    @param filedir: The path to the non-db files.
    @param backupdir: The path to which to save the backups.
    This directory should not already exist.
    """
    if os.path.exists(backupdir):
        raise BackupDirectoryExists(backupdir)
    try:
        os.makedirs(backupdir)
    except OSError, e:
        raise CannotCreateBackupDirectory(e)

    bfs = (_backupMisc('spambayes'), _backupMisc('cache'),
           _createEmptyIndex, backupEmail)

    for f in bfs:
        f(datadir, filedir, backupdir)

# Handy helper

def reportAndRun(message, f, *a, **kw):
    sys.stdout.write(message + '...')
    sys.stdout.flush()
    try:
        result = f(*a, **kw)
    except:
        sys.stdout.write('Error!\n')
        sys.stdout.flush()
        raise
    else:
        sys.stdout.write('Done\n')
        sys.stdout.flush()
        return result

##
## Be-all, end-all
##

def duplicate(datadir, filedir, backupdir):
    """Create a complete hot copy of an open database.

    @param datadir: The path to the database file.
    @param filedir: The path to the files associated with
    the database.

    @param backupdir: The path to which to save the backups.
    This directory should not already exist.
    """
    # First copy the db itself
    reportAndRun('Backing up database and environment',
                 backupDatabase, datadir, opj(backupdir, 'db'))
    reportAndRun('Recovering database and environment',
                 tidy, opj(backupdir, 'db'))
    reportAndRun('Backing up external file data',
                 backupFiles, opj(backupdir, 'db'),
                 filedir, opj(backupdir, 'files'))


from bisect import bisect

def _modeify(mode, (indexKeyObj, storeID, item)):
    if mode == STORE_ID:
        return storeID
    elif mode == ITEM:
        return item
    elif mode == INDEX:
        return indexKeyObj
    elif mode == INDEX_AND_ITEM:
        return (indexKeyObj, item)
    elif mode == INDEX_AND_STORE_ID:
        return (indexKeyObj, storeID)
    else:
        raise Inconsistency(
            "Invalid result-mode specified for index query: %s" % (mode,))

class DSUQuery(object):
    __slots__ = ['sequence',
                 'mode',
                 'position']
    def __init__(self, sequence, mode):
        self.sequence = sequence
        self.mode = mode
        self.position = 0

    def next(self):
        if self.position >= len(self.sequence):
            raise StopIteration
        result = _modeify(self.mode, self.sequence[self.position])
        self.position += 1
        return result

    def __iter__(self):
        return self


class DSUPool(object):
    """Implement the pool queryIndex method with a decorate-sort-undecorate list in memory.
    """
    implements(IInvertedIndexStorage, IPool)

    name = None
    def __init__(self, sequence, decorators={}, indexes=[], name=None):
        """Create a decorate-sort-undecorate Pool which uses a sequence of
        items and a mapping of decorators.

        @param sequence: a sequence of objects, most likely Item instances (if
        you want self.addIndex to work)

        @param decorators: a mapping of index-tag object to
        one-argument-callable.  Index tags can be any hashable object.
        """

        self.sequence = sequence
        self.decorators = decorators
        for index in indexes:
            self.addIndex(index)
        if name:
            self.name = name

    _noIndexItem = object()

    def getItemIndexKey(self, item, index):
        if index == UIDIndex:
            try:
                return self.sequence.index(item)
            except ValueError:
                return

    def getItemIndexKeys(self, item, index):
        return None

    ## TODO: Provide an API for accessing this that the TDB can use
    indexes = property(lambda self: self.decorators)
    def addIndex(self, index):
        def indexDecorator(item):
            return index.keyToSortableString(item.getIndexKey(self, index, None))
        self.decorators[index] = indexDecorator

    def hasIndex(self, index):
        return self.decorators.has_key(index)

    def iterIndexes(self):
        return self.decorators.iterkeys()

    def removeIndex(self, prefix):
        del self.decorators[prefix]

    def countIndex(self, index):
        # XXX This is not a complete implementation of countIndex().  It is complete
        # enough for the pool browser, though.
        return len(self.sequence)

    def getIndexItem(self, index, indexKey=None, default=_noIndexItem):
        # Should be identical to store.Pool.getIndexItem, but currently isn't
        # because of deprecation crud
        for k, v in self.queryIndex(index, startKey=indexKey, result=INDEX_AND_ITEM):
            # only iterates once, haha.
            if indexKey is None:
                return v
            if k != indexKey:
                if default is DSUPool._noIndexItem:
                    raise KeyError(indexKey)
                return default
            return v
        if default is DSUPool._noIndexItem:
            raise KeyError(indexKey)
        return default

    def queryIndex(self, index=UIDIndex, startKey=None, endKey=None, result=ITEM,
                   backwards=False, limit=None, startItemID=0, endItemID=0):
        """Query this sequence.
        """
        index = _deprecatedIndexLookup(index)
        if index != UIDIndex:
            decorator = self.decorators[index]
            decorated = [(decorator(item), getattr(item, 'storeID', id(item)), item)
                         for item in self.sequence] # [None] * len(self.sequence)
            decorated.sort()
        else:
            decorated = []
            for enum, item in enumerate(self.sequence):
                sortableUID = UIDIndex.keyToSortableString(enum)
                decorated.append((sortableUID, getattr(item, 'storeID', id(item)), item))
        if startKey is not None:
            sortableStartKey = index.keyToSortableString(startKey)
            startIndex = bisect(decorated, (sortableStartKey, startItemID))
        else:
            startIndex = 0
        if endKey is not None:
            sortableEndKey = index.keyToSortableString(endKey)
            endIndex = bisect(decorated, (sortableEndKey, endItemID))
        else:
            endIndex = len(decorated)
        if backwards:
            decorated.reverse()
            if endKey is None:
                endIndex = len(decorated)
            else:
                endIndex = len(decorated) - (endIndex+1)
            if startKey is None:
                startIndex = 0
            else:
                startIndex = len(decorated) - (startIndex+1)
        if startIndex > endIndex:
            raise Inconsistency("%squery starting at %s ending at %s will have no items: specify them in order!." %
                                (backwards and 'backwards ' or '', startIndex, endIndex))
        resultSet = decorated[startIndex:endIndex+1]
        if limit is not None:
            resultSet = resultSet[:limit]
        return DSUQuery([(index.sortableStringToKey(sstr, self), enum, item)
                         for (sstr, enum, item) in resultSet], result)


    def addItem(self, item):
        self.sequence.append(item)

    def removeItem(self, item):
        self.sequence.remove(item)

    def __contains__(self, item):
        return item in self.sequence

    def _getstore(self):
        return context.get('substore')

    store = property(_getstore,
                     doc="""
                     This property is an embarrassing backwards-compatibility
                     hack because so very, very many things depend on the
                     correct presence of a 'store' attribute on pools.  In the
                     future _everything_ should get the store from context, and
                     the argument to Item.__init__ should be eliminated.
                     """)
    storeID = -1

backwardsCompatImplements(DSUPool)


def asyncPoolQuery(query):
    """Exhaust a PoolQuery over multiple reactor iterations.

    Returns a Deferred that will fire with a list containing
    the results of the query.
    """
    L = []
    flag = []

    d = defer.Deferred()
    store = context.get('store')
    assert store, "asyncPoolQuery must be run in a transaction"
    from twisted.internet import reactor

    def appendNext():
        try:
            L.append(query.next())
        except StopIteration:
            flag.append(None)
            return True
        else:
            return False

    def collectSome():
        store.transact(store.batch, appendNext)
        if flag:
            d.callback(L)
        else:
            reactor.callLater(0, collectSome)
    collectSome()

    return d

class _structlike(list):
    __names__ = []
    __slots__ = []

    def _name2slot(self, name):
        return self.__names__.index(name)

    def __init__(self, *args, **kw):
        super(_structlike, self).__init__(args)
        self.extend([None] * (len(self.__names__) - len(args)))
        for k, v in kw.iteritems():
            self[self._name2slot(k)] = v

    def __getattr__(self, attr):
        try:
            return self[self._name2slot(attr)]
        except (IndexError, ValueError):
            raise AttributeError(attr)

    def __setattr__(self, attr, value):
        try:
            self[self._name2slot(attr)] = value
        except ValueError:
            raise AttributeError(attr)


class TieBreaker(_structlike):
    __names__ = ['decoration',
                 'inputOrder',
                 'value',
                 'iterator']

def across(iterators, decorator=lambda i, x: x, uniquify=True,reverse=False):
    """Iterate across several iterators which are assumed to be in the same order.

    @param decorator: a function which returns a result that will sort in the
    same order that the lists are sorted.

    @param iterators: an iterator of iterators

    @param uniquify: return only unique values (uniqueness is decided by the
    return value of decorator, not the items themselves.

    This will yield undefined results if the iterators are not sorted.
    """
    L = []
    for masterTieBreaker, iterable in enumerate(iterators):
        if reverse:
            masterTieBreaker = -masterTieBreaker
        iterator = iter(iterable)
        for value in iterator:
            L.append(TieBreaker(decorator(iterator, value), masterTieBreaker, value, iterator))
            break
    lastdec = None
    while L:
        L.sort()
        if not reverse:
            L.reverse()
        nextb = L[-1]
        if uniquify:
            if lastdec != nextb.decoration:
                yield nextb.value
                lastdec = nextb.decoration
        else:
            yield nextb.value
        for value in nextb.iterator:
            nextb.decoration = decorator(nextb.iterator, value)
            nextb.value = value
            break
        else:
            L.pop()


def multiQueryIndex(pools, index, startKey=None, endKey=None, result=ITEM, backwards=False, *a, **kw):
    """Query many pools on one index.

    The result values _must_ be items.
    """
    assert result == ITEM, "Not implemented, bug glyph"
    queries = [pool.queryIndex(index, startKey, endKey, ITEM, backwards, *a, **kw)
               for pool in pools]
    return list(across(queries, lambda q, x: x.getIndexKey(q.pool, index), reverse=backwards))

