# -*- test-case-name: "atop.test.test_versioning" -*-

from cPickle import Pickler, Unpickler, dumps
from cStringIO import StringIO

import copy_reg, types, gc

from twisted.python import reflect, context, log

PROFILING = False

def _aybabtu(c):
    """
    Return a list of all bases of class 'c', only including bases which are
    themselves subclasses of Versionable but excluding Versionable itself.
    """
    l = []
    for b in reflect.allYourBase(c, Versionable):
        if b not in l and b is not Versionable:
            l.append(b)
    return l

class Versionable(object):
    """A versionable interface.
    """
    persistenceForgets = ()

    def __computeVersionInfo(Class, registry):
        bases = _aybabtu(Class)
        bases.reverse()
        bases.append(Class) # don't forget me!!
        versionInfo = {}
        forgetSlots = []
        rememberSlots = []
        for base in bases:
            if 'persistenceForgets' in base.__dict__:
                forgetSlots.extend(base.persistenceForgets)
            if 'contextRemembers' in base.__dict__:
                rememberSlots.extend(base.contextRemembers)
            basename = reflect.qual(base)
            baseversion = registry.highestVersion(basename)
            if baseversion > 0:
                versionInfo[basename] = baseversion
        return versionInfo, forgetSlots, rememberSlots
    __computeVersionInfo = classmethod(__computeVersionInfo)

    def __currentVersionInfo(Class):
        '''Return a information regarding the expected versionable state of this class.

        A three-tuple is returned.  The first element is a dictionary mapping class names
        to version numbers.  The second element is a "persistence forgets" list.  The
        third element is a "context remembers" list.

        The "persistence forgets" list enumerates attribute names to be omitted from
        persistence.  It is collected from the "persistenceForgets" attribute of all
        the classes at and above this class in the inheritence hierarchy.

        The "context remembers" list consists of two-tuples and is collected from the
        "contextRemembers" attribute of all the class at and above this class in the
        inheritence hiararchy.  The first element of these tuples is a key in the
        context.  The second element of these tuples is the name of an attribute.
        Values are retrieved from the context and bound to their associated attribute
        after instances of this class are unserialized.
        '''
        registry = context.get(UpgradeRegistry)
        assert registry is not None
        currentVersionInfo = registry.getCurrentVersionInfo(Class)
        if currentVersionInfo is None:
            currentVersionInfo = Class.__computeVersionInfo(registry)
            registry.setCurrentVersionInfo(Class, currentVersionInfo)
        return currentVersionInfo

    __currentVersionInfo = classmethod(__currentVersionInfo)

    def __setstate__(self, state):
        """Set state, remembering from context.  If my version doesn't match the one in the pickle, raise an exception.
        """
        myVersionInfo, forgetSlots, rememberSlots = self.__currentVersionInfo()
        if '__version_info__' in state:
            storedVersionInfo = state['__version_info__']
            if myVersionInfo != storedVersionInfo:
                raise VersionMismatch(self.__class__,
                                      myVersionInfo, storedVersionInfo)
        else:
            raise VersionMismatch(self.__class__, myVersionInfo, None)
        self.__dict__ = state
        for contextName, attributeName in rememberSlots:
            setattr(self, attributeName, context.get(contextName))

    def __getstate__(self, state=None):
        """Get state, adding version numbering to it on its way out.

        @param state: None, or a dictionary representing state.
        """
        versionInfo, forgetSlots, rememberSlots = self.__currentVersionInfo()
        if state is None:
            state = self.__dict__
        state = state.copy()
        for forgetSlot in forgetSlots:
            if forgetSlot in state:
                del state[forgetSlot]
        for (contextName, forgetSlot) in rememberSlots:
            if forgetSlot in state:
                del state[forgetSlot]
        state['__version_info__'] = versionInfo
        return state

class VersionMismatch(Exception):
    pass

def unpickleAPI(stringPickle, persistent_load=None, find_global=None, only_once=False):
    u = Unpickler(StringIO(stringPickle))
    if persistent_load:
        u.persistent_load = persistent_load
    if find_global:
        if only_once:
            def fg(module, name):
                del u.find_global
                return find_global(module, name)
            u.find_global = fg
        else:
            u.find_global = find_global
    return u.load()

def pickleAPI(obj, persistent_id=None):
    io = StringIO()
    p = Pickler(io, 2)
    if persistent_id:
        p.persistent_id = persistent_id
    p.dump(obj)
    return io.getvalue()

def _namedProxyGuard(o):
    """This is the nastiest sin I've had to commit against pickle.  It dikes
    out the cursed save_global code path so that our NamedTypeProxy instances
    won't be compared against global types.  Grrrr.
    """
    if isinstance(o, _NamedTypeProxy):
        return reflect.qual(o)

def _getReverter(module, name):
    return Reverter

class Reverter(object):
    def __new__(self):
        return context.get('item')

pickleItem = dumps

def pickleIt(x):
    return dumps(x, 2)

def unpickleItem(stringPickle):
    """Unpickle an item from a string stored in the database, doing revert
    magic if appropriate.
    """
    item = context.get('item')
    if item is not None:
        # We are reverting.  Presumably, this means that the last time
        # that the pickle loaded it was successful...?
        vInfo, forgottenAttributes, rememberAttributes = item.currentVersionInfo()
        oldAttributes = {}
        for attr in forgottenAttributes:
            if attr in item.__dict__:
                value = item.__dict__[attr]
                oldAttributes[attr] = value
        i = loadWithVersions(stringPickle, withRevert=item, isItem=True)
        assert i is item
        i.reactivate(oldAttributes)
        return i
    else:
        return loadWithVersions(stringPickle, isItem=True)

PICKLE_PROTOCOL = 2
PROFILING = False

class UpgradeRegistry(object):
    def __init__(self, upgraders=None, versionInfoCache=None):
        if upgraders is None:
            upgraders = {}
        if versionInfoCache is None:
            versionInfoCache = {}
        self.upgraders = upgraders
        self.versionInfoCache = versionInfoCache

	# Invalidator for cached version info on Versionable classes.
        # Version info needs to be recomputed if registerUpgrader has
        # been called since it was computed.
        self.versionStateValidity = 0

    def clone(self):
        return UpgradeRegistry(self.upgraders.copy(),
                               self.versionInfoCache.copy())

    def registerUpgrader(self, classname, version, requirements, function):
        """Register an upgrade function to run to upgrade a particular class
        from one version to the next.

        @type classname: str
        @param classname: the full classname of the instances that the upgrader function will work on

        @type version: int
        @param version: the version number this function will upgrade classname instances to

        @type requirements: dict
        @param requirements: a dict with {classname: version} items to define what version classname must
        be at before function can run

        @type function: single-argument callable
        @param function: the function that will perform the upgrading
        """
        self.versionStateValidity += 1
        if classname in self.upgraders:
            upgraders = self.upgraders[classname]
        else:
            upgraders = []
            self.upgraders[classname] = upgraders
        upgraders.append((version, requirements, function))
        upgraders.sort()

    def highestVersion(self, classname):
        info = self.upgraders.get(classname)
        if info is None:
            return 0
        return info[-1][0]

    def getCurrentVersionInfo(self, Class):
        info = self.versionInfoCache.get(Class)
        if info is not None:
            validity, version = info
            if validity == self.versionStateValidity:
                return version
        return None

    def setCurrentVersionInfo(self, Class, info):
        self.versionInfoCache[Class] = (self.versionStateValidity, info)

    def upgradeToRequiredVersion(self, inert, classname, storedInfo, inRequiredVersion):
        # inRequiredVersion is the version _to_ which the inert is upgraded.
        if inRequiredVersion == 0:
            return
        for (version, requirements, function) in self.upgraders[classname]:
            if version <= storedInfo.get(classname,0):
                continue
            if version > inRequiredVersion:
                return
            for requiredClass, requiredVersion in requirements.iteritems():
                self.upgradeToRequiredVersion(inert, requiredClass, storedInfo, requiredVersion)
            function(inert)
            # this is actually mutable state from the inert, so it should work
            # out just fine.
            storedInfo[classname] = version

    def _fillOutBases(self, classNames):
        """Given a list of class names, add the names all bases of each class
        which are not already in the list to the list.  (This follows the same
        rules as _aybabtu, e.g. all bases which are subclasses of Versionable
        but not Versionable itself).
        """

        classesByName = {}
        for className in classNames:
            try:
                classesByName[className] = realClass = reflect.namedAny(className)
            except (AttributeError, ImportError, NameError):
                pass
        for realClass in classesByName.values():
            if isinstance(realClass, types.FunctionType):
                continue
            for baseClass in _aybabtu(realClass):
                baseName = reflect.qual(baseClass)
                if not classesByName.has_key(baseName):
                    classesByName[baseName] = baseClass
                    classNames.append(baseName)

    def getBaseCurrentVersions(self, l, s=None):
        self._fillOutBases(l)
        if s is None:
            s = {}
        for j in l:
            if not s.has_key(j) and self.upgraders.has_key(j):
                vers, reqs, func = self.upgraders[j][-1]
                s[j] = vers
                self.getBaseCurrentVersions(reqs.keys(), s)
        return s

    def runUpgraders(self, inert, storedInfo):
        """Run all the upgraders for an inert.

        @param inert: the Inert object for which to run upgraders

        @param storedInfo: The versions the inert is currently, a dict mapping
        fully qualified class names to version number integers.
        """
        # start at the top
        currentVersions = self.getBaseCurrentVersions(storedInfo.keys() + [inert.name])
        if not currentVersions:
            return
        topvers = currentVersions.pop(inert.name, 0)
        items = currentVersions.items()
        items.sort()
        # always upgrade the topmost class first.  beyond that it is
        # deterministic only in the sense that it uses the alphabet; but you
        # shouldn't have any sideways dependencies on state anyway.
        items.insert(0, (inert.name, topvers))
        for classname, targetVersion in items:
            self.upgradeToRequiredVersion(inert, classname, storedInfo, targetVersion)

    def loadWithVersions(self, s, isItem=True, withRevert=None, beforeAfterUpgrade=None):
        if isItem:
            specialFirstFactory = ItemInert
            if withRevert is not None:
                inCaseOfRevert = withRevert.__dict__.copy()
                find_global = _getReverter
                only_once = True
            else:
                find_global = None
                only_once = False
        else:
            specialFirstFactory = None
            find_global = None
            only_once = False
        ctx = {UpgradeRegistry: self}
        try:
            return context.call(ctx, unpickleAPI, s, find_global=find_global, only_once=only_once)
        except VersionMismatch, vm:
            # We need to upgrade.
            pass
        except:
            # Assume that whatever is wrong with the pickle can be
            # fixed: we will have a chance to blow up later, and it
            # will likely be more informative then anyway.
            log.err()
        if withRevert is not None:
            withRevert.__dict__ = inCaseOfRevert
	# If we blow up *here*, something is deeply and grievously wrong with the pickle
        inert = context.call(ctx, unpickleAPI, s, find_global=InertLoader(specialFirstFactory=specialFirstFactory))
        inert.inertUpgrade(self)

        # If we blow up here, there is something *VERY* wrong with the
        # upgrader, it has corrupted an Inert so as to make it
        # unpickleable.
        s2 = pickleAPI(inert, persistent_id=_namedProxyGuard)
        # If we blow up *here*, there is a bug in the upgrader.
        liveItem = context.call(ctx, unpickleAPI, s2, find_global=find_global, only_once=only_once, persistent_load=reflect.namedAny)

        if beforeAfterUpgrade is not None:
            beforeAfterUpgrade(liveItem)

        from atop.store import Item
        if isinstance(liveItem, Item):
            liveItem.store.transact(inert.afterUpgrade, liveItem)
        else:
            inert.afterUpgrade(liveItem)
        return liveItem

    def saveWithVersions(self, o):
        return context.call({UpgradeRegistry: self}, dumps, o, 2)

theUpgradeRegistry = UpgradeRegistry()
registerUpgrader = theUpgradeRegistry.registerUpgrader
loadWithVersions = theUpgradeRegistry.loadWithVersions
saveWithVersions = theUpgradeRegistry.saveWithVersions

def walker(input, seen=None):
    if seen is None:
        seen = {}
    for found in gc.get_referents(input):
        if isinstance(found, (list, dict, Inert)):
            idobj = id(found)
            if idobj not in seen:
                seen[idobj] = found
                yield found
                for subfound in walker(found, seen):
                    yield subfound

class Inert(object):
    def __init__(self, name, newargs=(), calledDirectly=False):
        self.name = name
        self.newargs = newargs
        self.state = None
        self.listitems = None
        self.dictitems = None
        self.calledDirectly = calledDirectly

    def afterUpgrade(self, other):
        pass

    def iterrefs(self, seen):
        for ref in walker(self, seen):
            if isinstance(ref, Inert):
                yield ref

    def inertUpgrade(self, theUpgradeRegistry, seen=None):
        if seen is None:
            seen = {}
        if isinstance(self.state, dict) and self.state.has_key('__version_info__'):
            storedInfo = self.state['__version_info__']
        else:
            storedInfo = {}
            try:
                actualClass = reflect.namedAny(self.name)
            except:
                
                # We'll deal with this later - it's possible to have an object
                # with registered upgraders that can be run later, but which no
                # longer actually exists in current code, I think...?  -gl
                
                actualClass = object
            if isinstance(actualClass, type) or type(actualClass) == types.ClassType:
                # raise NotImplementedError("Not a type: %r " % self.name)
                if issubclass(actualClass, Versionable):
                    self.state['__version_info__'] = storedInfo
            
        theUpgradeRegistry.runUpgraders(self, storedInfo)
        for inert in self.iterrefs(seen):
            inert.inertUpgrade(theUpgradeRegistry, seen)

    def append(self, value):
        if self.listitems is None:
            self.listitems = []
        self.listitems.append(value)

    def extend(self, value):
        for v in value:
            self.append(v)

    def __setitem__(self, key, value):
        if self.dictitems is None:
            self.dictitems = {}
        self.dictitems[key] = value

    def __setstate__(self, state):
        self.state = state

    def __reduce_ex__(self, version):
        """Depends slightly upon global state; assumes that there is an actual
        object in the current runtime that corresponds to what we want to
        pickle.  This is only really a concern in the case of classic classes.
        It's necessary to load actual classes / functions / types because
        cPickle checks to make sure whether the name an object reports is
        actually the same as its current name.
        """
        assert version >= 2
        typeobj = reflect.namedAny(self.name)
        if self.listitems is not None:
            listitems = iter(self.listitems)
        else:
            listitems = None
        if self.dictitems is not None:
            dictitems = self.dictitems.iteritems()
        else:
            dictitems = None
        samestate = (self.state, listitems, dictitems)
        
        if isinstance(typeobj, types.ClassType):
            return (newClassicInstance, (typeobj,) + self.newargs) + samestate
        elif isinstance(typeobj, types.FunctionType):
            return (typeobj, self.newargs) + samestate
        elif not self.calledDirectly:
            return (copy_reg.__newobj__, (typeobj,) + self.newargs) + samestate
        else:
            return (typeobj, self.newargs) + samestate
            
    def __repr__(self):
        return '<%s %s>' % (type(self).__name__, repr(self.__reduce_ex__(2)))

    # ARGH we need to fake out cPickle in so many places
    def _getclass(self):
        return reflect.namedAny(self.name)
    __class__ = property(_getclass)

class ItemInert(Inert):
    def __init__(self,*a,**kw):
        Inert.__init__(self,*a,**kw)
        self.store = context.get('substore') or context.get('store')
        self._afterUpgrade = []

    def callAfterUpgrade(self, interface, methname, *a, **kw):
        """
        Call a method of an interface on the real Item after it's been loaded.  Note
        that if you use this method, your upgrader becomes
        proportionally less durable.  However, no other mechanism has
        been worked out to cover database-state-changing upgrades.
        """
        self._afterUpgrade.append((interface, methname, a, kw))

    def afterUpgrade(self, liveItem):
        Inert.afterUpgrade(self, liveItem)
        liveItem.touch()
        for iface, methname, a, kw in self._afterUpgrade:
	    method = getattr(iface(liveItem), methname)
            method(*a, **kw)


def newClassicInstance(typeobj, *args):
    if isinstance(typeobj, types.ClassType):
        if not args and hasattr(typeobj, "__getinitargs__"):
            return typeobj(*args)
        return types.InstanceType(typeobj)
    else:
        return typeobj(*args)


# There are 4 ways for an object to get manipulated from [c]Pickle.
# __new__ then __setstate__ <-- default __reduce__
# direct call <-- non-default __reduce__, classic class
# .append and .extend <-- list subclass
# .__setitem__ <-- dict subclass


# If it USED to be a classic class, it will call us directly with zero (or
# len(__getinitargs__)) arguments.  we can notice the fact that we were called
# directly, as opposed to invoked from __new__, and return from __reduce__
# something semantically equivalent.  We CANNOT return a __newobj__
# __reduce_ex__ value.  HOWEVER, we can NOT tell the difference between a
# classic class (which should be instantiated with the 'new' module) and an
# actual reduction function (which should just be called).  Best we can do is
# to look at the object in the location in question and guess: if we are called
# directly and it is a classic class then we probably need a
# backwards-compatibility reconstructor which is semantically identical to the
# behavior of the OBJ opcode.

class _NamedTypeProxy(type):
    # Surprise!  It's inside-out boy!
    
    """I am a proxy for a pickleable type somewhere out there in
    Python object space.  I have two attributes: inertFactory and
    inertName, representing the factory to generate Inert instances
    with and the FQPN of the class I am a proxy for, respectively.  DO
    NOT TOUCH ME UNLESS YOU KNOW WHAT YOU ARE DOING.

    XXX TODO: It needs to be possible to change the .inertName attribute
    somehow to indicate that a class has moved.  However this will currently
    change the __hash__ of this _NTP, so it will break dictionary entries.
    When we have to start moving named indexes around, this will need to be
    fixed.  However, there are other, more serious problems with index name
    changes in the database that will have to be resolved first.  See 
    atop.test.etst_versioning.UpgradeTestCase.testClassReferenceChanging for
    an example of how changing .inertName is useful (and for the test that
    will pass when this deficiency is fixed).
    """
    
    def __new__(self, *args):
        return self.inertFactory(self.inertName, args, False)
    def __call__(self, *args):
        return self.inertFactory(self.inertName, args, True)
    def __repr__(self):
        return 'named(%r)' % self.inertName
    def __hash__(self):
        return hash(self.inertName)+1
    def __eq__(self, other):
        return isinstance(other, _NamedTypeProxy) and other.inertName == self.inertName

def _justraise(name, args, b):
    """This should never be called; _NamedTypeProxy instances generated by the
    user will never be called upon to generate instances of themselves, as that
    is an internal API feature for talking to pickle.
    """
    raise NotImplementedError("You can't instantiate _NamedTypeProxy instances you created yourself")

def _innernamed(module, name, factory):
    fqpn = module + '.' + name
    subtype = type.__new__(_NamedTypeProxy, name, (_NamedTypeProxy,), {'__module__': module})
    subtype.inertName = fqpn
    subtype.inertFactory = factory
    return subtype

def named(name):
    """
    Create a proxy for a named type, function, or other global/constant object
    to unpickle in the future.
    """
    names = name.split('.')
    module = '.'.join(names[:-1])
    cname = names[-1]
    return _innernamed(module, cname, _justraise)

class InertLoader(object):
    #Instances of InertLoader are callable with module and name.
    def __init__(self, factory=Inert, specialFirstFactory=None):
        self.calledYet = False
        self.factory = factory
        if specialFirstFactory is None:
            specialFirstFactory = factory
        self.specialFirstFactory = specialFirstFactory

    def __call__(self, module, name):
        if not self.calledYet:
            self.calledYet = True
            factory = self.specialFirstFactory
        else:
            factory = self.factory
        return _innernamed(module, name, factory)

def test():
    global A, upg1
    
    class A(Versionable):
        pass

    a = dumps(A(), 2)

    def upg1(inert):
        inert.state['y'] = 1

    class A(Versionable):
        persistenceVersion = 1
        def __init__(self):
            self.y = 2

    registerUpgrader("__main__.A", 1, {}, upg1)
    upgraded = loadWithVersions(a)
    print upgraded.y == 1
    a2 = dumps(upgraded,2)

    class B(Versionable):
        persistenceVersion = 3

    class A(B, Versionable):
        persistenceVersion = 7
        hello = False
        alltheway = False

    def ato5(inert):
        inert.state['hello'] = True
    def ato7(inert):
        inert.state['alltheway'] = True
    def bto2(inert):
        inert.state['goodbye'] = True
    def bto3(inert):
        inert.state['donebefore'] = True

    registerUpgrader("__main__.A", 5, {'__main__.B': 2}, ato5)
    registerUpgrader("__main__.A", 7, {'__main__.B': 3}, ato7)
    registerUpgrader("__main__.B", 2, {}, bto2)
    registerUpgrader("__main__.B", 3, {}, bto3)

    q = loadWithVersions(a2)
    print q.hello
    print q.alltheway
    print q.goodbye
    print q.donebefore


if __name__ == "__main__":
    test()
